# -*- coding: utf-8 -*-
require 'date'
require 'csv'
require 'sidfm_controller_support'

class SidfmHostsController < ApplicationController
  unloadable

  layout 'sidfm'

  self.main_menu = false

  protect_from_forgery with: :exception

  helper :sidfm

  private

  include SidfmControllerSupport
  before_action :update_session_updated_at, :except => [ :get_heat_map_data ]
  before_action :get_sidfm_roles
  before_action :create_common_data

  before_action :check_license
  before_action :check_roles

  before_action :get_view_state
  after_action :put_view_state

  before_action :get_vm_version
  before_action :clear_import_session
  before_action :clear_export_session

  # 権限チェック
  def check_roles
    unless @sidfm_user then
      redirect_to home_path, :notice => af_log_user('許可されていない操作です。') and return
    end
    unless @sidfm_roles.include?(:view) then
      redirect_to user_path(User.current), :notice => af_log_user('アクセス権限がありません。') and return
    end
    if [ 'update_status', 'update_due_dates', 'send_mail' ].include?(action_name) and !@sidfm_roles.include?(:use) then
      redirect_to sidfm_hosts_path, :notice => af_log_user('アクセス権限がありません。') and return
    end
    @role_admin = @sidfm_roles.include?(:admin)
    @user_category_ids = if @role_admin then
                           nil
                         else
                           SidfmdCategoriesUser.where(:user_id => User.current.id).select(:sidfmd_category_id).order(:sidfmd_category_id).distinct.collect {|x| x[:sidfmd_category_id] }
                         end
    @user_host_ids = if @role_admin then
                       nil
                     else
                       SidfmdCategoriesHost.joins(:sidfmd_host).where(:sidfmd_category_id => @user_category_ids).where(:sidfmd_hosts => { :flg_delete => false }).select(:sidfmd_host_id).order(:sidfmd_host_id).distinct.collect {|x| x[:sidfmd_host_id] }
                     end
    @cache_prefix = @role_admin ? '/hosts/admin/0' : "/hosts/user/#{@user_category_ids.join(",")}"
  end

  #SidfmdHostViewからcontentを生成する filterやsortもクエリで行う
  def create_all_host_view_content(category_id = 0)

    system_cat_ids = SidfmdCategory.where(:is_system => true).pluck(:id)

    host_view = SidfmdHostView.user_host_view(@role_admin, @user_host_ids)

    if @role_admin then
      if category_id > 0 then
        if system_cat_ids.include?(category_id) then
          @hosts[:all_content] =  host_view.where(:system_category_id => category_id)
        else
          @hosts[:all_content] =  host_view.where(
            :id => SidfmdHvCategory.where(:sidfmd_category_id => category_id).select(:sidfmd_host_view_id)
          )
        end
      else
        @hosts[:all_content] = host_view.all
      end
    else
      if category_id > 0 then
        if system_cat_ids.include?(category_id) then
          @hosts[:all_content] =  host_view.where(:system_category_id => category_id)
        elsif @user_category_ids.include?(category_id) then
          @hosts[:all_content] = host_view.where(
            :id => SidfmdHvCategory.where(:sidfmd_category_id => category_id).select(:sidfmd_host_view_id)
          )
        else
          redirect_to user_path(User.current), :notice => af_log_user('アクセス権限がありません。') and return
        end
      else
        @hosts[:all_content] = host_view
      end
    end
    
  end

  # 対象となるコンテンツ全てを含む配列を返す。
  # カテゴリ選択のみ、反映させる。
  def create_all_content(category_id = 0)
    @hosts[:all_content] = SidfmdHost.create_all_content(:category_id => category_id, :role_admin => @role_admin, :user_host_ids => @user_host_ids, :user_category_ids => @user_category_ids, :corresp_required => @corresp_required)
  end

  # パラメータ解析
  def parse_params
    pages_params = params.fetch(:pages, {}).permit(:order, :offset, :per_page, :category_id, :grp, :idx, :filters, :checked, :all_checked)

    @hosts = { :url => request.url }
    @hosts[:order] = (pages_params[:order] || 'hostname_asc')
    @hosts[:filters] = JSON.parse((pages_params[:filters] || '{}'), symbolize_names: true)
    @hosts[:category_id] = (pages_params[:category_id] || 0).to_i
    if @hosts[:category_id] != 0 and SidfmdCategory.system.where(:id => @hosts[:category_id]).size == 0 and !@role_admin and !@user_category_ids.include?(@hosts[:category_id]) then
      @hosts[:category_id] = 0
    end
    @hosts[:grp] = (pages_params[:grp] || 'count').to_sym
    @hosts[:idx] = (pages_params[:idx] || @view_state[:idx].to_s).to_sym
    @view_state[:idx] = @hosts[:idx]
    @hosts[:offset] = (pages_params[:offset] || 0).to_i
    @hosts[:per_page] = (pages_params[:per_page] || 10).to_i
    @hosts[:checked] = JSON.parse(pages_params[:checked] || '[]')
    @hosts[:all_checked] = pages_params[:all_checked] || ''

  end

  def create_content
    # 各種情報取得
    if @hosts[:category_id] > 0 then
      @hosts[:category_name] = SidfmdCategory.find(@hosts[:category_id]).name
    else
      @hosts[:category_name] = '全てのホスト'
    end
    @system_categories = SidfmdCategory.system.with_hosts
    @not_system_categories = SidfmdCategory.not_system.with_hosts
    if !@role_admin then
        @not_system_categories = @not_system_categories.where(:id => @user_category_ids)
    end
    
    @corresp_required = Sidfm.corresp_required
    @risks = SidfmdRisk.all.order(:id => :desc)
    @statuses = IssueStatus.all.order(:id)

    create_all_content(@hosts[:category_id])

    @hosts[:corresp_required_host_size] = @hosts[:all_content].size
    @hosts[:fixed_corresp_required_host_size] = 0
    @hosts[:not_fixed_corresp_required_host_size] = 0
    @hosts[:all_content].each do |x|
      @hosts[:fixed_corresp_required_host_size] += 1 if x[:fixed_warning][@hosts[:idx]]
      @hosts[:not_fixed_corresp_required_host_size] += 1 if x[:not_fixed_warning][@hosts[:idx]]
    end
    @hosts[:all_host_size] = @hosts[:all_content].size
  
  end

  def create_host_view_content
    # 各種情報取得
    if @hosts[:category_id] > 0 then
      @hosts[:category_name] = SidfmdCategory.find(@hosts[:category_id]).name
    else
      @hosts[:category_name] = '全てのホスト'
    end
    @system_categories = SidfmdCategory.system.with_hosts
    @not_system_categories = SidfmdCategory.not_system.with_hosts
    if !@role_admin then
        @not_system_categories = @not_system_categories.where(:id => @user_category_ids)
    end
    @corresp_required = Sidfm.corresp_required
    @risks = SidfmdRisk.all.order(:id => :desc)
    @statuses = IssueStatus.all.order(:id)

    create_all_host_view_content(@hosts[:category_id])
    @hosts[:corresp_required_host_size] = @hosts[:all_content].count

    if @hosts[:idx] == :sri then
      @hosts[:fixed_corresp_required_host_size] = @hosts[:all_content].where(:fixed_warning_sri => true).count
      @hosts[:not_fixed_corresp_required_host_size] = @hosts[:all_content].where(:not_fixed_warning_sri => true).count
    else
      @hosts[:fixed_corresp_required_host_size] = @hosts[:all_content].where(:fixed_warning_cvss => true).count
      @hosts[:not_fixed_corresp_required_host_size] = @hosts[:all_content].where(:not_fixed_warning_cvss => true).count
    end

    @hosts[:all_host_size] = @hosts[:all_content].count
  end

  FILTER_KEYS = {}

  # フィルター情報の初期化
  def init_filter_keys(is_host_view)

    #@due_dates = @hosts[:all_content].inject([]) {|r, x|
    #  r += x[:due_dates]
    #}.uniq.collect {|x| x.nil? ? '' : x }.sort
 
    @dist_versions = if is_host_view then
                       @hosts[:all_content].where('LENGTH(osname) >= 1').order(:osname).distinct.pluck(:osname)
                     else
                       @hosts[:all_content].collect {|x|
                         (x[:os] and x[:os][:osname] and x[:os][:osname].size > 0) ? x[:os][:osname] : nil
                       }.compact.uniq.sort
                     end
    if FILTER_KEYS.size == 0 then
      FILTER_KEYS[:warning] = { :values => [ 'warning', 'others', 'all' ].map(&:freeze).freeze, :default => 'all'.freeze }
      FILTER_KEYS[:status] = {
        :values => (SidfmdIssue::STATUS_ALL.collect {|x| x.to_s } +
                    [ SidfmdIssue::STATUS_UNFINISHED.collect {|x| x.to_s }.join(':') ] +
                    [ SidfmdIssue::STATUS_FINISHED.collect {|x| x.to_s }.join(':') ] +
                    [ '0' ]).map(&:freeze).freeze, :default => '0'.freeze }
      FILTER_KEYS[:sris] = { :values => nil, :default => [].freeze }
      FILTER_KEYS[:cvsses] = { :values => nil, :default => [].freeze }
      FILTER_KEYS[:area] = { :values => (SidfmdArea::VALUES.collect {|x| x.to_s } + [ '' ]).map(&:freeze).freeze, :default => ''.freeze }
      FILTER_KEYS[:exploitable] = { :values => [ 'true', '' ].map(&:freeze).freeze, :default => ''.freeze }
      FILTER_KEYS[:categories] = { 
        :values => nil, 
        :default => {
          :ids => [].freeze,
          :is_and => false,
          :is_except => false
        }
      }
      FILTER_KEYS[:os] = { :values => @dist_versions + [ ''.freeze ], :default => '' }

      FILTER_KEYS[:due_date] = { :value => nil, :default => '::' }

      FILTER_KEYS[:keyword] = { :values => nil, :default => '' }
      FILTER_KEYS[:date] = { :values => nil, :default => '::' }
      FILTER_KEYS[:product] = { :values => nil, :default => {}}
    end
  end

  # フィルターチェック。正規化も行う。
  def check_filters(is_host_view = false)
    init_filter_keys(is_host_view)
    f_keys = FILTER_KEYS.keys
    for i in 0..f_keys.size - 1 do
      if @hosts[:filters].has_key?(f_keys[i]) then
        if FILTER_KEYS[f_keys[i]][:values] && !FILTER_KEYS[f_keys[i]][:values].include?(@hosts[:filters][f_keys[i]]) then
          @hosts[:filters][f_keys[i]] = FILTER_KEYS[f_keys[i]][:default]
        end
      else
        @hosts[:filters][f_keys[i]] = FILTER_KEYS[f_keys[i]][:default]
      end
    end
    @hosts[:filters][:categories][:ids] = @hosts[:filters][:categories][:ids].map(&:to_i)
  end

  DAY_SECONDS = 24 * 60 * 60
  DATE_INVALID = Date.new(4000,12,31).freeze
  DATE_NONE = Date.new(3000,12,31).freeze

  # フィルタの適用
  def apply_filters
    #プロダクト検索は、クエリの発行が少なくなるように初めに該当するプロダクトだけ取得する
    prod_searched = false
    if @hosts[:filters][:product].size > 0 then
      prod_info = @hosts[:filters][:product].symbolize_keys
      prod_searched = true
      if prod_info[:pid] then
        #productが検索対象
        if prod_info[:pid].to_s.index('all') then
          #全バージョン
          prod_info[:pid] = prod_info[:pid].delete('^0-9')
          prods_related_host_ids = SidfmdHostApp.where(:product_id => prod_info[:pid]).map{|app| app.sidfmd_host_id}
        else
          #「全て」のバージョン
          prods_related_host_ids = SidfmdHostApp.where(:product_id => prod_info[:pid])
                                                .where('version_id = 0 OR version_id IS NULL')#バージョンは該当させない
                                                .map{|app| app.sidfmd_host_id}
        end
      elsif prod_info[:vid] then
        #versionが検索対象
        prods_related_host_ids = SidfmdHostApp.where(:version_id => prod_info[:vid]).map{|app| app.sidfmd_host_id}
      elsif prod_info[:lpid] then
        #lproductが検索対象
        if prod_info[:lpid].to_s.index('all') then
          #全バージョン
          prod_info[:lpid] = prod_info[:lpid].delete('^0-9')
          prods_related_host_ids = SidfmdHostLapp.where(:sidfmd_lproduct_id => prod_info[:lpid]).map{|app| app.sidfmd_host_id}
        else
          #「全て」のバージョン  
          prods_related_host_ids = SidfmdHostLapp.where(:sidfmd_lproduct_id => prod_info[:lpid])
                                                 .where(:sidfmd_lversion_id => 0)#バージョンは該当させない
                                                 .map{|app| app.sidfmd_host_id}
        end
      elsif prod_info[:lvid] then
        #lversionが検索対象
        prods_related_host_ids = SidfmdHostLapp.where(:sidfmd_lversion_id => prod_info[:lvid]).map{|app| app.sidfmd_host_id}
      else
        prod_searched = false #リセット時
      end 
    end
    ret = @hosts[:all_content].find_all {|item|
      #プロダクト検索
      if prod_searched then
        next false unless prods_related_host_ids.include?(item[:sidfmd_host_id]) 
      end
      # 要注意
      if @hosts[:filters][:warning] == 'warning' then
        next false unless item[:warning][@hosts[:idx]]
      elsif @hosts[:filters][:warning] == 'others' then
        next false if item[:warning][@hosts[:idx]]
      end
      sri_cvss_status_ids = if @hosts[:filters][:warning] == 'warning' then
                              item[:sri_cvss_status_ids].find_all {|x| case @hosts[:idx] when :sri then x[:sri] >= @corresp_required[:sri]; when :cvss then x[:cvss] ? x[:cvss] >= @corresp_required[:cvss] : true; end }
                            else
                              item[:sri_cvss_status_ids]
                            end
      # 対応状況(new)
      # min_status_idで判断
      if @hosts[:filters][:status] != '0' then
        if /\A(\d):(\d)\z/ =~ @hosts[:filters][:status] then
          status1 = Regexp.last_match(1).to_i
          status2 = Regexp.last_match(2).to_i
          next false unless (item[:min_status_id] == status1 or item[:min_status_id] == status2)
        else
          next false unless item[:min_status_id] == @hosts[:filters][:status].to_i
        end
      end

      # 対応状況(old)
      # sri_cvss_status_idsで判断
=begin
      if @hosts[:filters][:status] != '0' then
        if /\A(\d):(\d)\z/ =~ @hosts[:filters][:status] then
          status1 = Regexp.last_match(1).to_i
          status2 = Regexp.last_match(2).to_i
          next false unless ((sri_cvss_status_ids.size == 0 and (status1 == SidfmdIssue::STATUS_NOTAFFECT or status2 == SidfmdIssue::STATUS_NOTAFFECT)) or
                             (sri_cvss_status_ids.find {|x| x[:status_id] == status1} or sri_cvss_status_ids.find {|x| x[:status_id] == status2 }))
        else
          next false unless
            (sri_cvss_status_ids.size == 0 and @hosts[:filters][:status].to_i == SidfmdIssue::STATUS_NOTAFFECT) or
            (sri_cvss_status_ids.find {|x| x[:status_id] == @hosts[:filters][:status].to_i })
        end
      end
=end
      # 最大SRI
      if @hosts[:filters][:sris].any? then
        next false unless @hosts[:filters][:sris].include?(item[:max_sri].to_s)
      end
      # 最大CVSS
      if @hosts[:filters][:cvsses].any? then
        next false unless @hosts[:filters][:cvsses].include?(item[:max_cvss][:label].to_s)
      end
      # 影響範囲
      if @hosts[:filters][:area] != '' then
        next false unless (!item[:kb] and item[:status_ids].size > 0  and @hosts[:filters][:area].to_i == item[:max_area_id])
      end
      # 攻撃コード
      if @hosts[:filters][:exploitable] != '' then
        next false unless item[:exploitable]
      end
      #カテゴリ(複数)
      if @hosts[:filters][:categories][:ids].any? then
        pass = false
        host_all_cat_ids = item[:no_system_categories].map{|cc| cc[:id]}.push(item[:system_category][:id])
        if @hosts[:filters][:categories][:is_and] then
          if (@hosts[:filters][:categories][:ids] - host_all_cat_ids).empty? then
            pass = true
          end
        elsif @hosts[:filters][:categories][:is_except] then
          if (@hosts[:filters][:categories][:ids] - host_all_cat_ids).size == @hosts[:filters][:categories][:ids].size then
            pass = true
          end
        else
          if (@hosts[:filters][:categories][:ids] - host_all_cat_ids).size != @hosts[:filters][:categories][:ids].size then
            pass = true
          end
        end
        next false unless pass
      end
      # OS
      if @hosts[:filters][:os] != '' then
        osname = "#{item[:os] ? item[:os][:osname] : ''}"
        next false unless (item[:os] and osname and osname == @hosts[:filters][:os])
      end
      # 対策予定日設定済
      if @hosts[:filters][:due_date] != '::' then
        if /\A(|notset):(|\d{4}\/\d{2}\/\d{2}):(|\d{4}\/\d{2}\/\d{2})\z/ =~ @hosts[:filters][:due_date] then
          notset = ('notset' == Regexp.last_match(1))
          begindate = Regexp.last_match(2)
          enddate = Regexp.last_match(3)
          if notset then
            next false unless item[:due_dates].include?(nil)
          else
            begin
              begindate = Date.parse(begindate)
            rescue Exception => e
              begindate = nil
            end
            begin
              enddate = Date.parse(enddate)
            rescue Exception => e
              enddate = nil
            end
            due_date = nil
            if begindate and enddate then
              flag = false
              item[:due_dates].each do |due_date|
                next unless due_date
                d_due_date = nil
                begin
                  d_due_date = Date.parse(due_date)
                rescue Exception => e
                  d_due_date = nil
                end
                next unless d_due_date
                if d_due_date >= begindate and d_due_date <= enddate then
                  flag = true
                  break
                end
              end
              next false unless flag
            elsif begindate then
              flag = false
              item[:due_dates].each do |due_date|
                next unless due_date
                d_due_date = nil
                begin
                  d_due_date = Date.parse(due_date)
                rescue Exception => e
                  d_due_date = nil
                end
                next unless d_due_date
                if d_due_date >= begindate then
                  flag = true
                  break
                end
              end
              next false unless flag
            elsif enddate then
              flag = false
              item[:due_dates].each do |due_date|
                next unless due_date
                d_due_date = nil
                begin
                  d_due_date = Date.parse(due_date)
                rescue Exception => e
                  d_due_date = nil
                end
                next unless d_due_date
                if d_due_date <= enddate then
                  flag = true
                  break
                end
              end
              next false unless flag
            end
          end
        end
      end
      # 日付指定検索
      if @hosts[:filters][:date] != '::' then
        if /^(in|rev):(\d{4}\/\d{2}\/\d{2}|):(\d{4}\/\d{2}\/\d{2}|)$/ =~ @hosts[:filters][:date] then
          orig_flg_rev = Regexp.last_match(1)
          orig_begin_date = Regexp.last_match(2)
          orig_end_date = Regexp.last_match(3)
          if orig_flg_rev.size > 0 and orig_begin_date.size == 0 and orig_end_date.size == 0 then
            @hosts[:filters][:date] = '::'
            next false
          end
          flg_rev = orig_flg_rev == 'rev' ? true : false
          begin_date = orig_begin_date.size > 0 ? Date.strptime(orig_begin_date, '%Y/%m/%d') : nil
          end_date = orig_end_date.size > 0 ? Date.strptime(orig_end_date, '%Y/%m/%d') : nil
          if flg_rev then
            if begin_date and end_date then
              begin_date = begin_date.to_time
              end_date = end_date.to_time + DAY_SECONDS - 1
              next false unless item[:created_ons].find {|x| begin_date >= x or end_date <= x }
           elsif begin_date then
              begin_date = begin_date.to_time
              next false unless item[:created_ons].find {|x| begin_date >= x }
            elsif end_date then
              end_date = end_date.to_time + DAY_SECONDS - 1
              next false unless item[:created_ons].find {|x| end_date <= x }
            else
              # last resort
              @hosts[:filters][:date] = '::'
              next false
            end
          else
            if begin_date and end_date then
              begin_date = begin_date.to_time
              end_date = end_date.to_time + DAY_SECONDS - 1
              next false unless item[:created_ons].find {|x| begin_date <= x and end_date >= x }
            elsif begin_date then
              begin_date = begin_date.to_time
              next false unless item[:created_ons].find {|x| begin_date <= x }
            elsif end_date then
              end_date = end_date.to_time + DAY_SECONDS - 1
              next false unless item[:created_ons].find {|x| end_date >= x }
            else
              # last resort
              @hosts[:filters][:date] = '::'
              next false
            end
          end
        else
          @hosts[:filters][:date] = '::'
          next false
        end
      end
      # ホスト名検索
      if @hosts[:filters][:keyword] != '' then
        regexp = Regexp.new(@hosts[:filters][:keyword], Regexp::IGNORECASE)
        next false unless regexp =~ item[:hostname] || regexp =~ item[:display_name] 
      end
      true
    }
    # ソート処理
    case @hosts[:order]
    when 'warning_asc' then
      temp_ev = proc {|x|
        [ x[:display_name], !x[:warning][@hosts[:idx]] ? 100 : 500 ]
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'warning_desc' then
      temp_ev = proc {|x|
        [ x[:display_name], !x[:warning][@hosts[:idx]] ? 100 : 500 ]
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'statuses_asc' then
      temp_ev = proc {|x|
        m = x[:status_ids].min
        if m then
          [ m, 500, x[:display_name] ]
        else
          [ '', 0, x[:display_name] ]
        end
        #if m then
        #  [ m, 500 ]
        #else
        #  [ '', 0 ]
        #end
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
          if r == 0 then
            r = xx[2] <=> yy[2]
          end
        end
        r
      }
    when 'statuses_desc' then
      temp_ev = proc {|x|
        m = x[:status_ids].min
        if m then
          [ m, 500, x[:display_name] ]
        else
          [ '', 0, x[:display_name] ]
        end
        #if m then
        #  [ m, 500 ]
        #else
        #  [ '', 0 ]
        #end
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
          if r == 0 then
            r = xx[2] <=> yy[2]
          end
        end
        r
      }
    when 'maxsri_asc' then
      temp_ev = proc {|x|
        [ x[:display_name], x[:max_sri] ]
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'maxsri_desc' then
      temp_ev = proc {|x|
        [ x[:display_name], x[:max_sri] ]
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'maxcvss_asc' then
      temp_ev = proc {|x|
        [ x[:display_name], x[:max_cvss][:score] ? x[:max_cvss][:score] : 0.0 ]
      }
      ret.sort! {|x, y| 
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'maxcvss_desc' then
      temp_ev = proc {|x|
        [ x[:display_name], x[:max_cvss][:score] ? x[:max_cvss][:score] : 0.0 ]
      }
      ret.sort! {|y, x| 
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'area_asc' then
      temp_ev = proc {|x|
        [
         x[:display_name],
         x[:kb] ? -50 : (x[:status_ids].size > 0 ? x[:max_area_id] : -100),
        ]
      }
      ret.sort! {|x, y| 
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'area_desc' then
      temp_ev = proc {|x|
        [
         x[:display_name],
         x[:kb] ? -50 : (x[:status_ids].size > 0 ? x[:max_area_id] : -100),
        ]
      }
      ret.sort! {|y, x| 
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'exploitable_asc' then
      temp_ev = proc {|x|
        [ x[:display_name],
          x[:kb] ? 25 : (x[:status_ids].size > 0 ? (x[:max_cvss][:label] != :notreg ? (x[:exploitable] ? 500 : 100) : 50) : 15) ]
      }
      ret.sort! {|x, y| 
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'exploitable_desc' then
      temp_ev = proc {|x|
        [ x[:display_name],
          x[:kb] ? 25 : (x[:status_ids].size > 0 ? (x[:max_cvss][:label] != :notreg ? (x[:exploitable] ? 500 : 100) : 50) : 15) ]
      }
      ret.sort! {|y, x| 
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'hostname_asc' then
      ret.sort! {|x, y| x[:display_name].downcase <=> y[:display_name].downcase }
    when 'hostname_desc' then
      ret.sort! {|x, y| y[:display_name].downcase <=> x[:display_name].downcase }
    when 'categories_asc' then
      temp_ev = proc {|x|
        [ x[:display_name], x[:system_category] ? x[:system_category][:id] : 99999 ]
      }
      ret.sort! {|x, y| 
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[1] <=> yy[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'categories_desc' then
      temp_ev = proc {|x|
        [ x[:display_name], x[:system_category] ? x[:system_category][:id] : 99999 ]
      }
      ret.sort! {|y, x| 
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[1] <=> yy[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'os_asc' then
      temp_ev = proc {|x|
        osname = x[:os] ? (x[:os][:osname] ? x[:os][:osname] : '') : ''
        [ x[:display_name], osname, (osname == '' or osname == '---') ? 500 : 0 ]
      }
      ret.sort! {|x, y| 
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[2] <=> yy[2]
        if r == 0 then
          r = xx[1] <=> yy[1]
          if r == 0 then
            r = xx[0] <=> yy[0]
          end
        end
        r
      }
    when 'os_desc' then
      temp_ev = proc {|x|
        osname = x[:os] ? (x[:os][:osname] ? x[:os][:osname] : '') : ''
        [ x[:display_name], osname, (osname == '' or osname == '---') ? 500 : 0 ]
      }
      ret.sort! {|y, x| 
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[2] <=> yy[2]
        if r == 0 then
          r = xx[1] <=> yy[1]
          if r == 0 then
            r = xx[0] <=> yy[0]
          end
        end
        r
      }
    when 'duedates_asc' then
      temp_ev = proc {|x|
        p = view_context.display_due_dates_percent_value(x[:due_dates_unfinished])
        [ x[:display_name], (p ? p : -1)]
      }
      ret.sort! {|x, y| 
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'duedates_desc' then
      temp_ev = proc {|x|
        p = view_context.display_due_dates_percent_value(x[:due_dates_unfinished])
        [ x[:display_name], (p ? p : -1) ]
      }
      ret.sort! {|y, x| 
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[0] <=> yy[0]
        end
        r
      }
    when 'finished_asc' then
      temp_ev = proc {|x|
        p = x[:status_ids].size > 0 ? view_context.display_finished_percent_value(x[:status_ids]) : nil
        [ x[:display_name], p ? p : 0.0, p ? 100 : 500 ]
      }
      ret.sort! {|x, y| 
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[2] <=> yy[2]
        if r == 0 then
          r = yy[1] <=> xx[1]
          if r == 0 then
            r = xx[0] <=> yy[0]
          end
        end
        r
      }
    when 'finished_desc' then
      temp_ev = proc {|x|
        p = x[:status_ids].size > 0 ? view_context.display_finished_percent_value(x[:status_ids]) : nil
        [ x[:display_name], p ? p : 0.0, p ? 100 : 500 ]
      }
      ret.sort! {|y, x| 
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[2] <=> yy[2]
        if r == 0 then
          r = yy[1] <=> xx[1]
          if r == 0 then
            r = xx[0] <=> yy[0]
          end
        end
        r
      }
    end
    ret
  end

  def apply_filters_to_host_view_content
    content = @hosts[:all_content]
  
    #プロダクト検索
    if @hosts[:filters][:product].size > 0 then
      prods_related_host_ids = nil
      prod_info = @hosts[:filters][:product].symbolize_keys
      if prod_info[:pid] then
        #productが検索対象
        if prod_info[:pid].to_s.index('all') then
          #全バージョン
          prod_info[:pid] = prod_info[:pid].delete('^0-9')
          prods_related_host_ids = SidfmdHostApp.where(:product_id => prod_info[:pid]).pluck(:sidfmd_host_id)
        else
          #「全て」のバージョン
          prods_related_host_ids = SidfmdHostApp.where(:product_id => prod_info[:pid])
                                                .where('version_id = 0 OR version_id IS NULL')#バージョンは該当させない
                                                .pluck(:sidfmd_host_id)
        end
      elsif prod_info[:vid] then
        #versionが検索対象
        prods_related_host_ids = SidfmdHostApp.where(:version_id => prod_info[:vid]).pluck(:sidfmd_host_id)
      elsif prod_info[:lpid] then
        #lproductが検索対象
        if prod_info[:lpid].to_s.index('all') then
          #全バージョン
          prod_info[:lpid] = prod_info[:lpid].delete('^0-9')
          prods_related_host_ids = SidfmdHostLapp.where(:sidfmd_lproduct_id => prod_info[:lpid]).pluck(:sidfmd_host_id)
        else
          #「全て」のバージョン
          prods_related_host_ids = SidfmdHostLapp.where(:sidfmd_lproduct_id => prod_info[:lpid])
                                                 .where(:sidfmd_lversion_id => 0)#バージョンは該当させない
                                                 .pluck(:sidfmd_host_id)
        end
      elsif prod_info[:lvid] then
        #lversionが検索対象
        prods_related_host_ids = SidfmdHostLapp.where(:sidfmd_lversion_id => prod_info[:lvid]).pluck(:sidfmd_host_id)
      end
      
      if prods_related_host_ids then
        content = content.where(:sidfmd_host_id => prods_related_host_ids)
      end
    end

    # 要注意
    if @hosts[:filters][:warning] == 'warning' then
      if @hosts[:idx] == :sri then
        content = content.where(:warning_sri => true)
      else
        content = content.where(:warning_cvss => true)
      end
    elsif @hosts[:filters][:warning] == 'others' then
      if @hosts[:idx] == :sri then
        content = content.where.not(:warning_sri => true)
      else
        content = content.where.not(:warning_cvss => true)
      end
    end

    # 対応状況(new) 
    # min_status_idで判断
    if @hosts[:filters][:status] != '0' then
      if /\A(\d):(\d)\z/ =~ @hosts[:filters][:status] then
        status1 = Regexp.last_match(1).to_i
        status2 = Regexp.last_match(2).to_i
        content = content.where('min_status_id IN (?)', [status1, status2])
      else
        content = content.where(:min_status_id => @hosts[:filters][:status].to_i)
      end
    end

    # 最大SRI
    if @hosts[:filters][:sris].any? then
      content = content.where(:max_sri => @hosts[:filters][:sris])
    end

    # 最大CVSS
    if @hosts[:filters][:cvsses].any? then
      content = content.where(:max_cvss_label => @hosts[:filters][:cvsses].map{|cvss| cvss.to_s})
    end 

    # 影響範囲
    if @hosts[:filters][:area] != '' then
      content = content.where(:id => SidfmdHvScsid.where('status_id > 0').select(:sidfmd_host_view_id))
                       .where(:kb => false)
                       .where(:max_area_id => @hosts[:filters][:area].to_i)
    end

    # 攻撃コード
    if @hosts[:filters][:exploitable] != '' then
      content = content.where(:exploitable => true)
    end

    #カテゴリ(複数)
    if @hosts[:filters][:categories][:ids].any? then
      target_host_ids = []
      if @hosts[:filters][:categories][:is_and] then
        SidfmdHost.where(
          :id => SidfmdCategoriesHost.where(:sidfmd_category_id => @hosts[:filters][:categories][:ids]).select(:sidfmd_host_id)
        ).where(:flg_delete => false)
        .each do |host|
          host_cat_ids = host.sidfmd_categories.pluck(:id)
          if (@hosts[:filters][:categories][:ids] - host_cat_ids).empty? then
            target_host_ids << host.id
          end
        end
      else
        target_host_ids = SidfmdCategoriesHost.where(:sidfmd_category_id => @hosts[:filters][:categories][:ids])
                                              .select(:sidfmd_host_id)
      end
      content = if @hosts[:filters][:categories][:is_except] then
                  content.where.not(:sidfmd_host_id => target_host_ids)
                else
                  content.where(:sidfmd_host_id => target_host_ids)
                end
    end

    # OS
    if @hosts[:filters][:os] != '' then
      content = content.where(:osname => @hosts[:filters][:os])
    end

    # 対策予定日設定検索
    if @hosts[:filters][:due_date] != '::' then
      if /\A(|notset):(|\d{4}\/\d{2}\/\d{2}):(|\d{4}\/\d{2}\/\d{2})\z/ =~ @hosts[:filters][:due_date] then
        notset = ('notset' == Regexp.last_match(1))
        begindate = Regexp.last_match(2)
        enddate = Regexp.last_match(3)
        if notset then
          #対策予定が設定されていない脆弱性をもつホスト
          content = content.where(:id => SidfmdHvDueDate.where(:d => nil).select(:sidfmd_host_view_id)) 
        else
          #対策予定が設定されている脆弱性をもつホスト
          begin
            begindate = Date.parse(begindate)
          rescue Exception => e
            begindate = nil
          end
          begin
            enddate = Date.parse(enddate)
          rescue Exception => e
            enddate = nil
          end
          due_date = nil
          if begindate and enddate then
            content = content.where(:id => SidfmdHvDueDate.where(:d => begindate..enddate).select(:sidfmd_host_view_id))
          elsif begindate then
            content = content.where(:id => SidfmdHvDueDate.where('d >= ?', begindate).select(:sidfmd_host_view_id))
          elsif enddate then
            content = content.where(:id => SidfmdHvDueDate.where('d <= ?', enddate).select(:sidfmd_host_view_id))
          end
        end
      end
    end

    # 日付指定検索
    if @hosts[:filters][:date] != '::' then
      if /^(in|rev):(\d{4}\/\d{2}\/\d{2}|):(\d{4}\/\d{2}\/\d{2}|)$/ =~ @hosts[:filters][:date] then
        orig_flg_rev = Regexp.last_match(1)
        orig_begin_date = Regexp.last_match(2)
        orig_end_date = Regexp.last_match(3)
        if orig_flg_rev.size > 0 and orig_begin_date.size == 0 and orig_end_date.size == 0 then
          @hosts[:filters][:date] = '::'
        else
          flg_rev = orig_flg_rev == 'rev' ? true : false
          begin_date = orig_begin_date.size > 0 ? Date.strptime(orig_begin_date, '%Y/%m/%d') : nil
          end_date = orig_end_date.size > 0 ? Date.strptime(orig_end_date, '%Y/%m/%d') : nil
          if flg_rev then
            if begin_date and end_date then
              begin_date = begin_date.to_time
              end_date = end_date.to_time + DAY_SECONDS - 1
              content = content.where(:id => SidfmdHvCreatedOn.where('dt <= ? OR dt >= ?', begin_date, end_date).select(:sidfmd_host_view_id))
            elsif begin_date then
              begin_date = begin_date.to_time
              content = content.where(:id => SidfmdHvCreatedOn.where('dt <= ?', begin_date).select(:sidfmd_host_view_id))
            elsif end_date then
              end_date = end_date.to_time + DAY_SECONDS - 1
              content = content.where(:id => SidfmdHvCreatedOn.where('dt >= ?', end_date).select(:sidfmd_host_view_id))
            else
              # last resort
              @hosts[:filters][:date] = '::'
            end
          else
            if begin_date and end_date then
              begin_date = begin_date.to_time
              end_date = end_date.to_time + DAY_SECONDS - 1
              content = content.where(:id => SidfmdHvCreatedOn.where(:dt => begin_date..end_date).select(:sidfmd_host_view_id))
            elsif begin_date then
              begin_date = begin_date.to_time
              content = content.where(:id => SidfmdHvCreatedOn.where('dt >= ?', begin_date).select(:sidfmd_host_view_id))
            elsif end_date then
              end_date = end_date.to_time + DAY_SECONDS - 1
              content = content.where(:id => SidfmdHvCreatedOn.where('dt <= ?', end_date).select(:sidfmd_host_view_id))
            else
              # last resort
              @hosts[:filters][:date] = '::'
            end
          end
        end
      else
        @hosts[:filters][:date] = '::'
      end
    end
    
    #ホスト名検索
    if @hosts[:filters][:keyword] != '' then
      kwd = @hosts[:filters][:keyword]
      content = content.where('hostname ILIKE ? OR display_name ILIKE ?', "%#{ActiveRecord::Base.send(:sanitize_sql_like, kwd)}%", "%#{ActiveRecord::Base.send(:sanitize_sql_like, kwd)}%")
    end

    #ソート処理
    case @hosts[:order]
    when 'warning_asc' then
      content = content.order("warning_#{@hosts[:idx].to_s} DESC").order('LOWER(display_name COLLATE "C")')
    when 'warning_desc' then
      content = content.order("warning_#{@hosts[:idx].to_s}").order('LOWER(display_name COLLATE "C") DESC')
    when 'statuses_asc' then
      content = content.order(:min_status_id).order('LOWER(display_name COLLATE "C")')
    when 'statuses_desc' then
      content = content.order(:min_status_id => :desc).order('LOWER(display_name COLLATE "C") DESC')
    when 'maxsri_asc' then
      content = content.order(:max_sri => :desc).order('LOWER(display_name COLLATE "C")')
    when 'maxsri_desc' then
      content = content.order(:max_sri).order('LOWER(display_name COLLATE "C") DESC')
    when 'maxcvss_asc', 'maxcvss_desc' then
      content = content.order(%|
                  CASE 
                  WHEN max_cvss_score IS NULL THEN
                    0.0
                  ELSE
                    max_cvss_score
                  END
                  #{@hosts[:order] == 'maxcvss_asc' ? 'DESC': 'ASC'}
                |)
                .order(%| LOWER(display_name COLLATE "C")  #{@hosts[:order] == 'maxcvss_asc' ? 'ASC': 'DESC'} |)
    when 'area_asc' then
      content = content.order("CASE min_status_id WHEN #{SidfmdIssue::STATUS_NOTAFFECT} THEN 2 ELSE 1 END")
                       .order(:max_area_id => :desc)
                       .order('LOWER(display_name COLLATE "C")')
    when 'area_desc' then
      content = content.order("CASE min_status_id WHEN #{SidfmdIssue::STATUS_NOTAFFECT} THEN 1 ELSE 2 END")
                       .order(:max_area_id)
                       .order('LOWER(display_name COLLATE "C") DESC')
    when 'exploitable_asc', 'exploitable_desc' then
      content = content.order(%|
        CASE
        WHEN max_cvss_label != 'notreg' AND exploitable IS TRUE THEN
          1
        WHEN exploitable IS FALSE THEN
          2
        ELSE
          3
        END
        #{@hosts[:order] == 'exploitable_asc' ? 'ASC' : 'DESC'}
      |).order(%| LOWER(display_name COLLATE "C") #{@hosts[:order] == 'exploitable_asc' ? 'ASC' : 'DESC'}|)
    when 'hostname_asc' then
      content = content.order('LOWER(display_name COLLATE "C")')
    when 'hostname_desc' then
      content = content.order('LOWER(display_name COLLATE "C") DESC')
    when 'categories_asc' then
      content = content.order(:system_category_id).order('LOWER(display_name COLLATE "C")')
    when 'categories_desc' then
      content = content.order(:system_category_id => :desc).order('LOWER(display_name COLLATE "C") DESC')
    when 'os_asc' then
      content = content.order(:osname).order('LOWER(display_name COLLATE "C")')
    when 'os_desc' then
      content = content.order(:osname => :desc).order('LOWER(display_name COLLATE "C") DESC')
    when 'duedates_asc' then
      content = content.order('ROUND(CAST(duedates_percent as numeric), 1) ASC')
                       .order('LOWER(display_name COLLATE "C")') 
    when 'duedates_desc' then
      content = content.order('ROUND(CAST(duedates_percent as numeric), 1) DESC') 
                       .order('LOWER(display_name COLLATE "C") DESC')
    when 'finished_asc' then
      content = content.order("CASE min_status_id WHEN #{SidfmdIssue::STATUS_NOTAFFECT} THEN 1 ELSE 0 END")
                       .order('ROUND(CAST(fixed_percent as numeric), 1) ASC')
                       .order('CASE fixed_percent WHEN NULL THEN 1 ELSE 0 END')
                       .order('LOWER(display_name COLLATE "C")')
    when 'finished_desc' then
      content = content.order("CASE min_status_id WHEN #{SidfmdIssue::STATUS_NOTAFFECT} THEN 0 ELSE 1 END")
                       .order('ROUND(CAST(fixed_percent as numeric), 1) DESC')
                       .order('CASE fixed_percent WHEN NULL THEN 0 ELSE 1 END')
                       .order('LOWER(display_name COLLATE "C") DESC')
    end
 
    content
  end

  TIMELINE_SPAN = 6
  TIMELINE_SPAN_UNIT = :month

  # タイムラインのデータを返す
  def get_timeline_data
    ret = {}
    ret[:now] = Time.current
    ret[:localtime] = ret[:now].localtime
    ret[:timescale] = []
    TIMELINE_SPAN.times do |i|
      z = ret[:localtime].ago((TIMELINE_SPAN - 1 - i).send(TIMELINE_SPAN_UNIT))
      r = {
        :time => z,
        :year => z.year,
        :month => z.month,
        :step => TIMELINE_SPAN - 1 - i,
        :is_first => (i == 0),
        :is_last => (i == (TIMELINE_SPAN - 1)),
        :name => z.strftime('%Y/%m'),
      }
      ret[:timescale] << r
    end
    ret[:status_counts] = {
      SidfmdIssue::STATUS_NOTCONFIRMED => [0] * ret[:timescale].size,
      SidfmdIssue::STATUS_ONGOING      => [0] * ret[:timescale].size,
      SidfmdIssue::STATUS_FIXED        => [0] * ret[:timescale].size,
      SidfmdIssue::STATUS_NOTAFFECT    => [0] * ret[:timescale].size,
    }
    ret[:status_names] = {}
    SidfmdIssue::STATUS_ALL.each do |s|
      ret[:status_names][s] = SidfmdIssue.status_id_to_s(s)
    end
    case @hosts[:idx]
    when :sri then
      ret[:risk_syms] = {
        0 => 'none', 1 => 'low', 2 => 'medium', 3 => 'high', 4 => 'critical', 5 => 'fixed',
      }
      ret[:risk_names] = {
        0 => '無', 1 => '低', 2 => '中', 3 => '高', 4 => '重大', 5 => '済',
      }
      ret[:risk_counts] = ret[:risk_syms].each_key.collect {|x|
        [0] * ret[:timescale].size
      }
    when :cvss then
      ret[:risk_labels] = [ :notreg, :d4, :d3, :d2, :d1, :fixed ]
      ret[:risk_names] = {
        :notreg => '未登録',
        :d4 => '10.0〜9.0',
        :d3 => '8.9〜7.0',
        :d2 => '6.9〜4.0',
        :d1 => '3.9〜0.0',
        :fixed => '済',
      }
      ret[:risk_counts] = {}
      ret[:risk_labels].each do |i|
        ret[:risk_counts][i] = [0] * ret[:timescale].size
      end
    end
    flg_fixed = (SidfmdHostTimeline.column_names.include?('sri_fixed') && SidfmdHostTimeline.column_names.include?('cvss_fixed'))
    ret[:timescale].each do |timescale|
      stat = SidfmdHostTimeline.where('dt >= ?', timescale[:time])
      unless timescale[:is_first] then
        stat = stat.where('dt <= ?', ret[:timescale][timescale[:step] - 1][:time])
      end
      stat = stat.order(:dt => :desc).limit(1).first
      if stat then
        ret[:status_counts][SidfmdIssue::STATUS_NOTCONFIRMED][timescale[:step]] = stat.status_notconfirmed
        ret[:status_counts][SidfmdIssue::STATUS_ONGOING][timescale[:step]] = stat.status_ongoing
        ret[:status_counts][SidfmdIssue::STATUS_FIXED][timescale[:step]] = stat.status_fixed
        ret[:status_counts][SidfmdIssue::STATUS_NOTAFFECT][timescale[:step]] = stat.status_notaffect
        case @hosts[:idx]
        when :sri then
          ret[:risk_counts][5][timescale[:step]] = stat.sri_fixed if flg_fixed
          ret[:risk_counts][4][timescale[:step]] = stat.sri_critical
          ret[:risk_counts][3][timescale[:step]] = stat.sri_high
          ret[:risk_counts][2][timescale[:step]] = stat.sri_medium
          ret[:risk_counts][1][timescale[:step]] = stat.sri_low
          ret[:risk_counts][0][timescale[:step]] = stat.sri_none
        when :cvss then
          ret[:risk_counts][:fixed][timescale[:step]] = stat.cvss_fixed if flg_fixed
          ret[:risk_counts][:d4][timescale[:step]] = stat.cvss_d4
          ret[:risk_counts][:d3][timescale[:step]] = stat.cvss_d3
          ret[:risk_counts][:d2][timescale[:step]] = stat.cvss_d2
          ret[:risk_counts][:d1][timescale[:step]] = stat.cvss_d1
          ret[:risk_counts][:notreg][timescale[:step]] = stat.cvss_notreg
        end
      else
        ret[:status_counts][SidfmdIssue::STATUS_NOTCONFIRMED][timescale[:step]] = 0
        ret[:status_counts][SidfmdIssue::STATUS_ONGOING][timescale[:step]] = 0
        ret[:status_counts][SidfmdIssue::STATUS_FIXED][timescale[:step]] = 0
        ret[:status_counts][SidfmdIssue::STATUS_NOTAFFECT][timescale[:step]] = 0
        case @hosts[:idx]
        when :sri then
          ret[:risk_syms].each_key do |sri|
            ret[:risk_counts][sri][timescale[:step]] = 0
          end
        when :cvss then
          ret[:risk_labels].each do |cvss|
            ret[:risk_counts][cvss][timescale[:step]] = 0
          end
        end
      end
    end

    ret
  end

  def filter_content
    @hosts[:distribution_status] = { :notconfirmed => 0, :ongoing => 0, :fixed => 0, :notaffect => 0 }
    @hosts[:distribution_risk] =
      case @hosts[:idx]
      when :sri then
        { :critical => 0, :high => 0, :medium => 0, :low => 0, :none => 0, :fixed => 0, }
      when :cvss then
        { :notreg => 0, :d4 => 0, :d3 => 0, :d2 => 0, :d1 => 0, :fixed => 0, }
      end
    @hosts[:all_content].each do |i|
        case i[:min_status_id]
        when SidfmdIssue::STATUS_NOTCONFIRMED then @hosts[:distribution_status][:notconfirmed] += 1
        when SidfmdIssue::STATUS_ONGOING then @hosts[:distribution_status][:ongoing] += 1
        when SidfmdIssue::STATUS_FIXED then @hosts[:distribution_status][:fixed] += 1
        when SidfmdIssue::STATUS_NOTAFFECT then @hosts[:distribution_status][:notaffect] += 1
        else @hosts[:distribution_status][:notaffect] += 1
        end
      if SidfmdIssue::STATUS_UNFINISHED.include?(i[:min_status_id]) then
        case @hosts[:idx]
        when :sri then
          case i[:max_sri]
          when SidfmdIssueSri::VALUES[0] then
            @hosts[:distribution_risk][:critical] += 1
          when SidfmdIssueSri::VALUES[1] then
            @hosts[:distribution_risk][:high] += 1
          when SidfmdIssueSri::VALUES[2] then
            @hosts[:distribution_risk][:medium] += 1
          when SidfmdIssueSri::VALUES[3] then
            @hosts[:distribution_risk][:low] += 1
          when SidfmdIssueSri::VALUES[4] then
            @hosts[:distribution_risk][:none] += 1
          end
        when :cvss then
          case i[:max_cvss][:label]
          when :notreg then
            @hosts[:distribution_risk][:notreg] += 1
          when :d4 then
            @hosts[:distribution_risk][:d4] += 1
          when :d3 then
            @hosts[:distribution_risk][:d3] += 1
          when :d2 then
            @hosts[:distribution_risk][:d2] += 1
          when :d1 then
            @hosts[:distribution_risk][:d1] += 1
          end
        end
      else
        @hosts[:distribution_risk][:fixed] += 1
      end
    end
    @hosts[:dist_cat_hosts] = {}
    [ :public_server, :private_server,
      :mobile, :client, :shared_server, :virtual_host_server, :public_network,
      :private_network ].each_with_index do |v, i|
      @hosts[:dist_cat_hosts][v] =
        @hosts[:all_content].sum {|x| x[:system_category][:id] == i + 1 ? 1 : 0 }
    end
    @hosts[:dist_os_hosts_hash] = {}
    @hosts[:dist_os_hosts_size] = 0
    @hosts[:all_content].each do |host|
      os = '（不明）'
      if host[:os] and host[:os][:osname] and host[:os][:osname].size > 0 then
        os = host[:os][:osname]
      end
      if @hosts[:dist_os_hosts_hash].has_key?(os) then
        @hosts[:dist_os_hosts_hash][os] += 1
      else
        @hosts[:dist_os_hosts_hash][os] = 1
      end
    end
    @hosts[:dist_os_hosts] = []
    @hosts[:dist_os_hosts_hash].keys.each do |k|
      @hosts[:dist_os_hosts] << { :name => k, :value => @hosts[:dist_os_hosts_hash][k] }
    end
    @hosts[:dist_os_hosts_size] = @hosts[:dist_os_hosts].sum {|x| x[:value] }
    temp_ev = proc {|x| [ (x[:name] == '（不明）' ? 1000 : 100), x[:value], x[:name] ]}
    @hosts[:dist_os_hosts].sort! {|x, y|
      xx = temp_ev.call(x)
      yy = temp_ev.call(y)
      r = xx[0] <=> yy[0]
      if r == 0 then
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[2] <=> yy[2]
        end
      end
      r
    }

    check_filters
    @hosts[:content_filtered] = apply_filters
    @hosts[:total_size] = @hosts[:content_filtered].size
    @hosts[:content] = @hosts[:content_filtered].slice(@hosts[:offset], @hosts[:per_page])
    @hosts[:content] = [] unless @hosts[:content]

    @hosts[:statuses] = SidfmdIssue.statuses

    @alert = {
      :warning => @hosts[:not_fixed_corresp_required_host_size] > 0,
    }

    @timeline_data = (@hosts[:grp] == :distribution) ? get_timeline_data : []
  end

  def filter_host_view_content
    @hosts[:distribution_status] = {}
    @hosts[:distribution_risk] = {}
    @hosts[:distribution_status][:notconfirmed] = @hosts[:all_content].where(:min_status_id => SidfmdIssue::STATUS_NOTCONFIRMED).count
    @hosts[:distribution_status][:ongoing] = @hosts[:all_content].where(:min_status_id => SidfmdIssue::STATUS_ONGOING).count
    @hosts[:distribution_status][:fixed] = @hosts[:all_content].where(:min_status_id => SidfmdIssue::STATUS_FIXED).count
    @hosts[:distribution_status][:notaffect] = (
      @hosts[:all_content].count - (
        @hosts[:distribution_status][:notconfirmed] +
        @hosts[:distribution_status][:ongoing] +
        @hosts[:distribution_status][:fixed]
      )
    )

    unfinished_hosts = @hosts[:all_content].where(:min_status_id => SidfmdIssue::STATUS_UNFINISHED)
    if @hosts[:idx] == :sri then
      @hosts[:distribution_risk][:critical] = unfinished_hosts.where(:max_sri => SidfmdIssueSri::VALUES[0]).count
      @hosts[:distribution_risk][:high] = unfinished_hosts.where(:max_sri => SidfmdIssueSri::VALUES[1]).count
      @hosts[:distribution_risk][:medium] = unfinished_hosts.where(:max_sri => SidfmdIssueSri::VALUES[2]).count
      @hosts[:distribution_risk][:low] = unfinished_hosts.where(:max_sri => SidfmdIssueSri::VALUES[3]).count
      @hosts[:distribution_risk][:none] = unfinished_hosts.where(:max_sri => SidfmdIssueSri::VALUES[4]).count
    else
      @hosts[:distribution_risk][:notreg] = unfinished_hosts.where(:max_cvss_label => 'notreg').count
      @hosts[:distribution_risk][:d4] = unfinished_hosts.where(:max_cvss_label => 'd4').count
      @hosts[:distribution_risk][:d3] = unfinished_hosts.where(:max_cvss_label => 'd3').count
      @hosts[:distribution_risk][:d2] = unfinished_hosts.where(:max_cvss_label => 'd2').count
      @hosts[:distribution_risk][:d1] = unfinished_hosts.where(:max_cvss_label => 'd1').count
    end
    @hosts[:distribution_risk][:fixed] = @hosts[:all_content].where.not(:min_status_id => SidfmdIssue::STATUS_UNFINISHED).count

    @hosts[:dist_cat_hosts] = {}
    scat_ids = SidfmdCategory.where(:is_system => true).pluck(:id)
    scat_ids.each do |scat_id|
      label = SidfmdCategory.id2label(scat_id)
      @hosts[:dist_cat_hosts][label] = @hosts[:all_content].where(:system_category_id => scat_id).count
    end

    @hosts[:dist_os_hosts_hash] = @hosts[:all_content].group(:osname).count
    if @hosts[:dist_os_hosts_hash][nil] then
      undef_osname_count = @hosts[:dist_os_hosts_hash][nil]
      @hosts[:dist_os_hosts_hash].delete(nil)
      @hosts[:dist_os_hosts_hash]['（不明）'] = undef_osname_count
    end
    @hosts[:dist_os_hosts_size] = 0
    @hosts[:dist_os_hosts] = []
    @hosts[:dist_os_hosts_hash].keys.each do |k|
      @hosts[:dist_os_hosts] << { :name => k, :value => @hosts[:dist_os_hosts_hash][k] }
    end
    @hosts[:dist_os_hosts_size] = @hosts[:dist_os_hosts].sum {|x| x[:value] }
    temp_ev = proc {|x| [ (x[:name] == '（不明）' ? 1000 : 100), x[:value], x[:name] ]}
    @hosts[:dist_os_hosts].sort! {|x, y|
      xx = temp_ev.call(x)
      yy = temp_ev.call(y)
      r = xx[0] <=> yy[0]
      if r == 0 then
        r = yy[1] <=> xx[1]
        if r == 0 then
          r = xx[2] <=> yy[2]
        end
      end
      r
    }

    check_filters true
    @hosts[:content_filtered] = apply_filters_to_host_view_content
    @hosts[:total_size] = @hosts[:content_filtered].size
    @hosts[:content] = @hosts[:content_filtered].limit(@hosts[:per_page]).offset(@hosts[:offset])
    @hosts[:content] = [] unless @hosts[:content]

    @hosts[:statuses] = SidfmdIssue.statuses

    @alert = {
      :warning => @hosts[:not_fixed_corresp_required_host_size] > 0,
    }

    @timeline_data = (@hosts[:grp] == :distribution) ? get_timeline_data : []
  end

  public

  def index
    parse_params
  end

  def index_r
    parse_params
    create_host_view_content
    filter_host_view_content
    @self_user_categories = SidfmdCategory.where(:id => (SidfmdCategoriesUser.where(:user_id => User.current).map(&:sidfmd_category_id))).map{|c| c.is_system ? Hash["system_c", c.name] : Hash["custom_c", c.name]}
    @filter_inherit = {
      :'pages[idx]' => @hosts[:idx],
      :'pages[filters]' => @hosts[:filters].reject {|k, v|
        [ :keyword ].include?(k)
      }.to_json,
    }
    @display_not_system_categories = SidfmdCategory.where(:id => @not_system_categories.map{|c| c.id}).order('name COLLATE "C"')
    #ブラウザバックかどうか判断するための処理
    #if !session['chkbok_ids'] then
    #  session['chkbok_ids'] = {}
    #end
    #if !session['chkbok_ids']['hosts'] then
    #  session['chkbok_ids']['hosts'] = {}
    #end
    #session['chkbok_ids']['hosts']['is_browser_back'] = false
  end

  def update_status
=begin    f_params = params.fetch(:status, {}).permit(:hostids, :prev_status, :new_status, :comment, :dump)
    dump = {}
    begin
      dump = JSON.parse(f_params[:dump])
    rescue JSON::ParserError => e
      log_exception(e)
      dump[:action] = :index
      dump[:notice] = '入力データが不正です。'
      redirect_to dump and return
    end
    dump[:action] = :index
    hostids = nil
    begin
      hostids = JSON.parse(f_params[:hostids])
    rescue JSON::ParserError => e
      log_exception(e)
      dump[:notice] = '入力データが不正です。'
      redirect_to dump and return
    end
    if !hostids.respond_to?(:'[]') or !hostids.respond_to?(:size) or hostids.size == 0 then
      dump[:notice] = '入力データが不正です。'
      redirect_to dump and return
    end
    if !@role_admin and (@user_host_ids & hostids).size == 0 then
      dump[:notice] = 'アクセスを許可されていません。'
      redirect_to dump and return
    end
    prev_status = f_params[:prev_status].to_i
    if prev_status == 0 and f_params[:prev_status] != '0' then
      dump[:notice] = '設定前ステータスが不正です。'
      redirect_to dump and return
    end
    if !(SidfmdIssue::STATUS_ALL + [0]).include?(prev_status) then
      dump[:notice] = '設定前ステータスが不正です。'
      redirect_to dump and return
    end
    new_status = f_params[:new_status].to_i
    if new_status == 0 and f_params[:new_status] != '0' then
      dump[:notice] = '設定前ステータスが不正です。'
      redirect_to dump and return
    end
    if !(SidfmdIssue::STATUS_ALL + [0]).include?(new_status) then
      dump[:notice] = '設定前ステータスが不正です。'
      redirect_to dump and return
    end
    comment = f_params[:comment]
    comment = 'ステータス変更' if (comment.nil? or comment.size == 0)
    if new_status == 0 then
      dump[:notice] = '設定後ステータスが設定されなかったため、操作は行われておりません。'
      redirect_to dump and return
    end
    begin
      SidfmdIssue.transaction do
        old_status_ids = []
        issues = SidfmdIssue
          .joins(:issue)
          .where(:sidfmd_host_id => hostids)
        if prev_status == 0 then
          issues = issues.where.not(:issues => { :status_id => SidfmdIssue::STATUS_FINISHED })
          old_status_ids = SidfmdIssue::STATUS_UNFINISHED
        else
          issues = issues.where(:issues => { :status_id => prev_status } )
          old_status_ids << prev_status
        end
        #issues.each do |issue|
        #  issue.with_lock do
        #    issue.change_status_with_journal(User.current.id, new_status, comment)
        #  end
        #end
        issue_ids = issues.map{|i| i.id}
        num = (issue_ids.size / 3) > 1 ? (issue_ids.size / 3) : 1
        while (p_issue_ids = issue_ids.shift(num)).size > 0 do
          SidfmdIssue.change_statuses_with_journal(p_issue_ids, old_status_ids, new_status, comment, User.current.id)
        end
      end
      Rails.cache.clear
    rescue ActiveRecord::StatementInvalid => e
      log_exception(e)
      dump[:notice] = 'ステータスの変更に失敗しました。'
      redirect_to dump and return
    end

    redirect_to dump
=end
    f_params = params.permit(:issue_ids,:shift_num, :ids, :prev_status, :new_status, :comment)
    err = false
    err_info = {}
    hostids = nil
    shift_num = f_params[:shift_num]
    begin
      issue_ids = JSON.parse(f_params[:issue_ids])
      hostids = JSON.parse(f_params[:ids])
    rescue JSON::ParserError => e
      err = true
    end
    if !hostids.respond_to?(:'[]') or !hostids.respond_to?(:size) or hostids.size == 0 then
      err = true
    end
    if !@role_admin then
      if (hostids - @user_host_ids).size > 0 then
        err = true
      end
    end
    prev_status = f_params[:prev_status].to_i
    if prev_status == 0 and f_params[:prev_status] != '0' then
      err = true
    end
    if !(SidfmdIssue::STATUS_ALL + [0]).include?(prev_status) then
      err = true
    end
    new_status = f_params[:new_status].to_i
    if new_status == 0 and f_params[:new_status] != '0' then
      err = true
    end
    if !(SidfmdIssue::STATUS_ALL + [0]).include?(new_status) then
      err = true
    end
    comment = f_params[:comment]
    comment = 'ステータス変更' if (comment.nil? or comment.size == 0)
    if new_status == 0 then
      err = true
    end
    if err then
      render :layout => false, :json => { :err => 'invalid' } and return
    end
    return_value = {}
    p_issue_ids = []

    target_html_ids = []
    begin
      SidfmdIssue.transaction do
        old_status_ids = []
        if prev_status == 0 then
          old_status_ids = SidfmdIssue::STATUS_UNFINISHED
        else
          old_status_ids << prev_status
        end
        if issue_ids.size == 0 then
          issues = SidfmdIssue
            .joins(:issue)
            .where(:sidfmd_host_id => hostids)
          if prev_status == 0 then
            issues = issues.where.not(:issues => { :status_id => SidfmdIssue::STATUS_FINISHED })
          else
            issues = issues.where(:issues => { :status_id => prev_status } )
          end
          issue_ids = issues.map{|i| i.issue.id}
          return_value[:total_size] = issue_ids.size
          shift_num = 100
          return_value[:shift_num] = shift_num
        end

        p_issue_ids = issue_ids.shift(shift_num.to_i)
        target_html_ids = SidfmdIssue.where(:issue_id => p_issue_ids).distinct.pluck(:html_id)
        SidfmdIssue.change_statuses_with_journal(p_issue_ids, old_status_ids, new_status, comment, User.current.id)
      end
      Rails.cache.clear
    rescue ActiveRecord::StatementInvalid => e
      logger.info(e)
      err = true
    end
    return_value[:issue_ids] = issue_ids
    return_value[:executed_vuls_size] = p_issue_ids.size
    if err then
      return_value[:err] = 'db_error'
      return_value[:err_vuls_size] = p_issue_ids.size
    else
      #View更新
      SidfmdVulView.sync_data(target_html_ids)
      SidfmdHostView.sync_data(hostids)
    end
    render :layout => false, :json => return_value
  end

  def update_due_dates
=begin    f_params = params.fetch(:schedule, {}).permit(:hostids, :force, :due_date, :comment, :dump)
    dump = {}
    begin
      dump = JSON.parse(f_params[:dump])
    rescue JSON::ParserError => e
      log_exception(e)
      dump[:action] = :index_r
      dump[:notice] = '入力データが不正です。'
      redirect_to dump and return
    end
    dump[:action] = :index_r
    hostids = nil
    begin
      hostids = JSON.parse(f_params[:hostids])
    rescue JSON::ParserError => e
      log_exception(e)
      dump[:notice] = '入力データが不正です。'
      redirect_to dump and return
    end
    if hostids.size == 0 or !hostids.respond_to?(:'[]') then
      dump[:notice] = '入力データが不正です。'
      redirect_to dump and return
    end
    if !@role_admin and (@user_host_ids & hostids).size == 0 then
      dump[:notice] = 'アクセスを許可されていません。'
      redirect_to dump and return
    end
    force = !f_params[:force].nil?
    due_date = f_params[:due_date]
    due_date = nil if (due_date and due_date.size == 0)
    if due_date then
      temp = nil
      begin
        temp = Date.strptime(due_date, '%Y/%m/%d')
      rescue Exception => e
        log_exception(e)
        logger.info('date parse failed: ' + due_date.inspect)
      end
      unless temp then
        begin
          temp = Date.strptime(due_date, '%m/%d/%Y')
        rescue Exception => e
          log_exception(e)
          logger.info('date parse failed: ' + due_date.inspect)
        end
      end
      if temp then
        due_date = temp
      else
        dump[:notice] = '不正な日付です。'
        redirect_to dump and return
      end
    end
    comment = f_params[:comment]
    comment = '対策予定日設定変更' if (comment.nil? or comment.size == 0)
    begin
      SidfmdIssue.transaction do
        issues = SidfmdIssue
          .joins(:issue)
          .where(:sidfmd_host_id => hostids)
          .where.not(:issues => { :status_id => SidfmdIssue::STATUS_FINISHED })
        if !force then
          issues = issues
            .where(:issues => { :due_date => nil })
        end
        issue_ids = issues.map{|i| i.issue.id}
        SidfmdIssue.change_due_dates_with_journal(issue_ids, force, due_date, comment, User.current.id)
      end
      Rails.cache.clear
    rescue ActiveRecord::StatementInvalid => e
      log_exception(e)
      dump[:notice] = '不正な日付です。'
      redirect_to dump and return
    end

    redirect_to dump
=end
    f_params = params.permit(:issue_ids, :shift_num, :ids, :force, :due_date, :comment)
    err = false
    err_info = {}
    hostids = nil
    shift_num = f_params[:shift_num]
    begin
      issue_ids = JSON.parse(f_params[:issue_ids])
      hostids = JSON.parse(f_params[:ids])
    rescue JSON::ParserError => e
      logger.info(e)
      err = true
    end
    if hostids.size == 0 or !hostids.respond_to?(:'[]') then
      err = true
    end
    if !@role_admin then
      if (hostids - @user_host_ids).size > 0 then
        err = true
      end
    end
    force = f_params[:force] == 'true' ? true : false
    due_date = f_params[:due_date]
    due_date = nil if (due_date and due_date.size == 0)
    if due_date then
      temp = nil
      begin
        temp = Date.strptime(due_date, '%Y/%m/%d')
      rescue Exception => e
        logger.info(e)
        err = true
      end
      unless temp then
        begin
          temp = Date.strptime(due_date, '%m/%d/%Y')
        rescue Exception => e
          logger.info(e)
          err = true
        end
      end
      if temp then
        due_date = temp
      else
        err = true
      end
    end
    comment = f_params[:comment]
    comment = '対策予定日設定変更' if (comment.nil? or comment.size == 0)
    if err then
      render :layout => false, :json => { :err => 'invalid' } and return
    end
    return_value = {}
    p_issue_ids = []

    target_html_ids = []
    begin
      SidfmdIssue.transaction do
        if issue_ids.size == 0 then
          issues = SidfmdIssue
            .joins(:issue)
            .where(:sidfmd_host_id => hostids)
            .where.not(:issues => { :status_id => SidfmdIssue::STATUS_FINISHED })
          if !force then
            issues = issues.where(:issues => { :due_date => nil })
          end
          issue_ids = issues.map{|i| i.issue.id}
          return_value[:total_size] = issue_ids.size
          shift_num = 100
          return_value[:shift_num] = shift_num
        end

        p_issue_ids = issue_ids.shift(shift_num.to_i)
        target_html_ids = SidfmdIssue.where(:issue_id => p_issue_ids).distinct.pluck(:html_id)
        SidfmdIssue.change_due_dates_with_journal(p_issue_ids, force, due_date, comment, User.current.id)
      end
      Rails.cache.clear
    rescue ActiveRecord::StatementInvalid => e
      logger.info(e)
      err = true
    end
    return_value[:issue_ids] = issue_ids
    return_value[:executed_vuls_size] = p_issue_ids.size
    if err then
      return_value[:err] = 'db_error'
      return_value[:err_vuls_size] = p_issue_ids.size
    else
      #View更新
      SidfmdVulView.sync_data(target_html_ids)
      SidfmdHostView.sync_data(hostids)
    end
    render :layout => false, :json => return_value
  end

  CSV_BOM = "\xff\xfe".force_encoding('UTF-16LE').freeze

=begin
  #HostViewからデータを取得するdownload
  def download
    f_params = params.fetch(:download, {}).permit(:format, :category_id, :idx, :order, :filters, :hostids)
    format = f_params[:format]
    if ![ 'json', 'tsv', 'tsv-package-info' ].include?(format) then
      redirect_to :action => :index ,:notice => '不正な入力です。' and return
    end 

    @hosts = { :url => request.url }
    @hosts[:order] = (f_params[:order] || 'id_asc')
    @hosts[:filters] = JSON.parse(f_params[:filters] || '{}').symbolize_keys
    @hosts[:category_id] = (f_params[:category_id] || 0).to_i
    if @hosts[:category_id] > 0 then
      @hosts[:category_name] = SidfmdCategory.find(@hosts[:category_id]).name
    else
      @hosts[:category_name] = '全てのホスト'
    end
    @hosts[:grp] = :count
    @hosts[:idx] = (f_params[:idx] || 'sri').to_sym
    @hosts[:offset] = 0
    @hosts[:per_page] = 10
    @hosts[:hostids] = []
    begin
      @hosts[:hostids] = JSON.parse(f_params[:hostids] || '[]')
    rescue JSON::ParserError => e
      logger.error(e)
    end
    @system_categories = SidfmdCategory.system.with_hosts
    @not_system_categories = SidfmdCategory.not_system.with_hosts
    if !@role_admin then
      @not_system_categories = @not_system_categories.where(:id => @user_category_ids)
    end
    @corresp_required = Sidfm.corresp_required
    @risks = SidfmdRisk.all.order(:id => :desc)
    @statuses = IssueStatus.all.order(:id)
    create_host_view_content
    filter_host_view_content

    case format
    when 'json' then
      json = {}
      temp_order = @hosts[:order].split(/_/)
      json[:order] = { :by => temp_order[0], :type => temp_order[1], }
      json[:filters] = @hosts[:filters]
      json[:category] = { :id => @hosts[:category_id], :name => @hosts[:category_name], }
      json[:idx] = @hosts[:idx]
      json[:content] = []
      @hosts[:content_filtered].each do |host|
        if @hosts[:hostids].size != 0 and !@hosts[:hostids].include?(host[:sidfmd_host_id]) then
          next
        end
        item = {}
        item[:host] = {
          :id => host[:sidfmd_host_id],
          :hostname => host[:hostname] ,
          :ext_name => host[:ext_name],
          :display_name => host[:display_name]
        }
        item[:warning] = if @hosts[:idx] == :sri then
                           host[:warning_sri]
                         else
                           host[:warning_cvss]
                         end
        status_ids = host.sidfmd_hv_scsids.pluck(:status_id)
        count_status_ids = SidfmdIssue.count_status_ids(status_ids)
        item[:statuses] = count_status_ids
          .sort {|x, y| x[:id] <=> y[:id] }
          .collect {|x| { :name => x[:name], :value => x[:value] } }
        item[:max_sri] = SidfmdIssueSri::RISKS[host[:max_sri]]
        item[:max_cvss] = (host[:max_cvss_label] == 'notreg' ? '' : sprintf('%.1f', (host[:max_cvss_score] ? host[:max_cvss_score] : 0).to_f))
        item[:max_area] = SidfmdArea.where(:area_id => host[:max_area_id]).first.name
        item[:exploitable] = (host[:max_cvss_label] != 'notreg' ? host[:exploitable] : nil)
        ns_cats = if @role_admin then
                    host.sidfmd_hv_categories.order('LOWER(name COLLATE "C")')
                  else
                    host.sidfmd_hv_categories.where(:sidfmd_category_id => @user_category_ids)
                                             .order('LOWER(name COLLATE "C")')
                  end
        cats = []
        cats << host[:system_category_name]
        ns_cats.each do |cat|
          cats << cat[:name]
        end
        cats.uniq!
        cats.compact!
        item[:categories] = cats
        item[:os] = ((host[:osname] and host[:osname].size > 0) ? host[:osname] : '')
      
        item[:due_dates_percent] = (host.duedates_percent ? display_percent(host.duedates_percent) : '')
        item[:finished_percent] = (status_ids.any? ? display_finished_percent(status_ids) : '')
        apps = []
        pkgs = []
        happs = []
        sidfmd_host = SidfmdHost.find(host[:sidfmd_host_id])
        if sidfmd_host then
          queue = sidfmd_host.get_last_queue
          if queue then
            pkgs = queue.sidfmd_queue_pkgs.order(:pkg).collect {|x| x.pkg }
            apps = queue.sidfmd_queue_apps.order(:name,:version).collect {|x| {
                :name => x.name, :version => x.version, :path => x.path,
              }}
          end
          happs = sidfmd_host.sidfmd_host_apps.order(:product_id,:version_id).collect {|x| x.get_name }
        end
        item[:packages] = pkgs
        item[:products] = apps
        item[:manual_products] = happs
        json[:content] << item
      end
      send_data(json.to_json, :filename => "vmlist-hosts-#{Time.current.strftime('%Y%m%d%H%M%S')}.json", :disposition => 'attachment')
      return
    when 'tsv' then
      # TSV, UTF-16LE, BOMあり, 拡張子 .csv
      header = [ 'ホストID', '要対応', '対応状況', '最大SRI', '最大CVSS', '影響範囲', '攻撃コード', 'ホスト名', '表示名', 'カテゴリ', 'OS名', '対策予定日設定済', '対処率' ]

      header = header.collect {|x| x.encode('UTF-16LE') }
      csv = CSV.generate("#{CSV_BOM}".force_encoding('UTF-16LE'), :col_sep => "\t".encode('UTF-16LE')) do |o|
        o << header
        item = []
        @hosts[:content_filtered].each do |host|
          if @hosts[:hostids].size != 0 and !@hosts[:hostids].include?(host[:sidfmd_host_id]) then
            next
          end
          item.clear
          item << host[:sidfmd_host_id].to_s.encode('UTF-16LE')
          if @hosts[:idx] == :sri then
            item << host[:warning_sri].to_s.encode('UTF-16LE')
          else
            item << host[:warning_cvss].to_s.encode('UTF-16LE')
          end
          status_ids = host.sidfmd_hv_scsids.pluck(:status_id)
          count_status_ids = SidfmdIssue.count_status_ids(status_ids)
          item << count_status_ids.sort {|x, y| x[:id] <=> y[:id] }.collect {|x| "#{x[:name]}:#{x[:value]}" }.join(',').encode('UTF-16LE')
          item << SidfmdIssueSri::RISKS[host[:max_sri]].encode('UTF-16LE')
          item << (host[:max_cvss_label] == 'notreg' ? '' : sprintf('%.1f', (host[:max_cvss_score] ? host[:max_cvss_score] : 0).to_f)).encode('UTF-16LE')
          item << SidfmdArea.where(:area_id => host[:max_area_id]).first.name.encode('UTF-16LE')
          item << host[:exploitable].to_s.encode('UTF-16LE')
          item << host[:hostname].encode('UTF-16LE')
          item << host[:display_name].encode('UTF-16LE')
          ns_cats = if @role_admin then
                    host.sidfmd_hv_categories.order('LOWER(name COLLATE "C")')
                  else
                    host.sidfmd_hv_categories.where(:sidfmd_category_id => @user_category_ids)
                                             .order('LOWER(name COLLATE "C")')
                  end
          cats = []
          cats << host[:system_category_name].encode('UTF-16LE')
          ns_cats.each do |cat|
            cats << cat[:name].encode('UTF-16LE')
          end
          cats.uniq!
          cats.compact!
          item << cats.join(','.encode('UTF-16LE')).encode('UTF-16LE')
          item << ((host[:osname] and host[:osname].size > 0) ? host[:osname] : '').encode('UTF-16LE')
          item << (host.duedates_percent ? display_percent(host.duedates_percent) : '').encode('UTF-16LE')
          item << (status_ids.any? ? display_finished_percent(status_ids) : '').encode('UTF-16LE')
          o << item
        end
      end
      send_data(csv, :filename => "vmlist-hosts-tsv-#{Time.current.strftime('%Y%m%d%H%M%S')}.csv", :disposition => 'attachment')
      return
    when 'tsv-package-info' then
      header = [ 'ホストID', 'ホスト名', 'パッケージ名' ]
      header = header.collect {|x| x.encode('UTF-16LE') }
      csv = CSV.generate("#{CSV_BOM}".force_encoding('UTF-16LE'), :col_sep => "\t".encode('UTF-16LE')) do |o|
        o << header
        item = []
        @hosts[:content_filtered].each do |host|
          if @hosts[:hostids].size != 0 and !@hosts[:hostids].include?(host[:sidfmd_host_id]) then
            next
          end
          last_queue = SidfmdHost.find_by_hostname(host[:hostname]).get_last_queue
          if last_queue then
            pkgs = last_queue.sidfmd_queue_pkgs
            if pkgs.size > 0 then
              pkgs.each do |x|
                item.clear
                item << host[:sidfmd_host_id].to_s.encode('UTF-16LE')
                item << host[:hostname].encode('UTF-16LE')
                item << x.pkg.encode('UTF-16LE')
                o << item
              end
            end
          end
        end
      end
      send_data(csv, :filename => "vmlist-hosts-packages-tsv-#{Time.current.strftime('%Y%m%d%H%M%S')}.csv", :disposition => 'attachment')
      return
    end

    redirect_to :action => :index, :notice => 'ファイルの出力に失敗しました。'
  end
=end

  def download
    f_params = params.fetch(:download, {}).permit(:format, :category_id, :idx, :order, :filters, :hostids)
    format = f_params[:format]
    if ![ 'json', 'tsv', 'tsv-package-info', 'tsv-product-info' ].include?(format) then
      redirect_to :action => :index ,:notice => '不正な入力です。' and return
    end
    ### 以降 index と同様の処理。
    @hosts = { :url => request.url }
    @hosts[:order] = (f_params[:order] || 'id_asc')
    @hosts[:filters] = JSON.parse((f_params[:filters] || '{}'), symbolize_names: true)
    @hosts[:category_id] = (f_params[:category_id] || 0).to_i
    if @hosts[:category_id] > 0 then
      @hosts[:category_name] = SidfmdCategory.find(@hosts[:category_id]).name
    else
      @hosts[:category_name] = '全てのホスト'
    end
    @hosts[:grp] = :count
    @hosts[:idx] = (f_params[:idx] || 'sri').to_sym
    @hosts[:offset] = 0
    @hosts[:per_page] = 10
    @hosts[:hostids] = []
    begin
      @hosts[:hostids] = JSON.parse(f_params[:hostids] || '[]')
    rescue JSON::ParserError => e
      logger.error(e)
    end
    @system_categories = SidfmdCategory.system.with_hosts
    @not_system_categories = SidfmdCategory.not_system.with_hosts
    if !@role_admin then
      @not_system_categories = @not_system_categories.where(:id => @user_category_ids)
    end
    @corresp_required = Sidfm.corresp_required
    @risks = SidfmdRisk.all.order(:id => :desc)
    @statuses = IssueStatus.all.order(:id)
    create_all_content(@hosts[:category_id])
    @hosts[:corresp_required_host_size] =
      @hosts[:all_content].sum {|x| x[:warning][@hosts[:idx]] ? 1 : 0 }
    @hosts[:fixed_corresp_required_host_size] =
      @hosts[:all_content].sum {|x| x[:fixed_warning][@hosts[:idx]] ? 1 : 0 }
    @hosts[:not_fixed_corresp_required_host_size] =
      @hosts[:corresp_required_host_size] - @hosts[:fixed_corresp_required_host_size]
    @hosts[:all_host_size] = @hosts[:all_content].size
    filter_content
    ###
    case format
    when 'json' then
      json = {}
      temp_order = @hosts[:order].split(/_/)
      json[:order] = { :by => temp_order[0], :type => temp_order[1], }
      json[:filters] = @hosts[:filters]
      json[:category] = { :id => @hosts[:category_id], :name => @hosts[:category_name], }
      json[:idx] = @hosts[:idx]
      json[:content] = []
      @hosts[:content_filtered].each do |host|
        if @hosts[:hostids].size != 0 and !@hosts[:hostids].include?(host[:sidfmd_host_id]) then
          next
        end
        item = {}
        item[:host] = { 
          :id => host[:sidfmd_host_id], 
          :hostname => host[:hostname] , 
          :ext_name => host[:ext_name], 
          :display_name => host[:display_name]
        }
        item[:warning] = host[:warning][@hosts[:idx]]
        count_status_ids = SidfmdIssue.count_status_ids(host[:status_ids])
        item[:statuses] = count_status_ids
          .sort {|x, y| x[:id] <=> y[:id] }
          .collect {|x| { :name => x[:name], :value => x[:value] } }
        item[:max_sri] = SidfmdIssueSri::RISKS[host[:max_sri]]
        t_cvss = host[:max_cvss]
        item[:max_cvss] = (t_cvss[:label] == :notreg ? '' : sprintf('%.1f', (t_cvss[:score] ? t_cvss[:score] : 0).to_f))
        item[:max_area] = SidfmdArea.where(:area_id => host[:max_area_id]).first.name
        item[:exploitable] = (t_cvss[:label] != :notreg ? host[:exploitable] : nil)
        sc = host[:system_category]
        ns_cats = host[:no_system_categories]
        cats = []
        cats << sc[:name]
        ns_cats.each do |cat|
          cats << cat[:name]
        end
        cats.uniq!
        cats.compact!
        item[:categories] = cats
        item[:os] = ((host[:os] and host[:os][:osname] and host[:os][:osname].size > 0) ? host[:os][:osname] : '')
        item[:due_dates_percent] = (host[:due_dates_unfinished].size > 0 ? display_due_dates_percent(host[:due_dates_unfinished]) : '0.0%')
        item[:finished_percent] = (host[:status_ids].size > 0 ? display_finished_percent(host[:status_ids]) : '100.0%')
        apps = []
        pkgs = []
        happs = []
        hlapps = []
        sidfmd_host = SidfmdHost.find(host[:sidfmd_host_id])
        if sidfmd_host then
          queue = sidfmd_host.get_last_queue
          if queue then
            pkgs = queue.sidfmd_queue_pkgs.order(:pkg).collect {|x| x.pkg }
            apps = queue.sidfmd_queue_apps.order(:name,:version).collect {|x|
              { :name => x.name, :version => x.version, :path => x.path }
            }
          end
          happs = sidfmd_host.sidfmd_host_apps.order(:product_id,:version_id).collect {|x| x.get_name }.compact
          hlapps = sidfmd_host.sidfmd_host_lapps.order(:sidfmd_lproduct_id, :sidfmd_lversion_id).collect {|x| x.get_name }.compact
        end
        item[:packages] = pkgs
        item[:products] = apps
        item[:manual_products] = happs
        item[:local_products] = hlapps
        json[:content] << item
      end
      send_data(json.to_json, :filename => "vmlist-hosts-#{Time.current.strftime('%Y%m%d%H%M%S')}.json", :disposition => 'attachment')
      return
    when 'tsv' then
      # TSV, UTF-16LE, BOMあり, 拡張子 .csv
      header = [ 'ホストID', '要対応', '対応状況', '最大SRI', '最大CVSS', '影響範囲', '攻撃コード', 'ホスト名', '表示名', 'カテゴリ', 'OS名', '対策予定日設定済', '対処率' ]

      header = header.collect {|x| x.encode('UTF-16LE') }
      csv = CSV.generate("#{CSV_BOM}".force_encoding('UTF-16LE'), :col_sep => "\t".encode('UTF-16LE')) do |o|
        o << header
        item = []
        @hosts[:content_filtered].each do |host|
          if @hosts[:hostids].size != 0 and !@hosts[:hostids].include?(host[:sidfmd_host_id]) then
            next
          end
          item.clear
          item << host[:sidfmd_host_id].to_s.encode('UTF-16LE')
          item << host[:warning][@hosts[:idx]].to_s.encode('UTF-16LE')
          count_status_ids = SidfmdIssue.count_status_ids(host[:status_ids])
          item << count_status_ids.sort {|x, y| x[:id] <=> y[:id] }.collect {|x| "#{x[:name]}:#{x[:value]}" }.join(',').encode('UTF-16LE')
          item << SidfmdIssueSri::RISKS[host[:max_sri]].encode('UTF-16LE')
          t_cvss = host[:max_cvss]
          item << (t_cvss[:label] == :notreg ? '' : sprintf('%.1f', (t_cvss[:score] ? t_cvss[:score] : 0).to_f)).encode('UTF-16LE')
          item << SidfmdArea.where(:area_id => host[:max_area_id]).first.name.encode('UTF-16LE')
          item << host[:exploitable].to_s.encode('UTF-16LE')
          item << host[:hostname].encode('UTF-16LE')
          item << host[:display_name].encode('UTF-16LE')
          sc = host[:system_category]
          ns_cats = host[:no_system_categories]
          cats = []
          cats << escape_category_tsv(sc[:name]).encode('UTF-16LE')
          ns_cats.each do |cat|
            cats << escape_category_tsv(cat[:name]).encode('UTF-16LE')
          end
          cats.uniq!
          cats.compact!
          item << cats.join(','.encode('UTF-16LE')).encode('UTF-16LE')
          item << ((host[:os] and host[:os][:osname] and host[:os][:osname].size > 0) ? host[:os][:osname] : '').encode('UTF-16LE')
          item << (host[:due_dates_unfinished].size > 0 ? display_due_dates_percent(host[:due_dates_unfinished]) : '0.0%').encode('UTF-16LE')
          item << (host[:status_ids].size > 0 ? display_finished_percent(host[:status_ids]) : '100.0%').encode('UTF-16LE')
          o << item
        end
      end
      send_data(csv, :filename => "vmlist-hosts-tsv-#{Time.current.strftime('%Y%m%d%H%M%S')}.csv", :disposition => 'attachment')
      return
    when 'tsv-package-info' then
      header = [ 'ホストID', 'ホスト名', 'パッケージ名' ]
      header = header.collect {|x| x.encode('UTF-16LE') }
      csv = CSV.generate("#{CSV_BOM}".force_encoding('UTF-16LE'), :col_sep => "\t".encode('UTF-16LE')) do |o|
        o << header
        item = []
        @hosts[:content_filtered].each do |host|
          if @hosts[:hostids].size != 0 and !@hosts[:hostids].include?(host[:sidfmd_host_id]) then
            next
          end
          last_queue = SidfmdHost.find_by_hostname(host[:hostname]).get_last_queue
          if last_queue then
            pkgs = last_queue.sidfmd_queue_pkgs
            if pkgs.size > 0 then
              pkgs.each do |x|
                item.clear
                item << host[:sidfmd_host_id].to_s.encode('UTF-16LE')
                item << host[:hostname].encode('UTF-16LE')
                item << x.pkg.encode('UTF-16LE')
                o << item
              end
            end
          end
        end
      end
      send_data(csv, :filename => "vmlist-hosts-packages-tsv-#{Time.current.strftime('%Y%m%d%H%M%S')}.csv", :disposition => 'attachment')
      return
    when 'tsv-product-info' then
      header = [ 'ホストID', 'ホスト名', '種類', 'プロダクト名' ]
      header = header.collect {|x| x.encode('UTF-16LE') }
      csv = CSV.generate("#{CSV_BOM}".force_encoding('UTF-16LE'), :col_sep => "\t".encode('UTF-16LE')) do |o|
        o << header
        @hosts[:content_filtered].each do |host|
          if @hosts[:hostids].size != 0 and !@hosts[:hostids].include?(host[:sidfmd_host_id]) then
            next
          end
          products = []
          lproducts = []
          SidfmdHostApp.where(:sidfmd_host_id => host[:sidfmd_host_id]).each do |app|
            item = []
            if app.version_id and app.version_id > 0 then
              version = SidfmdVersion.find_by(:vid => app.version_id)
              if version then
                item << host[:sidfmd_host_id].to_s.encode('UTF-16LE')
                item << host[:hostname].encode('UTF-16LE') 
                item << 'manual_product'.encode('UTF-16LE')
                item << version.fullname.encode('UTF-16LE')
                products << item
              end
            else
              product = SidfmdProduct.find_by(:pid => app.product_id)
              if product then
                item << host[:sidfmd_host_id].to_s.encode('UTF-16LE')
                item << host[:hostname].encode('UTF-16LE') 
                item << 'manual_product'.encode('UTF-16LE')
                item << product.name.encode('UTF-16LE')
                products << item
              end
            end
          end
          SidfmdHostLapp.where(:sidfmd_host_id => host[:sidfmd_host_id]).each do |lapp|
            item = []
            if lapp.sidfmd_lversion_id > 0 then
              lversion = SidfmdLversion.find_by(:id => lapp.sidfmd_lversion_id)
              if lversion then
                item << host[:sidfmd_host_id].to_s.encode('UTF-16LE')
                item << host[:hostname].encode('UTF-16LE')
                item << 'local_product'.encode('UTF-16LE')
                item << lversion.fullname.encode('UTF-16LE')
                lproducts << item
              end
            else
              lproduct = SidfmdLproduct.find_by(:id => lapp.sidfmd_lproduct_id)
              if lproduct then
                item << host[:sidfmd_host_id].to_s.encode('UTF-16LE')
                item << host[:hostname].encode('UTF-16LE')
                item << 'local_product'.encode('UTF-16LE')
                item << lproduct.name.encode('UTF-16LE')
                lproducts << item
              end
            end
          end
          products.sort_by!{|prd| prd[3]}
          lproducts.sort_by!{|lprd| lprd[3]}
          products.each do |prd|
            o << prd
          end
          lproducts.each do |lprd|
            o << lprd
          end
        end
      end
      send_data(csv, :filename => "vmlist-hosts-products-tsv-#{Time.current.strftime('%Y%m%d%H%M%S')}.csv", :disposition => 'attachment')
      return
    end
  
    redirect_to :action => :index, :notice => 'ファイルの出力に失敗しました。'
  end

  def send_mail
    filtered_params = params.fetch(:form, {}).permit(:recipients, :subject, :message)
    begin
      mail = {}
      mail[:from] = User.current.mail
      email_with_name = []
      recipients = JSON.parse(filtered_params[:recipients])
      recipients.each do |address, name|
        email_name = %("#{name}" <#{address}>)
        email_with_name << email_name
      end
      #mail[:to] = email_with_name
      mail[:subject] = filtered_params[:subject]
      mail[:message] = filtered_params[:message]
      send_mail_num = SidfmdMailer::SEND_MAIL_NUM
      time_out_second = SidfmdMailer::SEND_MAIL_TIMEOUT_SECOND
      while (send_mail_targets = email_with_name.shift(send_mail_num)).size > 0 do
        threads = send_mail_targets.map{|fullname_and_address|
          Thread.new(mail.deep_dup, fullname_and_address){|cp_mail, data|
            ret = {}
            cp_mail[:to] = data
            begin
              SidfmdMailer.send_mail(cp_mail).deliver_now
            rescue Exception => e
              ret[:exception] = e
            end
            ret
         }
        }
        threads.each do |result|
          result = result.join(time_out_second)
          if result then
            result = result.value
            if result[:exception] then logger.error("failed send mail. #{result[:exception]}") end
          else
            logger.error("time-out of sending a mail")
          end
        end
      end
      #SidfmdMailer.send_mail(mail).deliver_now
    rescue Exception => e
      log_exception(e)
      redirect_to :action => :index, :notice => e and return
    end
    redirect_to :action => :index and return
  end

end
