# -*- coding: utf-8 -*-

require 'date'
require 'csv'
require 'tempfile'
require 'zip'
require 'active_record'

require 'sidfmdlm_api'
require 'sidfm_controller_support'

class SidfmLicensesController < ApplicationController
  unloadable

  layout 'sidfm'

  self.main_menu = false

  protect_from_forgery with: :exception

  helper :sidfm

  # before_action :api_init, :except => [ :license_setting, :license_update ]

  private

  include SidfmControllerSupport
  before_action :update_session_updated_at
  before_action :get_sidfm_roles
  before_action :create_common_data

  before_action :check_license, :except => [ :license, :license_setting, :license_update ]
  before_action :check_roles

  before_action :get_vm_version
  before_action :clear_import_session, :except => [ :index, :index_r, :import, :new_import, :progress_import_opts, :progress_import_hosts ]  
  before_action :clear_export_session, :except => [ :new_export ]

  # 権限チェック
  def check_roles
    unless @sidfm_user then
      redirect_to home_path, :notice => af_log_user('許可されていない操作です。') and return
    end
    @role_admin = @sidfm_roles.include?(:admin)
    unless @sidfm_roles.include?(:view) then
      redirect_to user_path(User.current), :notice => af_log_user('アクセス権限がありません。') and return
    end
    # SIDfmVM管理者、SIDfmVM報告者、SIDfmVM利用者権限でアクセス可能な範囲
    if [ 'set_display_name', 'update_memo', 'update_memo_bulk', 'update_applications', 'update_applications_bulk', 'set_last_checked_at' ].include?(action_name) and (!@sidfm_roles.include?(:host_admin) and !@sidfm_roles.include?(:host_limited_admin)) then
      redirect_to home_path, :notice => af_log_user('許可されていない操作です。') and return
    end
    # SIDfmVM管理者、SIDfmVM報告者でアクセス可能な範囲
    if [ 'register_host', 'toggle_host_delete', 'rename_host',
         'update_refdate', 'update_refdate_bulk', 
         'update_categories', 'update_categories_bulk',
         'load_offline_data', 'import', 'new_import',
         'new_export', 'progress_import_opts', 'progress_import_hosts',
         'new_import_apply', 
       ].include?(action_name) and !@sidfm_roles.include?(:host_admin)
    then
      redirect_to home_path, :notice => af_log_user('許可されていない操作です。') and return
    end
    if [ 'license', 'license_setting', 'license_update', 'license_apikey_update', 'host_apikey_update', 'register_apikey_update', 'register_apikey_remove', 'import_apply', 'export' ].include?(action_name) and !@role_admin then
      redirect_to home_path, :notice => af_log_user('許可されていない操作です。') and return
    end
    @user_category_ids = SidfmdCategoriesUser.where(:user_id => User.current.id).select(:sidfmd_category_id).order(:sidfmd_category_id).distinct.collect {|x| x[:sidfmd_category_id] }
    @user_host_ids = SidfmdCategoriesHost.joins(:sidfmd_host).where(:sidfmd_category_id => @user_category_ids).where(:sidfmd_hosts => { :flg_delete => false }).select(:sidfmd_host_id).order(:sidfmd_host_id).distinct.collect {|x| x[:sidfmd_host_id] }
    @cache_prefix = @role_admin ? '/license/admin/0' : "/license/user/#{@user_category_ids.join(",")}"
  end
  
  #ホスト名チェック
  def check_hostname(hostname)
    if /\A[0-9a-zA-Z._-]+\z/ !~ hostname then
      return 'ホスト名に使用できない文字が含まれています。'
    end
    if hostname.bytesize > SidfmdHost::MAX_HOSTNAME_BYTE_LENGTH then
      return SidfmdHost::MAX_HOSTNAME_BYTE_LENGTH.to_s + 'バイトを超えるホスト名は登録できません。'
    end
    return true
  end

  #ホスト付属情報の、SidfmdQueueに紐づくデータの登録
  def regist_host_queue_opts(host_id, package_names, app_names, dist_name)
    begin 
      queue = SidfmdQueue.create(sidfmd_host_id: host_id)
      if package_names then
        regist_pkgs = package_names.map{|pkg|
          SidfmdQueuePkg.new(sidfmd_queue_id: queue.id, pkg: pkg)
        }
        SidfmdQueuePkg.import(regist_pkgs) if regist_pkgs.size > 0
      end
      if app_names then
        regist_apps = []
        app_names.each do |app_name|
          app = app_name.split(':')
          if app.size == 3 then #"#{SidfmdQueueApp#name}:#{SidfmdQueueApp#version}:#{SidfmdQueueApp#path}"
            regist_apps << SidfmdQueueApp.new(
              sidfmd_queue_id: queue.id, 
              name: app[0],
              version: app[1],
              path: app[2]
            )
          end  
        end
        SidfmdQueueApp.import(regist_apps) if regist_apps.size > 0
      end
      if dist_name then
        dist = dist_name.split(':')
        if dist.size == 3 then #"#{SidfmdQueueDist#ident}:#{SidfmdQueueDist#version}:#{SidfmdQueueDist:name}"
          SidfmdQueueDist.create(
            sidfmd_queue_id: queue.id,
            ident: dist[0],
            version: dist[1],
            name: dist[2]
          )
        end
      end
    rescue Exception => e
      log_exception(e)
      return false
    end
    return true
  end
 
  #プロダクトを登録する。インポートで使用する
  def regist_products(manual_product_names, local_product_names, host_id)
    host_products = []
    if manual_product_names.any? then
      unregistrable_product_ids = SidfmdProduct.unregistrable_product_ids
      SidfmdProduct.where(:name => manual_product_names).each{|p| 
        next if unregistrable_product_ids.include?(p.pid)
        host_products << SidfmdHostApp.new(sidfmd_host_id: host_id, product_id: p.pid, version_id: 0)
      }
    end
    host_versions = SidfmdVersion.where(:fullname => manual_product_names).map{|v|
      r = nil
      prd = SidfmdProduct.find_by(:id => v.sidfmd_product_id)
      if prd then
        r = SidfmdHostApp.new(sidfmd_host_id: host_id, product_id: prd.pid, version_id: v.vid)
      end
      r
    }.compact
    regist_products = host_products + host_versions
    SidfmdHostApp.import(regist_products) if regist_products.size > 0
    
    local_products = SidfmdLproduct.where(:name => local_product_names).map{|lp|
      SidfmdHostLapp.new(sidfmd_host_id: host_id, sidfmd_lproduct_id: lp.id, sidfmd_lversion_id: 0)
    }
    local_versions = SidfmdLversion.where(:fullname => local_product_names).map{|lv|
      SidfmdHostLapp.new(sidfmd_host_id: host_id, sidfmd_lproduct_id: lv.sidfmd_lproduct_id, sidfmd_lversion_id: lv.id)
    }
    regist_lproducts = local_products + local_versions
    SidfmdHostLapp.import(regist_lproducts) if regist_lproducts.size > 0
  end

  # get_hosts_by_lmのitemをCSVの1行分のitemに変換する。
  def item2csv(host)
    item = []
    item << (host[:sidfmd_host] ? host[:sidfmd_host].id : '').to_s.encode('UTF-16LE')
    item << host[:apikey].encode('UTF-16LE')
    item << host[:hostname].encode('UTF-16LE')
    item << (host[:sidfmd_host] ? host[:sidfmd_host].sidfmd_categories.distinct : []).collect {|x| x.name.encode('UTF-16LE') }.join(','.encode('UTF-16LE')).encode('UTF-16LE')
    item << (host[:sidfmd_host] ? (host[:sidfmd_host].refdate ? host[:sidfmd_host].refdate.strftime('%Y/%m/%d') : '') : '').encode('UTF-16LE')
    item << host[:flg_delete].to_s.encode('UTF-16LE')
    item << (host[:sidfmd_host] ? host[:sidfmd_host].is_agent.to_s : '').encode('UTF-16LE')
    item << (host[:sidfmd_host] ? (host[:sidfmd_host].memo ? host[:sidfmd_host].memo : '') : '').encode('UTF-16LE')
    if host[:sidfmd_host] then
      host[:sidfmd_host].sidfmd_host_apps.each do |app|
        if app.version_id then
          version = SidfmdVersion.where(:vid => app.version_id).first
          if version then
            item << version.fullname.encode('UTF-16LE')
          elsif app.product_id then
            product = SidfmdProduct.where(:pid => app.product_id).first
            if product then
              item << product.name.encode('UTF-16LE')
            end
          end
        elsif app.product_id then
          product = SidfmdProduct.where(:pid => app.product_id).first
          if product then
            item << product.name.encode('UTF-16LE')
          end
        end
      end
      host[:sidfmd_host].sidfmd_host_lapps.each do |lapp|
        if lapp.sidfmd_lversion_id == 0 then
          lproduct = SidfmdLproduct.find_by(:id => lapp.sidfmd_lproduct_id)
          item << lproduct.name.encode('UTF-16LE') if lproduct
        else 
          lversion = SidfmdLversion.find_by(:id => lapp.sidfmd_lversion_id)
          item << lversion.fullname.encode('UTF-16LE') if lversion
        end
      end
    end
    item
  end

  #get_hosts_by_lmのitemをnew_export用のCSVの1行分のitemに変換する。
  def create_new_export_hosts_info(host, sidfmd_host)
    #列(項目)は下記の通り
    #ホスト削除、ホストID、トークン、ホスト名、表示名、カテゴリ、基準日、削除、Agentフラグ、メモ
    item = []
    if !@role_admin then
      if !sidfmd_host then
         return item
      end
      host_cats = sidfmd_host.sidfmd_categories.distinct.select{|cat| (cat.is_system || @user_category_ids.include?(cat.id))}
    else
      if sidfmd_host then
        host_cats = sidfmd_host.sidfmd_categories.distinct
      end
    end
    item << ''
    item << ''
    item << (sidfmd_host ? sidfmd_host.id : '').to_s
    item << host['apikey']
    item << host['hostname']
    item << (sidfmd_host ? (sidfmd_host.ext_name ? sidfmd_host.ext_name : '') : '')
    item << (sidfmd_host ? host_cats : []).sort{|a, b|
              r = (a.is_system ? 1 : 10) <=> (b.is_system ? 1 : 10)
              if r == 0 then
                r = a.name <=> b.name
              end
              r
            }
            .map{|x| x.name}
            .join(',')
    item << (sidfmd_host ? (sidfmd_host.refdate ? sidfmd_host.refdate.strftime('%Y/%m/%d') : '') : '')
    item << host['flg_delete'].to_s
    item << (sidfmd_host ? sidfmd_host.is_agent.to_s : '')
    item << (sidfmd_host ? (sidfmd_host.memo ? sidfmd_host.memo : '') : '')

    item
  end

  #new_exportのホストパッケージ情報
  def create_host_pkgs_info_for_new_export(sidfmd_host)
    items = []
    return items if !sidfmd_host
    queue = sidfmd_host.get_last_queue
    if queue then
      queue.sidfmd_queue_pkgs.sort_by{|pkg| pkg.pkg}.each do |pkg|
        item = []
        item << sidfmd_host.id.to_s
        item << 'packages'
        item << pkg.pkg
        items << item
      end
    end
    items
  end

  #new_exportのホスト手動登録プロダクト情報
  def create_host_manual_products_info_for_new_export(sidfmd_host)
    items = []
    return items if !sidfmd_host
    host_id = sidfmd_host.id
    product_id_with_version_ids = {}
    prods_data = {}
    prods_for_sort = {}
    sidfmd_host.sidfmd_host_apps.each do |app|
      prod = SidfmdProduct.find_by(:pid => app.product_id)
      if prod then
        prods_for_sort[prod.pid] = prod if !prods_for_sort[prod.pid]
        if app.version_id == nil || app.version_id == 0 then
          item = []
          item << host_id.to_s
          item << 'manual_products'
          item << prod.name
          prods_data[app.product_id] = item
        else
          product_id_with_version_ids[app.product_id] = [] unless product_id_with_version_ids[app.product_id]
          product_id_with_version_ids[app.product_id] << app.version_id
        end
      end
    end
    prods_for_sort.values.sort_by{|prod| prod.name}.each do |prod|
      items << prods_data[prod.pid] if prods_data[prod.pid]
      version_ids = product_id_with_version_ids[prod.pid]
      SidfmdVersion.where(:vid => version_ids).order(:version).each do |ver|
        item = []
        item << host_id.to_s
        item << 'manual_products'
        item << ver.fullname
        items << item
      end
    end
    items
  end

  #new_exportのホストローカルプロダクト情報
  def create_host_local_products_info_for_new_export(sidfmd_host)
    items = []
    return items if !sidfmd_host
    host_id = sidfmd_host.id
    lproduct_id_with_lversion_ids = {}
    lprods_data = {}
    lprods_for_sort = {}
    sidfmd_host.sidfmd_host_lapps.each do |lapp|
      lprod = SidfmdLproduct.find_by(:id => lapp.sidfmd_lproduct_id)
      if lprod then
        lprods_for_sort[lprod.id] = lprod if !lprods_for_sort[lprod.id]
        if lapp.sidfmd_lversion_id == 0 then
          item = []
          item << host_id.to_s
          item << 'local_products'
          item << lprod.name
          lprods_data[lapp.sidfmd_lproduct_id] = item
        else
          lproduct_id_with_lversion_ids[lapp.sidfmd_lproduct_id] = [] unless lproduct_id_with_lversion_ids[lapp.sidfmd_lproduct_id]
          lproduct_id_with_lversion_ids[lapp.sidfmd_lproduct_id] << lapp.sidfmd_lversion_id
        end
      end
    end
    lprods_for_sort.values.sort_by{|lprod| lprod.name}.each do |lprod|
      items << lprods_data[lprod.id] if lprods_data[lprod.id]
      lversion_ids = lproduct_id_with_lversion_ids[lprod.id]
      SidfmdLversion.where(:id => lversion_ids).order(:priority).each do |lver|
        item = []
        item << host_id.to_s
        item << 'local_products'
        item << lver.fullname
        items << item
      end
    end
    items
  end

  #new_exportの最新同期アプリケーション情報
  def create_host_apps_info_for_new_export(sidfmd_host)
    items = []
    return items if !sidfmd_host
    host_id = sidfmd_host.id
    queue = sidfmd_host.get_last_queue
    if queue then
      queue.sidfmd_queue_apps.sort{|a, b| 
        r = a.name <=> b.name 
        if r == 0 then
          r = a.version <=> b.version
        end
        r
      }.each do |app|
        item = []
        item << host_id.to_s
        item << 'queue_apps'
        item << "#{app.name}:#{app.version}:#{app.path}"
        items << item
      end
    end
    items
  end
 
  #new_exportの最新同期OS情報
  def create_host_os_info_for_new_export(sidfmd_host)
    item = []
    return item if !sidfmd_host
    host_id = sidfmd_host.id
    queue = sidfmd_host.get_last_queue
    if queue then
      dist = queue.sidfmd_queue_dist
      if dist then
        item << host_id.to_s
        item << 'queue_dist'
        item << "#{dist.ident}:#{dist.version}:#{dist.name}"
      end
    end
    item
  end

  def check_filters
    @hosts[:filters][:category_id] = '0' unless @hosts[:filters][:category_id]
    @hosts[:filters][:hostname] = '' unless @hosts[:filters][:hostname]
  end

  REFDATE_INVALID = Date.new(4000,12,31).freeze
  REFDATE_NONE = Date.new(3000,12,31).freeze

  def apply_filters 
    filter_hostname = @hosts[:filters][:hostname].downcase
    ret = @hosts[:hosts].find_all {|item| 
      # カテゴリ検索
      if @hosts[:filters][:category_id] != '0' then
        category_id = @hosts[:filters][:category_id].to_i
        if item[:sidfmd_host] then
          next false if category_id == -2
          categories = item[:sidfmd_host].sidfmd_categories
          categories = categories.where(:id => category_id) if category_id > 0
          categories_count = categories.count
          next false if (category_id == -1 and categories_count != 0)
          next false if (category_id > 0 and categories_count == 0)
        else
          next false if category_id != -2
        end
      end
      # ホスト名照合
      if @hosts[:filters][:hostname] != '' then
        next false unless item[:hostname].downcase[filter_hostname] || (item[:sidfmd_host] && item[:sidfmd_host][:ext_name] && item[:sidfmd_host][:ext_name].downcase[filter_hostname])
        #next false unless item[:hostname].downcase[@hosts[:filters][:hostname].downcase]
      end
      # ホスト詳細画面からリンクした時に、表示するホストを１つに絞る
      if @hostid_for_filter_one_host != '' then
        @hostid_for_filter_one_host = @hostid_for_filter_one_host.to_i
        if item[:sidfmd_host] && item[:sidfmd_host].id == @hostid_for_filter_one_host then
          target_host = SidfmdHost.find_by(:id => @hostid_for_filter_one_host)
          if target_host then
            if !@sidfm_roles.include?(:admin)then
              user_host_ids = SidfmdCategoriesHost.where(:sidfmd_category_id => @user_category_ids)
                                .map{|h| h.sidfmd_host_id}
              if !user_host_ids.include?(@hostid_for_filter_one_host) then
                next false
              else
                @hosts[:filters][:hostname] = target_host.hostname
              end
            else
              @hosts[:filters][:hostname] = target_host.hostname
            end
          else
            next false
          end
        else
          next false
        end
      end                    
      true
    }
    case @hosts[:order]
    when 'hostname_asc' then
      temp_ev = proc {|x|
        display_name = x[:sidfmd_host] ? (x[:sidfmd_host][:ext_name] ? x[:sidfmd_host][:ext_name] : x[:hostname]) : x[:hostname]
        [ x[:flg_delete] ? 1000 : 100, display_name.downcase ]
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[0] <=> yy[0]
        if r == 0 then
          r = xx[1] <=> yy[1]
        end
        r
      }
    when 'hostname_desc' then
      temp_ev = proc {|x|
        display_name = x[:sidfmd_host] ? (x[:sidfmd_host][:ext_name] ? x[:sidfmd_host][:ext_name] : x[:hostname]) : x[:hostname]
        [ x[:flg_delete] ? 1000 : 100, display_name.downcase ]
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[0] <=> yy[0]
        if r == 0 then
          r = xx[1] <=> yy[1]
        end
        r
      }
    when 'category_asc' then
      temp_ev = proc {|x|
        if x and x[:sidfmd_host] then
          cats = x[:sidfmd_host].sidfmd_categories.system
          if cats.size > 0 then
            [ x[:flg_delete] ? 1000 : 100, 100, cats.first.id ]
          else
            [ x[:flg_delete] ? 1000 : 100, 200, '' ]
          end
        else
          [ x[:flg_delete] ? 1000 : 100, 300, '' ]
        end
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[0] <=> yy[0]
        if r == 0 then
          r = xx[1] <=> yy[1]
          if r == 0 then
            r = xx[2] <=> yy[2]
          end
        end
        r
      }
    when 'category_desc' then
      temp_ev = proc {|x|
        if x and x[:sidfmd_host] then
          cats = x[:sidfmd_host].sidfmd_categories.system
          if cats.size > 0 then
            [ x[:flg_delete] ? 1000 : 100, 100, cats.first.id ]
          else
            [ x[:flg_delete] ? 1000 : 100, 200, '' ]
          end
        else
          [ x[:flg_delete] ? 1000 : 100, 300, '' ]
        end
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[0] <=> yy[0]
        if r == 0 then
          r = xx[1] <=> yy[1]
          if r == 0 then
            r = xx[2] <=> yy[2]
          end
        end
        r
      }
    when 'os_asc' then
      temp_ev = proc {|x|
        if x and x[:sidfmd_host] then
          os = x[:sidfmd_host].get_os
          if os and os.has_key?(:osname) then
            [ x[:flg_delete] ? 1000 : 100, 100, os[:osname] ]
          else
            [ x[:flg_delete] ? 1000 : 100, 200, '' ]
          end
        else
          [ x[:flg_delete] ? 1000 : 100, 300, '' ]
        end
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[0] <=> yy[0]
        if r == 0 then
          r = xx[1] <=> yy[1]
          if r == 0 then
            r = xx[2] <=> yy[2]
          end
        end
        r
      }
    when 'os_desc' then
      temp_ev = proc {|x|
        if x and x[:sidfmd_host] then
          os = x[:sidfmd_host].get_os
          if os and os.has_key?(:osname) then
            [ x[:flg_delete] ? 1000 : 100, 100, os[:osname] ]
          else
            [ x[:flg_delete] ? 1000 : 100, 200, '' ]
          end
        else
          [ x[:flg_delete] ? 1000 : 100, 300, '' ]
        end
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[0] <=> yy[0]
        if r == 0 then
          r = xx[1] <=> yy[1]
          if r == 0 then
            r = xx[2] <=> yy[2]
          end
        end
        r
      }
    when 'agentprds_asc' then
      temp_ev = proc {|x|
        if x and x[:sidfmd_host] then
          queue = x[:sidfmd_host].sidfmd_queues.order(:created_at => :desc).first
          if queue then
            [ x[:flg_delete] ? 1000 : 100, queue.sidfmd_queue_pkgs.size + queue.sidfmd_queue_apps.size ]
          else
            [ x[:flg_delete] ? 1000 : 100, -100 ]
          end
        else
          [ x[:flg_delete] ? 1000 : 100, -200 ]
        end
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[0] <=> yy[0]
        if r == 0 then
          r = yy[1] <=> xx[1]
        end
        r
      }
    when 'agentprds_desc' then
      temp_ev = proc {|x|
        if x and x[:sidfmd_host] then
          queue = x[:sidfmd_host].sidfmd_queues.order(:created_at => :desc).first
          if queue then
            [ x[:flg_delete] ? 1000 : 100, queue.sidfmd_queue_pkgs.size + queue.sidfmd_queue_apps.size ]
          else
            [ x[:flg_delete] ? 1000 : 100, -100 ]
          end
        else
          [ x[:flg_delete] ? 1000 : 100, -200 ]
        end
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[0] <=> yy[0]
        if r == 0 then
          r = yy[1] <=> xx[1]
        end
        r
      }
    when 'prds_asc' then
      temp_ev = proc {|x|
        if x and x[:sidfmd_host] then
          lapps = x[:sidfmd_host].sidfmd_host_apps
          if lapps then
            [ x[:flg_delete] ? 1000 : 100, lapps.count ]
          else
            [ x[:flg_delete] ? 1000 : 100, -100 ]
          end
        else
          [ x[:flg_delete] ? 1000 : 100, -200 ]
        end
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[0] <=> yy[0]
        if r == 0 then
          r = yy[1] <=> xx[1]
        end
        r
      }
    when 'prds_desc' then
      temp_ev = proc {|x|
        if x and x[:sidfmd_host] then
          lapps = x[:sidfmd_host].sidfmd_host_apps
          if lapps then
            [ x[:flg_delete] ? 1000 : 100, lapps.count ]
          else
            [ x[:flg_delete] ? 1000 : 100, -100 ]
          end
        else
          [ x[:flg_delete] ? 1000 : 100, -200 ]
        end
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[0] <=> yy[0]
        if r == 0 then
          r = yy[1] <=> xx[1]
        end
        r
      }
    when 'refdate_asc' then 
      temp_ev = proc {|x|
        if x and x[:sidfmd_host] then
          if x[:sidfmd_host].refdate then
            [ x[:flg_delete] ? 1000 : 100, x[:sidfmd_host].refdate ]
          else
            [ x[:flg_delete] ? 1000 : 100, REFDATE_NONE ]
          end
        else
          [ x[:flg_delete] ? 1000 : 100, REFDATE_INVALID ]
        end
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[0] <=> yy[0]
        if r == 0 then
          r = xx[1] <=> yy[1]
        end
        r
      }
    when 'refdate_desc' then
      temp_ev = proc {|x|
        if x and x[:sidfmd_host] then
          if x[:sidfmd_host].refdate then
            [ x[:flg_delete] ? 1000 : 100, x[:sidfmd_host].refdate ]
          else
            [ x[:flg_delete] ? 1000 : 100, REFDATE_NONE ]
          end
        else
          [ x[:flg_delete] ? 1000 : 100, REFDATE_INVALID ]
        end
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[0] <=> yy[0]
        if r == 0 then
          r = xx[1] <=> yy[1]
        end
        r
      }
    when 'memo_asc' then
      temp_ev = proc {|x|
        if x and x[:sidfmd_host] then
          if x[:sidfmd_host].memo then
            [ x[:flg_delete] ? 1000 : 100, 100 ]
          else
            [ x[:flg_delete] ? 1000 : 100, 200 ]
          end
        else
          [ x[:flg_delete] ? 1000 : 100, 300 ]
        end
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[0] <=> yy[0]
        if r == 0 then
          r = xx[1] <=> yy[1]
        end
        r
      }
    when 'memo_desc' then
      temp_ev = proc {|x|
        if x and x[:sidfmd_host] then
          if x[:sidfmd_host].memo then
            [ x[:flg_delete] ? 1000 : 100, 100 ]
          else
            [ x[:flg_delete] ? 1000 : 100, 200 ]
          end
        else
          [ x[:flg_delete] ? 1000 : 100, 300 ]
        end
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[0] <=> yy[0]
        if r == 0 then
          r = xx[1] <=> yy[1]
        end
        r
      }
    when 'manual_asc' then
      temp_ev = proc {|x|
        display_name = x[:sidfmd_host] ? (x[:sidfmd_host][:ext_name] ? x[:sidfmd_host][:ext_name] : x[:hostname]) : x[:hostname]
        [ x[:flg_delete] ? 1000 : 100, (x[:sidfmd_host] and !x[:sidfmd_host].is_agent ? 100 : 1000), display_name ]
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[0] <=> yy[0]
        if r == 0 then
          r = xx[1] <=> yy[1]
          if r == 0 then
            r = xx[2] <=> yy[2]
          end
        end
        r
      }
    when 'manual_desc' then
      temp_ev = proc {|x|
        display_name = x[:sidfmd_host] ? (x[:sidfmd_host][:ext_name] ? x[:sidfmd_host][:ext_name] : x[:hostname]) : x[:hostname]
        [ x[:flg_delete] ? 1000 : 100, (x[:sidfmd_host] and !x[:sidfmd_host].is_agent ? 100 : 1000), display_name ]
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[0] <=> yy[0]
        if r == 0 then
          r = xx[1] <=> yy[1]
          if r == 0 then
            r = xx[2] <=> yy[2]
          end
        end
        r
      }
    when 'agent_asc' then
      temp_ev = proc {|x| 
        display_name = x[:sidfmd_host] ? (x[:sidfmd_host][:ext_name] ? x[:sidfmd_host][:ext_name] : x[:hostname]) : x[:hostname]
        [ x[:flg_delete] ? 1000 : 100, (x[:sidfmd_host] and x[:sidfmd_host].is_agent ? 100 : 1000), display_name ]
      }
      ret.sort! {|x, y|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[0] <=> yy[0]
        if r == 0 then
          r = xx[1] <=> yy[1]
          if r == 0 then
            r = xx[2] <=> yy[2]
          end
        end
        r
      }
    when 'agent_desc' then
      temp_ev = proc {|x|
        display_name = x[:sidfmd_host] ? (x[:sidfmd_host][:ext_name] ? x[:sidfmd_host][:ext_name] : x[:hostname]) : x[:hostname]
        [ x[:flg_delete] ? 1000 : 100, (x[:sidfmd_host] and x[:sidfmd_host].is_agent ? 100 : 1000), display_name ]
      }
      ret.sort! {|y, x|
        xx = temp_ev.call(x)
        yy = temp_ev.call(y)
        r = xx[0] <=> yy[0]
        if r == 0 then
          r = xx[1] <=> yy[1]
          if r == 0 then
            r = xx[2] <=> yy[2]
          end
        end
        r
      }
    else
      # どれにも該当しない場合はホスト名検索とする。
      ret.sort! {|x, y| (x[:sidfmd_host] ? (x[:sidfmd_host][:ext_name] ? x[:sidfmd_host][:ext_name] : x[:hostname]) : x[:hostname]).downcase <=> (y[:sidfmd_host] ? (y[:sidfmd_host][:ext_name] ? y[:sidfmd_host][:ext_name] : y[:hostname]) : y[:hostname]).downcase }
      #ret.sort! {|x, y| x[:hostname].downcase <=> y[:hostname].downcase }
    end
    ret
  end

  def create_content
    ret = SidfmdHost.get_hosts_by_lm(:'sort[hostname]' => :asc)
    if ret then
      if @role_admin then
        ret[:hosts].delete_if {|x| x[:sidfmd_host].nil? }
      else
        ret[:hosts].delete_if {|x| x[:sidfmd_host].nil? or !@user_host_ids.include?(x[:sidfmd_host].id) }
      end
    end
    ret
  end

  public

  def index
    pages_params = params.fetch(:pages, {}).permit(:offset, :per_page, :filters, :order, :checked, :all_checked)
    @hosts = {}
    @hosts[:offset] = (pages_params[:offset] || 0).to_i
    @hosts[:per_page] = (pages_params[:per_page] || 100).to_i
    @hosts[:order] = (pages_params[:order] || 'hostname_asc')
    @hosts[:checked] = JSON.parse(pages_params[:checked] || '[]')
    @hosts[:all_checked] = ((pages_params[:all_checked] || '') == 'true')
    @hosts[:filters] = JSON.parse(pages_params[:filters] || '{}').symbolize_keys
  end

  def index_r
    pages_params = params.fetch(:pages, {}).permit(:offset, :per_page, :filters, :order, :checked, :all_checked, :host_id, :clear_session)
    if pages_params[:clear_session] then
      clear_import_session
      redirect_to :action => :index_r
    end
    @license_key = Sidfm.sidfmdlm_license_key
    @license_apikey = Sidfm.sidfmdlm_license_apikey
    @register_apikey = Sidfm.sidfmdlm_register_apikey
    @hosts = Rails.cache.fetch(@cache_prefix + '/hosts/0', SidfmControllerSupport::CACHE_OPTIONS) do
      create_content
    end
    if @hosts.nil? then
      Rails.cache.clear
      redirect_to :action => :license, :notice => 'ライセンスが不正です。' and return
    end
    if !@sidfm_roles.include?(:admin) then
      @hosts[:hosts].delete_if {|x| x[:flg_delete] }
    end
    @hosts[:offset] = (pages_params[:offset] || 0).to_i
    @hosts[:per_page] = (pages_params[:per_page] || 100).to_i
    @hosts[:order] = (pages_params[:order] || 'hostname_asc')
    @hosts[:checked] = JSON.parse(pages_params[:checked] || '[]')
    @hosts[:all_checked] = ((pages_params[:all_checked] || '') == 'true')
    @hosts[:filters] = JSON.parse(pages_params[:filters] || '{}').symbolize_keys
    @hostid_for_filter_one_host = (pages_params[:host_id] || '')#ホスト詳細画面からリンクした時、１つのホストに絞るための変数
    check_filters
    @hosts[:hosts_filtered] = apply_filters
    @all_hostnames =  @hosts[:hosts_filtered].map{|host| if host[:flg_delete] then nil else host[:hostname] end}.compact
    @hosts[:content] = @hosts[:hosts_filtered].slice(@hosts[:offset], @hosts[:per_page])
    @hosts[:content] = [] unless @hosts[:content]
    @hosts[:total_size] = @hosts[:hosts_filtered].size
 
    @system_categories = SidfmdCategory.system
    @not_system_categories = @role_admin ? SidfmdCategory.not_system : SidfmdCategory.where(:id => @user_category_ids)
    @new_import_notices = session[:import_notices] ? session[:import_notices] : nil
    clear_import_session
    #ブラウザバックかどうか判断するための処理
    #if !session['chkbok_ids'] then
    #  session['chkbok_ids'] = {}
    #end
    #if !session['chkbok_ids']['hostadmin'] then
    #  session['chkbok_ids']['hostadmin'] = {}
    #end
    #session['chkbok_ids']['hostadmin']['is_browser_back'] = false
  end

  def license
    @license_key = Sidfm.sidfmdlm_license_key
    @license_apikey = Sidfm.sidfmdlm_license_apikey
    @register_apikey = Sidfm.sidfmdlm_register_apikey
    #filtered_params = params.fetch(:params, {}).permit(:offset, :limit, :hostname, :'sort[hostname]', :'sort[flg_delete]', :'sort[created_at]')
    @hosts = SidfmdHost.get_hosts_by_lm(:'sort[hostname]' => :asc)
  end

  def license_setting
    @license_key = Sidfm.sidfmdlm_license_key
    @license_apikey = Sidfm.sidfmdlm_license_apikey
  end

  def license_update
    filtered_params = params.fetch(:params, {}).permit(:license_key, :license_password)
    @license_key = filtered_params[:license_key]
    license_password = filtered_params[:license_password]
    begin
      if !SidfmdHost.auth_api(@license_key, license_password) then
        redirect_to :action => :license_setting, :notice => 'ライセンス認証に失敗しました。' and return
      end
    rescue Exception => e
      log_exception(e)
      redirect_to :action => :license_setting, :notice => 'ライセンスサーバへのアクセスに失敗しました。' and return
    end
    Rails.cache.clear

    redirect_to :action => :license
  end

  def license_apikey_update
    res = SidfmdHost.license_apikey_update
    if res.respond_to?(:[]) and res[:success] then
      redirect_to :action => :license and return
    end
    redirect_to :action => :license, :notice => 'APIキーの更新に失敗しました。'
  end

  # ホストAPIキーの更新
  # deprecated:使用せず。
  def host_apikey_update
    filtered_params = params.fetch(:params, {}).permit(:id, :hostname)
    host_id = filtered_params[:id]
    hostname = filtered_params[:hostname]
    if host_id and hostname then
      SidfmdHost.transaction do
        res = SidfmdHost.update_host_apikey(host_id.to_i)
        if res.respond_to?(:[]) and res[:success] then
          host = SidfmdHost.where(:hostname => hostname).first
          if host then
            host.update_attribute :host_apikey, res[:apikey]
          else
            res = SidfmdHost.establish_consistency
          end
          redirect_to :action => :index and return
        end
      end
    end
    redirect_to :action => :index_r, :notice => 'APIキーの更新に失敗しました。'
  end

  # 削除状態の変更
  def toggle_host_delete
    filtered_params = params.permit(:host_id, :sidfmd_host_id, :rule_info)
    err_message = ''
    r = {}
    host_id = filtered_params[:host_id]
    sidfmd_host_id = filtered_params[:sidfmd_host_id]
    if host_id && sidfmd_host_id then
      host = SidfmdHost.find_by(:id => sidfmd_host_id)
      if host then
        if !@role_admin and !@user_host_ids.include?(host.id) then
          err_message = '自分の属するカスタムカテゴリのホスト以外は操作できません。'
          render :layout => false, :json => { :err_message => err_message } and return
        end
      end
      res = nil
      begin
        rule_info = JSON.parse(filtered_params[:rule_info] || '{}').symbolize_keys
        res = SidfmdHost.toggle_host_delete(host_id.to_i)
        unless res then
          err_message = 'ライセンスサーバとの通信に失敗しました。'
          render :layout => false, :json => { :err_message => err_message } and return
        end
        unless res[:success] then
          err_message = 'ライセンス数を超過するため、操作に失敗しました。'
          render :layout => false, :json => { :err_message => err_message } and return
        end
      rescue Exception => e
        log_exception(e)
        err_message = 'ライセンスサーバへのアクセスに失敗しました。'
        render :layout => false, :json => { :err_message => err_message } and return
      end
      host = nil
      last_checked_at = nil
      SidfmdHost.transaction do
        host = SidfmdHost.find_by(:id => sidfmd_host_id)
        if host then
          host.lock!
          hostid = host.id
          new_flg_delete = !host[:flg_delete]
          host.update_attribute :flg_delete, new_flg_delete
          if new_flg_delete then
            # データを削除する。Redmineのチケットは削除しない（VM上からは直接参照できない）。
            last_checked_at = host.last_checked_at
            host.last_checked_at = nil
            host.memo = nil
            host.is_agent = false
            host.save
            # カテゴリの削除
            SidfmdCategoriesHost.where(:sidfmd_host_id => hostid).delete_all
            # プロダクトの削除
            SidfmdHostApp.where(:sidfmd_host_id => hostid).delete_all
            SidfmdHostLapp.where(:sidfmd_host_id => hostid).delete_all
            # キューの削除
            sidfmd_queue_ids = SidfmdQueue.where(:sidfmd_host_id => hostid).collect {|x| x.id }.uniq
            sidfmd_queue_ids.each do |sidfmd_queue_id|
              SidfmdQueueDist.where(:sidfmd_queue_id => sidfmd_queue_id).delete_all
              SidfmdQueuePkg.where(:sidfmd_queue_id => sidfmd_queue_id).delete_all
              SidfmdQueueApp.where(:sidfmd_queue_id => sidfmd_queue_id).delete_all
              SidfmdQueueLapp.where(:sidfmd_queue_id => sidfmd_queue_id).delete_all
              SidfmdQueueWin.where(:sidfmd_queue_id => sidfmd_queue_id).delete_all
              sidfmd_queue_vul_ids = SidfmdQueueVul.where(:sidfmd_queue_id => sidfmd_queue_id).collect {|x| x.id }.uniq
              sidfmd_queue_vul_ids.each do |sidfmd_queue_vul_id|
                SidfmdQueueVulCve.where(:sidfmd_queue_vul_id => sidfmd_queue_vul_id).delete_all
                SidfmdQueueVulSri.where(:sidfmd_queue_vul_id => sidfmd_queue_vul_id).delete_all
                SidfmdQueueVulAffprod.where(:sidfmd_queue_vul_id => sidfmd_queue_vul_id).delete_all
                SidfmdQueueVulAffver.where(:sidfmd_queue_vul_id => sidfmd_queue_vul_id).delete_all
                SidfmdQueueVulAffpkg.where(:sidfmd_queue_vul_id => sidfmd_queue_vul_id).delete_all
              end
              SidfmdQueueVul.where(:sidfmd_queue_id => sidfmd_queue_id).delete_all
            end
            SidfmdQueue.where(:id => sidfmd_queue_ids).delete_all
            # チケット（脆弱性）の削除
            sidfmd_issue_ids = SidfmdIssue.where(:sidfmd_host_id => hostid).collect {|x| x.id }.uniq
            sidfmd_issue_ids.each do |sidfmd_issue_id|
              SidfmdIssueCve.where(:sidfmd_issue_id => sidfmd_issue_id).delete_all
              SidfmdIssueSri.where(:sidfmd_issue_id => sidfmd_issue_id).delete_all
              SidfmdIssueAffprod.where(:sidfmd_issue_id => sidfmd_issue_id).delete_all
              SidfmdIssueAffver.where(:sidfmd_issue_id => sidfmd_issue_id).delete_all
              SidfmdIssueAffpkg.where(:sidfmd_issue_id => sidfmd_issue_id).delete_all
            end
            SidfmdIssue.where(:id => sidfmd_issue_ids).delete_all
            #ルールまたはルールの条件削除
            delete_rule_ids = rule_info[:delete_rule_ids]
            delete_cond_ids = rule_info[:delete_cond_ids]
            delete_host_cond_ids = rule_info[:delete_host_cond_ids]
            if delete_rule_ids && delete_rule_ids.size > 0 then
              SidfmdRule.where(:id => delete_rule_ids).destroy_all
            end 
            if delete_cond_ids && delete_cond_ids.size > 0 then
              SidfmdRuleCond.where(:id => delete_cond_ids).destroy_all
            end
            if delete_host_cond_ids && delete_host_cond_ids.size > 0 then
              SidfmdRuleCmHostSub.where(:id => delete_host_cond_ids).destroy_all
            end
          end
        end
      end

      unless host then
        SidfmdHost.transaction do
          res = SidfmdHost.establish_consistency
          logger.info("SidfmdHost.establish_consistensy=#{res.inspect}")
        end
      end
      Rails.cache.clear
      r = { :err_message => err_message }
      
    else
      err_message = '削除状態の変更に失敗しました。'
      r = { :err_message => err_message }
    end

    render :layout => false, :json => r
  end

  # ホスト名変更
  def rename_host
    filtered_params = params.permit(:hostid, :hostname, :old_hostname)

    hostid = filtered_params[:hostid]
    hostname = filtered_params[:hostname]
    old_hostname = filtered_params[:old_hostname]
    err_message = ''
    unless hostid or hostname or old_hostname then
      err_message = 'パラメータが不正です。'
      render :layout => false, :json => { :err_message => err_message } and return
    end
    if hostid.size == 0 or hostname.size == 0 or old_hostname.size == 0 then
      err_message = 'パラメータが不正です。'
      render :layout => false, :json => { :err_message => err_message } and return
    end
   
    if hostid and hostname then
      if /\A[0-9a-zA-Z._-]+\z/ !~ hostname then
        err_message = 'ホスト名が不正です。'
        render :layout => false, :json => { :err_message => err_message } and return
      end
      if hostname.bytesize > SidfmdHost::MAX_HOSTNAME_BYTE_LENGTH then
        err_message = SidfmdHost::MAX_HOSTNAME_BYTE_LENGTH.to_s + 'バイトを超えるホスト名は登録できません。'
        render :layout => false, :json => { :err_message => err_message } and return
      end
      unless @role_admin then
        host = SidfmdHost.where(:hostname => old_hostname).first
        if host then
          if !@user_host_ids.include?(host.id) then
            err_message = '自分の属するカスタムカテゴリのホスト以外は操作できません。'
            render :layout => false, :json => { :err_message => err_message } and return
          end
        else
          err_message = 'ホストが存在しません。'
          render :layout => false, :json => { :err_message => err_message } and return
        end
      end
      SidfmdHost.transaction do
        res = SidfmdHost.rename_host(hostid, hostname)
        if res.respond_to?(:[]) then
          if res[:success] then
            host = SidfmdHost.where(:hostname => old_hostname).first
            if host then
              host.update_attribute :hostname, hostname
            else
              res = SidfmdHost.establish_consistency
            end
            Rails.cache.clear
          else
            err_message = '登録済のホスト名、もしくは不正な文字を指定したため、操作に失敗しました。'
            render :layout => false, :json => { :err_message => err_message } and return
          end
        else
          err_message = 'ライセンスサーバとの通信に失敗しました。'
          render :layout => false, :json => { :err_message => err_message } and return
        end
      end
    else
      err_message = 'ホスト名の変更に失敗しました。'
      render :layout => false, :json => { :err_message => err_message } and return
    end
    render :layout => false, :json => { :err_message => err_message }
  end

  def set_display_name
    filtered_params = params.permit(:hostid, :display_name)
    hostid = filtered_params[:hostid]
    display_name = filtered_params[:display_name]
    err_message = ''
    unless hostid || display_name then
      err_message = 'パラメータが不正です。'
      render :layout => false, :json => { :err_message => err_message } and return
    end
    if hostid.size == 0 then
      err_message = 'パラメータが不正です。'
      render :layout => false, :json => { :err_message => err_message } and return
    end
    if display_name.size > SidfmdHost::MAX_DISPLAY_NAME_LENGTH then
      err_message = '100文字を超える表示名は登録できません。'
      render :layout => false, :json => { :err_message => err_message } and return
    end
    if display_name =~ /<|>|"|'/ then
      err_message = '表示名に使用できない文字が含まれています。'
      render :layout => false, :json => { :err_message => err_message } and return
    elsif display_name.size > 0 && display_name.strip.size === 0 then
      err_message = '空白文字のみの表示名は登録できません。'
      render :layout => false, :json => { :err_message => err_message } and return 
    elsif display_name =~ /&&&&&/ then
      err_message = '「&&&&&」の文字列(5つ連なるアンド)は使用できません。'
      render :layout => false, :json => { :err_message => err_message } and return
    end
    if hostid && display_name then
      host = SidfmdHost.find_by(:id => hostid)
      unless @role_admin then
        if host && !@user_host_ids.include?(host.id) then
          err_message = '自分の属するカスタムカテゴリのホスト以外は操作できません。'
          render :layout => false, :json => { :err_message => err_message } and return
        end
      end
      SidfmdHost.transaction do
        if host then
          display_name = nil if display_name.size == 0
          host.update_attribute :ext_name, display_name
          Rails.cache.clear
        else
          err_message = 'ホストが存在しません。'
          render :layout => false, :json => { :err_message => err_message } and return
        end
      end
    else
      err_message = '表示名の変更に失敗しました。'
      render :layout => false, :json => { :err_message => err_message } and return
    end
    render :layout => false, :json => { :err_message => err_message } and return
  end

  # メモ更新
  def update_memo
    filtered_params = params.fetch(:params, {}).permit(:hostname, :memo, :dump)
    dump = {}
    begin
      dump = JSON.parse(filtered_params[:dump])
    rescue JSON::ParserError => e
      log_exception(e)
      dump[:action] = :index_r
      dump[:notice] = '入力が不正です。'
      redirect_to dump and return
    end
    if dump["pages[filters]"] then
      dump["pages[filters]"] = JSON.generate(dump["pages[filters]"])
    end
    dump[:action] = :index_r
    hostname = filtered_params[:hostname]
    memo = filtered_params[:memo]
    unless hostname or memo then
      dump[:notice] = '入力が不正です。'
      redirect_to dump and return
    end
    if hostname.size == 0 then
      dump[:notice] = '入力が不正です。'
      redirect_to dump and return
    end
    if memo.length > SidfmdHost::MEMO_MAX_LEN then
      redirect_to :action => :index_r, :notice => 'メモに入力可能な最大文字数を超えています' and return
    end
    SidfmdHost.transaction do
      host = SidfmdHost.where(:hostname => hostname).first
      unless host then
        dump[:notice] = 'ホストが存在しません。'
        redirect_to dump and return
      end
      host.update_attribute :memo, memo.size > 0 ? memo : nil
      Rails.cache.clear
    end
    redirect_to dump
  end

  # メモ一括更新
  def update_memo_bulk
    filtered_params = params.fetch(:params, {}).permit(:hostnames, :memo, :dump)
    dump = {}
    begin
      dump = JSON.parse(filtered_params[:dump])
    rescue JSON::ParserError => e
      log_exception(e)
      dump[:action] = :index_r
      dump[:notice] = '入力が不正です。'
      redirect_to dump and return
    end
    dump[:action] = :index_r
    hostnames = filtered_params[:hostnames]
    memo = filtered_params[:memo]
    unless hostnames or memo then
      dump[:notice] = '入力が不正です。'
      redirect_to dump and return
    end
    if memo.size > SidfmdHost::MEMO_MAX_LEN then
      dump[:notice] = '入力が不正です。'
      redirect_to dump and return
    end
    begin
      hostnames = JSON.parse(hostnames)
    rescue JSON::ParserError => e
      log_exception(e)
      dump[:notice] = '入力が不正です。'
      redirect_to dump and return
    end
    if hostnames.size == 0 then
      dump[:notice] = '入力が不正です。'
      redirect_to dump and return
    end
    memo = nil if memo.size == 0
    SidfmdHost.transaction do
      hostnames.each do |hostname|
        host = SidfmdHost.where(:hostname => hostname).first
        if host then
          host.update_attribute :memo, memo
        end
      end
      Rails.cache.clear
    end
    redirect_to dump
  end

  # 基準日更新
  def update_refdate
    filtered_params = params.fetch(:params, {}).permit(:hostname, :refdate, :dump)
    dump = {}
    begin
      dump = JSON.parse(filtered_params[:dump])
    rescue JSON::ParserError => e
      log_exception(e)
      dump[:action] = :index_r
      dump[:notice] = '入力が不正です。'
      redirect_to dump and return
    end
    if dump["pages[filters]"] then
      dump["pages[filters]"] = JSON.generate(dump["pages[filters]"])
    end
    dump[:action] = :index_r
    hostname = filtered_params[:hostname]
    refdate = filtered_params[:refdate]
    unless hostname or refdate then
      dump[:notice] = '入力が不正です。'
      redirect_to dump and return
    end
    if hostname.size == 0 then
      dump[:notice] = '入力が不正です。'
      redirect_to dump and return
    end
    SidfmdHost.transaction do
      host = SidfmdHost.where(:hostname => hostname).first
      unless host then 
        dump[:notice] = 'ホストが存在しません。'
        redirect_to dump and return
      end
      if refdate.size > 0 then
        begin
          refdate = Date.parse(refdate) if refdate.size > 0
        rescue ArgumentError => e
          log_exception(e)
          dump[:notice] = '不正な日付です。'
          redirect_to dump and return
        end
      else
        refdate = nil
      end
      begin
        host.update_attribute :refdate, refdate
        Rails.cache.clear
      rescue ActiveRecord::StatementInvalid => e
        log_exception(e)
        dump[:notice] = '不正な日付です。'
        redirect_to dump and return
      end
    end
    redirect_to dump
  end

  # 基準日一括更新
  def update_refdate_bulk
    filtered_params = params.fetch(:params, {}).permit(:hostnames, :refdate, :dump)
    dump = {}
    begin
      dump = JSON.parse(filtered_params[:dump])
    rescue JSON::ParserError => e
      log_exception(e)
      dump[:action] = :index_r
      dump[:notice] = '入力が不正です。'
      redirect_to dump and return
    end
    dump[:action] = :index_r
    hostnames = filtered_params[:hostnames]
    refdate = filtered_params[:refdate]
    unless hostnames or refdate then
      dump[:notice] = '入力が不正です。'
      redirect_to dump and return
    end
    begin
      hostnames = JSON.parse(hostnames)
    rescue JSON::ParserError => e
      log_exception(e)
      dump[:notice] = '入力が不正です。'
      redirect_to dump and return
    end
    if hostnames.size == 0 then
      dump[:notice] = '入力が不正です。'
      redirect_to dump and return
    end
    if refdate.size > 0 then
      begin
        refdate = Date.parse(refdate) if refdate.size > 0
      rescue ArgumentError => e
        log_exception(e)
        dump[:notice] = '不正な日付です。'
        redirect_to dump and return
      end
    else
      refdate = nil
    end
    begin
      SidfmdHost.transaction do
        hostnames.each do |hostname|
          host = SidfmdHost.where(:hostname => hostname).first
          if host then
            host.update_attribute :refdate, refdate
          end
        end
      end
      Rails.cache.clear
    rescue ActiveRecord::StatementInvalid => e
      log_exception(e)
      dump[:notice] = '不正な日付です。'
      redirect_to dump and return
    end
    redirect_to dump
  end

  # カテゴリ更新
  def update_categories
    #filtered_params = params.fetch(:categories, {}).permit(:sidfmdhostid, :hostname, :selected_category_ids, :system_category_id, :dump, :rule_info)
    #dump = {}
    
    filtered_params = params.permit(:sidfmdhostid, :hostname, :selected_category_ids, :system_category_id, :rule_info)
    err_message = ''
    begin
    #  dump = JSON.parse(filtered_params[:dump])
      rule_info = JSON.parse(filtered_params[:rule_info] || '{}').symbolize_keys
    rescue JSON::ParserError => e
      log_exception(e)
      err_message = '入力が不正です。'
      render :layout => false, :json => { :err_message => err_message } and return
    end
    #if dump["pages[filters]"] then
    #  dump["pages[filters]"] = JSON.generate(dump["pages[filters]"])
    #end
    #dump[:action] = :index_r
    sidfmdhostid = filtered_params[:sidfmdhostid].to_i
    hostname = filtered_params[:hostname]
    selected_category_ids = filtered_params[:selected_category_ids]
    system_category_id = filtered_params[:system_category_id].to_i
    if sidfmdhostid == 0 or system_category_id == 0 then
      err_message = '入力が不正です。'
      render :layout => false, :json => { :err_message => err_message } and return
    end
    if !hostname or !selected_category_ids then
      err_message = '入力が不正です。'
      render :layout => false, :json => { :err_message => err_message } and return
    end
    begin
      selected_category_ids = JSON.parse(selected_category_ids)
    rescue JSON::ParserError => e
      log_exception(e)
      err_message = '入力が不正です。'
      render :layout => false, :json => { :err_message => err_message } and return
    end
    if !@role_admin then
      if !@user_host_ids.include?(sidfmdhostid) then
        err_message = '入力が不正です。'
        render :layout => false, :json => { :err_message => err_message } and return
      end
      if selected_category_ids.size == 0 then
        err_message = '少なくとも一つはカスタムカテゴリを指定する必要があります。'
        render :layout => false, :json => { :err_message => err_message } and return
      else
        diff = selected_category_ids & @user_category_ids
        if diff.size == 0 then
          err_message = '自分の属するカスタムカテゴリ以外を指定することはできません。'
          render :layout => false, :json => { :err_message => err_message } and return
        end
        selected_category_ids = diff
      end
    end
    begin
      host = nil
      SidfmdHost.transaction do
        host = SidfmdHost.where(:id => sidfmdhostid).first
        unless host then
          host = SidfmdHost.where(:hostname => hostname).first
          unless host then
            err_message = '対象ホストが見つかりません。'
            render :layout => false, :json => { :err_message => err_message } and return
          end
        end
        #関連するルールの削除・変更
        delete_rule_ids = rule_info[:delete_rule_ids]
        delete_cond_ids = rule_info[:delete_cond_ids]
        delete_host_cond_ids = rule_info[:delete_host_cond_ids]
        if delete_rule_ids && delete_rule_ids.size > 0 then
          SidfmdRule.where(:id => delete_rule_ids).destroy_all
        end
        if delete_cond_ids && delete_cond_ids.size > 0 then
          SidfmdRuleCond.where(:id => delete_cond_ids).destroy_all
        end
        if delete_host_cond_ids && delete_host_cond_ids.size > 0 then
          SidfmdRuleCmHostSub.where(:id => delete_host_cond_ids).destroy_all
        end
        host.sidfmd_categories = []
        res = host.modify_categories(system_category_id, selected_category_ids,
                                     @role_admin ? nil : @user_category_ids)
        unless res then
          raise StandardError, 'modify_categories failed'
        end
      end

    rescue Exception => e
      log_exception(e)
      err_message = 'カテゴリ更新に失敗しました。'
      render :layout => false, :json => { :err_message => err_message } and return
    end
    Rails.cache.clear
    render :layout => false, :json => { :err_message => err_message }
  end

  # カテゴリ一括変更
  def update_categories_bulk
    #filtered_params = params.fetch(:categories, {}).permit(:hostnames, :selected_category_ids, :system_category_id, :dump, :rule_info)
    #dump = {}
    filtered_params = params.permit(:hostnames, :selected_category_ids, :system_category_id, :rule_info)
    err_message = ''
    begin
      #dump = JSON.parse(filtered_params[:dump])
      rule_info = JSON.parse(filtered_params[:rule_info] || '{}').symbolize_keys
    rescue JSON::ParserError => e
      log_exception(e)
      err_message = '入力が不正です。'
      render :layout => false, :json => { :err_message => err_message } and return
    end
    #dump[:action] = :index_r
    hostnames = filtered_params[:hostnames]
    unless hostnames then
      err_message = '入力が不正です。'
      render :layout => false, :json => { :err_message => err_message } and return
    end
    begin
      hostnames = JSON.parse(hostnames)
    rescue JSON::ParserError => e
      log_exception(e)
      err_message = '入力が不正です。'
      render :layout => false, :json => { :err_message => err_message } and return
    end
    system_category_id = filtered_params[:system_category_id]
    unless system_category_id then
      err_message = '入力が不正です。'
      render :layout => false, :json => { :err_message => err_message } and return
    end
    selected_category_ids = filtered_params[:selected_category_ids]
    unless selected_category_ids then
      err_message = '入力が不正です。'
      render :layout => false, :json => { :err_message => err_message } and return
    end
    begin
      selected_category_ids = JSON.parse(selected_category_ids)
    rescue JSON::ParserError => e
      log_exception(e)
      err_message = '入力が不正です。'
      render :layout => false, :json => { :err_message => err_message } and return
    end
    if !@role_admin then
      target_host_ids = SidfmdHost.where(:hostname => hostnames).pluck(:id)
      if (target_host_ids - @user_host_ids).size > 0 then
        err_message = '入力が不正です。'
        render :layout => false, :json => { :err_message => err_message } and return
      end
      if selected_category_ids.size == 0 then
        err_message = '少なくとも一つはカスタムカテゴリを指定する必要があります。'
        render :layout => false, :json => { :err_message => err_message } and return
      else
        diff = selected_category_ids & @user_category_ids
        if diff.size == 0 then
          err_message = '自分の属するカスタムカテゴリ以外を指定することはできません。'
          render :layout => false, :json => { :err_message => err_message } and return
        end
        selected_category_ids = diff
      end
    end
    begin
      host_ids = []
      SidfmdHost.transaction do
        #関連するルールの削除・変更
        delete_rule_ids = rule_info[:delete_rule_ids]
        delete_cond_ids = rule_info[:delete_cond_ids]
        delete_host_cond_ids = rule_info[:delete_host_cond_ids]
        if delete_rule_ids && delete_rule_ids.size > 0 then
          SidfmdRule.where(:id => delete_rule_ids).destroy_all
        end
        if delete_cond_ids && delete_cond_ids.size > 0 then
          SidfmdRuleCond.where(:id => delete_cond_ids).destroy_all
        end
        if delete_host_cond_ids && delete_host_cond_ids.size > 0 then
          SidfmdRuleCmHostSub.where(:id => delete_host_cond_ids).destroy_all
        end

        SidfmdHost.where(:hostname => hostnames).each do |host|
          host.sidfmd_categories = []
          res = host.modify_categories(system_category_id, selected_category_ids, @role_admin ? nil : @user_category_ids)
          unless res then
            logger.info('modify_categories failed')
            raise StandardError
          else
            host_ids << host.id if host.last_checked_at
          end
        end
      end
      Rails.cache.clear
    rescue Exception => e
      log_exception(e)
      err_message = 'カテゴリ更新に失敗しました。'
      render :layout => false, :json => { :err_message => err_message } and return
    end
    render :layout => false, :json => { :err_message => err_message } and return
  end

  def register_apikey_update
    res = Sidfm.sidfmdlm_register_apikey_update
    if res then
      redirect_to :action => :license
      return
    end
    redirect_to :action => :license, :notice => 'APIキーの登録・更新に失敗しました。'
  end

  def register_apikey_remove
    Sidfm.sidfmdlm_register_apikey = nil
    unless Sidfm.sidfmdlm_register_apikey then
      redirect_to :action => :license
      return
    end
    redirect_to :action => :license, :notice => 'APIキーの削除に失敗しました。'
  end

  # ホストを登録する。
  def register_host
    f_params = params.fetch(:params, {}).permit(:hostname, :display_name, :category_id, :selected, :selected_category_ids, :refdate, :memo, :selected_lproducts)
    hostname = f_params[:hostname]
    display_name = f_params[:display_name]
    if /\A[0-9a-zA-Z._-]+\z/ !~ hostname then
      redirect_to :action => :index_r, :notice => 'ホスト名が不正です。' and return
    end
    if hostname.bytesize > SidfmdHost::MAX_HOSTNAME_BYTE_LENGTH then
      redirect_to :action => :index_r, :notice => SidfmdHost::MAX_HOSTNAME_BYTE_LENGTH.to_s + 'バイトを超えるホスト名は登録できません。' and return
    end
    if display_name.size > SidfmdHost::MAX_DISPLAY_NAME_LENGTH then
      redirect_to :action => :index_r, :notice => '100文字を超える表示名は登録できません。' and return
    end
    if display_name =~ /<|>|"|'/ then
      redirect_to :action => :index_r, :notice => '表示名に使用できない文字が含まれています。' and return
    elsif display_name.size > 0 && display_name.strip.size === 0 then
      redirect_to :action => :index_r, :notice => '空白文字のみの表示名は登録できません。' and return
    elsif display_name =~ /&&&&&/ then
      redirect_to :action => :index_r, :notice => '「&&&&&」の文字列(5つ連なるアンド)は使用できません。' and return
    end
    category_id = f_params[:category_id].to_i
    unless category_id > 0 then
      redirect_to :action => :index_r, :notice => 'システムカテゴリが指定されていません。' and return
    end
    selected_category_ids = f_params[:selected_category_ids]
    unless selected_category_ids then
      redirect_to :action => :index_r, :notice => 'カスタムカテゴリが設定されていません。' and return
    end
    begin
      selected_category_ids = JSON.parse(selected_category_ids)
    rescue JSON::ParserError => e
      log_exception(e)
      redirect_to :action => :index_r, :notice => 'カスタムカテゴリが不正です。' and return
    end
    unless selected_category_ids then
      redirect_to :action => :index_r, :notice => '入力が不正です。' and return
    end
    if !@role_admin then
      if selected_category_ids.size == 0 then
        redirect_to :action => :index_r, :notice => '少なくとも一つはカスタムカテゴリを指定する必要があります。' and return
      else
        diff = selected_category_ids & @user_category_ids
        if diff.size == 0 then
          redirect_to :action => :index_r, :notice => '自分の属するカスタムカテゴリ以外を指定することはできません。' and return
        end
        selected_category_ids = diff
      end
    end
    selected = f_params[:selected]
    selected_lproducts = f_params[:selected_lproducts]
    if !selected || !selected_lproducts then
      redirect_to :action => :index_r, :notice => '入力が不正です。' and return
    end
    begin
      selected = JSON.parse(selected)
      selected_lproducts = JSON.parse(selected_lproducts)
    rescue JSON::ParserError => e
      log_exception(e)
      redirect_to :action => :index_r, :notice => '入力が不正です。' and return
    end
    refdate = f_params[:refdate]
    unless refdate then
      redirect_to :action => :index_r, :notice => '基準日が不正です。' and return
    end
    if refdate.size > 0 then
      begin
        refdate = Date.parse(refdate) if refdate.size > 0
      rescue ArgumentError => e
        log_exception(e)
        redirect_to :action => :index_r, :notice => '不正な日付です。' and return
      end
    else
      redirect_to :action => :index_r, :notice => '基準日は必須です。' and return
    end
    memo = f_params[:memo]
    unless memo then
      redirect_to :action => :index_r, :notice => '入力が不正です。' and return
    end
    if memo.length > SidfmdHost::MEMO_MAX_LEN then
      redirect_to :action => :index_r, :notice => 'メモに入力可能な最大文字数を超えています' and return
    end    
    memo = nil if memo.size == 0

    # 入力チェック
    if hostname.size == 0 then
      redirect_to :action => :index_r, :notice => 'ホスト名が入力されていません。'
      return
    end
    if category_id <= 0 then
      redirect_to :action => :index_r, :notice => 'カテゴリが不正です。'
      return
    end

    unregistrable_product_ids = SidfmdProduct.unregistrable_product_ids
    selected.each_pair do |k, v|
      v.each do |vv|
        pid = vv['pid'].to_i
        vid = vv['vid'].to_i
        if vid == 0 and unregistrable_product_ids.include?(pid) then
          redirect_to :action => :index_r, :notice => 'プロダクトに、「全て」を登録することができないプロダクトが含まれています。' and return
        end
      end
    end

    # 既に登録されている場合はreject
    # 削除済ならそのまま通り（有効になる）。
    old_host = SidfmdHost.where(:hostname => hostname).where(:flg_delete => false)
    if old_host and old_host.count > 0 then
      redirect_to :action => :index_r, :notice => '同名のホストが登録されています。' and return
    end

    @hosts = SidfmdHost.get_hosts_by_lm(:limit => 0)

    # ライセンスチェック
    # なお、ライセンスサーバ側でもチェックが行われるため、このチェックを除外しても
    # ライセンス数を超えてホストの登録を行うことはできない。
    if @hosts and @hosts[:license_number] <= @hosts[:used_license_number] then
      redirect_to :action => :index_r, :notice => 'ホスト数がライセンスの上限に達しています。'
      return
    end

    # ライセンスサーバとVMでホストの登録を行う。
    temp = SidfmdHost.register_host(hostname, nil, false)
    if temp.nil? then
      redirect_to :action => :index_r, :notice => 'ホスト登録に失敗しました。' and return
    end

    # 作成されたホストを取得
    host = SidfmdHost.where(:hostname => hostname).first
    if host.nil? then
      redirect_to :action => :index_r, :notice => 'ホスト登録に失敗しました。' and return
    end

    begin
      SidfmdHost.transaction do
        #表示名登録
        display_name = nil if display_name.size == 0
        host.update_attribute :ext_name, display_name

        # カテゴリ登録
        SidfmdCategoriesHost.create(:sidfmd_host_id => host.id,
                                    :sidfmd_category_id => category_id)
        res = host.modify_categories(category_id, selected_category_ids)
        unless res then
          raise StandardError
        end

        # 基準日登録
        host.update_attribute :refdate, refdate

        # メモ登録
        host.update_attribute :memo, memo

        # アプリ登録
        host.sidfmd_host_apps = []
        host.sidfmd_host_lapps = []
        host.save
 
        prods = []
        lprods = []
        selected.each_pair do |k, v|
          v.each do |vv|
            pid = vv['pid'].to_i
            vid = vv['vid'].to_i
            prods << SidfmdHostApp.new(
              :sidfmd_host_id => host.id,
              :product_id => pid,
              :version_id => vid
            )
          end
        end
        selected_lproducts.each do |k, v|
          v.each do |vv|
            pid = vv['pid'].to_i
            vid = vv['vid'].to_i
            lprods << SidfmdHostLapp.new(
              :sidfmd_host_id => host.id,
              :sidfmd_lproduct_id => pid,
              :sidfmd_lversion_id => vid
            )
          end
        end 
        if (prods.size + lprods.size) <= SidfmdHost::MAX_REGISTER_PRODUCTS then
          SidfmdHostApp.import(prods) if prods.any?
          SidfmdHostLapp.import(lprods) if lprods.any?
        end
      end
  
      SidfmdHostView.sync_data([ host.id ])#Viewに反映
      Rails.cache.clear
     
    rescue ActiveRecord::StatementInvalid => e
      log_exception(e)
      redirect_to :action => :index, :notice => '不正な日付です。' and return
    rescue StandardError => e
      log_exception(e)
      redirect_to :action => :index, :notice => 'カテゴリの更新に失敗しました。' and return
    end
    redirect_to :action => :index
  end

  # 設定されているアプリケーションを更新する。
  def update_applications
    f_params = params.permit(:hostname, :selected, :selected_lproducts)
    err_message = ''
    begin
      selected = JSON.parse(f_params[:selected])
      selected_lproducts = JSON.parse(f_params[:selected_lproducts])
    rescue JSON::ParserError => e
      log_exception(e)
      err_message = '入力が不正です。'
      render :layout => false, :json => { :err_message => err_message } and return 
    end
  
    unregistrable_product_ids = SidfmdProduct.unregistrable_product_ids 
    all_selected_products_size = 0
    selected.each_pair do |k, v|
      v.each do |vv|
        pid = vv['pid'].to_i
        vid = vv['vid'].to_i
        if vid == 0 and unregistrable_product_ids.include?(pid) then
          err_message = 'プロダクトに、「全て」を登録することができないプロダクトが含まれています。'
          render :layout => false, :json => { :err_message => err_message } and return
        end
      end
      all_selected_products_size += v.size
    end
    selected_lproducts.each_pair do |k, v|
      all_selected_products_size += v.size
    end

    if all_selected_products_size > SidfmdHost::MAX_REGISTER_PRODUCTS then
      err_message = '入力が不正です。'
      render :layout => false, :json => { :err_message => err_message } and return
    end

    hostname = f_params[:hostname]

    host = nil
    SidfmdHost.transaction do
      host = SidfmdHost.where(:hostname => hostname).first
      unless host then
        err_message = 'ホストが存在しません'
        render :layout => false, :json => { :err_message => err_message } and return
      end
      host.sidfmd_host_apps = []
      host.sidfmd_host_lapps = []
      host.save

      apps = []
      lapps = []
      selected.each_pair do |k, v|
        v.each do |vv|

          pid = vv['pid'].to_i
          vid = vv['vid'].to_i
          apps << SidfmdHostApp.new(:sidfmd_host_id => host.id,
                                    :product_id => pid,
                                    :version_id => vid)
        end
      end
      selected_lproducts.each do |k, v|
        v.each do |vv|
          pid = vv['pid'].to_i
          vid = vv['vid'].to_i
          lapps << SidfmdHostLapp.new(:sidfmd_host_id => host.id,
                                      :sidfmd_lproduct_id => pid,
                                      :sidfmd_lversion_id => vid)
        end
      end
    
      SidfmdHostApp.import(apps) if apps.any?
      SidfmdHostLapp.import(lapps) if lapps.any?
    end
    Rails.cache.clear
 
    render :layout => false, :json => { :err_message => err_message } and return
  end

  # プロダクト一括更新
  def update_applications_bulk
    filtered_params = params.permit(:hostnames, :selected, :selected_lproducts)
    err_message = ''
    hostnames = filtered_params[:hostnames]
    selected = filtered_params[:selected]
    selected_lproducts = filtered_params[:selected_lproducts]
    unless hostnames or selected or selected_lproducts then
      err_message = '入力が不正です。'
      render :layout => false, :json => { :err_message => err_message } and return
    end
    begin
      hostnames = JSON.parse(hostnames)
    rescue JSON::ParserError => e 
      log_exception(e)
      err_message = '入力が不正です。'
      render :layout => false, :json => { :err_message => err_message } and return
    end
    begin
      selected = JSON.parse(selected)
      selected_lproducts = JSON.parse(selected_lproducts)
    rescue JSON::ParserError => e
      log_exception(e)
      err_message = '入力が不正です。'
      render :layout => false, :json => { :err_message => err_message } and return
    end

    unregistrable_product_ids = SidfmdProduct.unregistrable_product_ids
    all_selected_products_size = 0
    selected.each_pair do |k, v|
      v.each do |vv|
        pid = vv['pid'].to_i
        vid = vv['vid'].to_i
        if vid == 0 and unregistrable_product_ids.include?(pid) then
          err_message = 'プロダクトに、「全て」を登録することができないプロダクトが含まれています。'
          render :layout => false, :json => { :err_message => err_message } and return
        end
      end
      all_selected_products_size += v.size
    end
    selected_lproducts.each_pair do |k, v|
      all_selected_products_size += v.size
    end

    if all_selected_products_size > SidfmdHost::MAX_REGISTER_PRODUCTS then
      err_message = '入力が不正です。'
      render :layout => false, :json => { :err_message => err_message } and return
    end

    host_ids = []
    apps = []
    lapps = []

    SidfmdHost.transaction do
      hostnames.each do |hostname|
        host = SidfmdHost.where(:hostname => hostname).first
        if host then
          host.sidfmd_host_apps = []
          host.sidfmd_host_lapps = []
          host.save

          selected.each_pair do |k, v|
            v.each do |vv|
              pid = vv['pid'].to_i
              vid = vv['vid'].to_i
              apps << SidfmdHostApp.new(:sidfmd_host_id => host.id,
                                        :product_id => pid,
                                        :version_id => vid)
            end
          end
          
          selected_lproducts.each do |k, v|
            v.each do |vv|
              pid = vv['pid'].to_i
              vid = vv['vid'].to_i
              lapps << SidfmdHostLapp.new(:sidfmd_host_id => host.id,
                                          :sidfmd_lproduct_id => pid,
                                          :sidfmd_lversion_id => vid)
            end
          end
        end
      end
    end

    SidfmdHostApp.import(apps) if apps.any?
    SidfmdHostLapp.import(lapps) if lapps.any?
   
    Rails.cache.clear
    
    render :layout => false, :json => { :err_message => err_message }
  end

  # ホスト一覧のインポートを行い、操作予定の内容を出力する
  def import
    @actions = session[:import_actions]
    @errs = session[:import_errs]
    @notices = session[:import_notices]
    @is_new_import = session[:is_new_import]
    #ルールに関係するデータ
    @rule_del_host_ids = []
    @rule_cat_change_hostnames = {}
    @actions.each do |action|
      action_name = action['action']
      if action_name == 'modify_host' then
        a_options = action['options']
        a_options.each do |op|
          if op['action'] == 'category_change' then
            hostname = action['old_data']['hostname']
            @rule_cat_change_hostnames[hostname] = op['new']
          end
        end
      elsif action_name == 'remove_host' then
        old_data = action['old_data']
        old_data.each do |od|
          host_info = od['host']['sidfmd_host']
          if host_info then
            host_id = host_info['id']
            @rule_del_host_ids << host_id
          end
        end
      end
    end
  end

  def new_import
    f_params = params.permit(:import_file)
    zip_file = f_params[:import_file]
    err = false
    unless zip_file then
      logger.info('invalid zip file')
      err = true
    end
    hosts_info_file = nil
    hosts_info_file_lines = []
    host_opts_info_files = []
    new_host_opts_info_files = []
    begin
      Zip::File.open(zip_file) do |zip|
        zip.each do |entry|
          next if entry.name && (entry.name.index('_MACOSX') || entry.name.index('.DS_Store'))
          if entry.file? then
            if entry.name && entry.name.index('hostinfo.csv') then
              hosts_info_file = entry.get_input_stream.read.gsub(/^$\n/, '')
            elsif entry.name && entry.name.index('regist') then
              file = entry.get_input_stream.read
              if file.getbyte(0) == 255 && file.getbyte(1) == 254 then
                file = file[2, file.length - 2]
              end
              file = file.force_encoding('UTF-16LE').encode('UTF-8')
              new_host_opts_info_files << {:name => entry.name, :file => file}
            elsif entry.name && entry.name.index('opt') then
              file = entry.get_input_stream.read
              if file.getbyte(0) == 255 && file.getbyte(1) == 254 then
                file = file[2, file.length - 2]
              end
              file = file.force_encoding('UTF-16LE').encode('UTF-8')
              host_opts_info_files << {:name => entry.name, :file => file}
            end
          end
        end
      end
    rescue Exception => e
      log_exception(e)
      err = true
    end

    unless hosts_info_file then
      logger.info('hostinfo.csv is not found')
      err = true
    else
      if hosts_info_file.getbyte(0) == 255 && hosts_info_file.getbyte(1) == 254 then
        hosts_info_file = hosts_info_file[2, hosts_info_file.length - 2]
      end
      hosts_info_file = hosts_info_file.force_encoding('UTF-16LE').encode('UTF-8')
      hosts_info_file = CSV.parse(hosts_info_file, :col_sep => "\t", :headers => :first_row)
      hosts_info_file.each do |line|
        hosts_info_file_lines << { 
          'ホスト登録ID' => (line['ホスト登録ID'] && line['ホスト登録ID'].size > 0) ? line['ホスト登録ID'] : nil,
          'ホスト削除' => (line['ホスト削除'] && line['ホスト削除'].size > 0) ? line['ホスト削除'] : nil,
          'ホストID' => (line['ホストID'] && line['ホストID'].size > 0) ? line['ホストID'] : nil,
          'トークン' => (line['トークン'] && line['トークン'].size > 0) ? line['トークン'] : nil,
          'ホスト名' => (line['ホスト名'] && line['ホスト名'].size > 0) ? line['ホスト名'] : nil,
          '表示名' => (line['表示名'] && line['表示名'].size > 0) ? line['表示名'] : nil, 
          'カテゴリ' => (line['カテゴリ'] && line['カテゴリ'].size > 0) ? line['カテゴリ'] : nil,
          '基準日' => (line['基準日'] && line['基準日'].size > 0) ? line['基準日'] : nil,
          '削除' => (line['削除'] && line['削除'].size > 0) ? line['削除'] : nil,
          'Agentフラグ' => (line['Agentフラグ'] && line['Agentフラグ'].size > 0) ? line['Agentフラグ'] : nil,
          'メモ' => (line['メモ'] && line['メモ'].size > 0) ? line['メモ'] : nil
        }
      end
    end
    clear_import_session
    session[:import_actions] = []
    session[:import_errs] = []
    session[:import_notices] = []
    session[:is_new_import] = true
    session[:hosts_info_file_lines] = hosts_info_file_lines
    session[:host_opts_info_files] = host_opts_info_files
    session[:new_host_opts_info_files] = new_host_opts_info_files
    session[:host_opts_info] = {}
    session[:new_host_opts_info] = {}
    begin
      lmhosts = SidfmdHost.get_hosts_by_lm(:'sort[hostname]' => :asc)
    rescue Exception => e
      log_exception(e)
      err = true
    end
    unless lmhosts and lmhosts[:success] then
      logger.info('failed to get hosts')
      err = true
    end
    session[:lmhosts] = lmhosts
    total_size = hosts_info_file_lines.size + host_opts_info_files.size + new_host_opts_info_files.size
   
    render :layout => false, :json => {:err => err, :total_size => total_size}
  end

  def progress_import_opts
    errs = []
    is_exception = false
    product_names = SidfmdProduct.all.map{|p| [ p.name.strip.tr('０-９ａ-ｚＡ-Ｚ','0-9a-zA-Z').gsub('：', ':'), p.name ] }.to_h
    version_names = SidfmdVersion.all.map{|v| [ v.fullname.tr('０-９ａ-ｚＡ-Ｚ','0-9a-zA-Z').gsub('：', ':'), v.fullname ] }.to_h
    lproduct_names = SidfmdLproduct.all.map{|lp| [ lp.name.strip.tr('０-９ａ-ｚＡ-Ｚ','0-9a-zA-Z').gsub('：', ':'), lp.name ] }.to_h
    lversion_names = SidfmdLversion.all.map{|lv| [ lv.fullname.tr('０-９ａ-ｚＡ-Ｚ','0-9a-zA-Z').gsub('：', ':'), lv.fullname ] }.to_h

    unregistrable_product_ids = SidfmdProduct.unregistrable_product_ids

    create_opts_info = ->(opts_file, file_name){
      opts_csv = CSV.parse(opts_file, :col_sep => "\t", :headers => :first_row)
      line_num = 1
      opts_csv.each do |line|
        line_num += 1
        new_host_id = line['ホスト登録ID']
        host_id = line['ホストID']
        if new_host_id && host_id then
          errs << "#{file_name}の#{line_num}行目:「ホスト登録ID」列と「ホストID」の両方が入力されています。"
          next
        end
        item_name = line['種類']
        item_data = line['データ']
        next if !item_data || item_data.strip.size == 0 || !item_name || item_name.strip.size == 0
        item_data = item_data.gsub("\n", '').strip.tr('０-９ａ-ｚＡ-Ｚ','0-9a-zA-Z').gsub('：', ':')
        item_name = item_name.gsub("\n", '').strip.tr('ａ-ｚＡ-Ｚ','a-zA-Z')
        if item_name != 'packages' && item_name != 'manual_products' && item_name != 'local_products' && item_name != 'queue_apps' && item_name != 'queue_dist'then
          errs << "#{file_name}の#{line_num}行目: 「種類」列の値が正しくありません。"
          next
        end
        info_hash = {
          :packages => [],
          :manual_products => [],
          :local_products => [],
          :queue_apps => [],
        }
        if new_host_id && new_host_id.strip.size > 0 then
          new_host_id = new_host_id.strip
          unless session[:new_host_opts_info][new_host_id] then
            session[:new_host_opts_info][new_host_id] = info_hash
          end
          if item_name == 'queue_dist' then
            if item_data.count(':') != 2 then
              errs << "#{file_name}の#{line_num}行目: queue_distの「データ」列の値が正しくありません。"
              next
            end
            if session[:new_host_opts_info][new_host_id][item_name.strip.to_sym] then
              session[:import_notices] << "#{file_name}の#{line_num}行目: ホスト登録ID「#{new_host_id}」のqueue_distのデータが複数あります。"
            end
            session[:new_host_opts_info][new_host_id][item_name.strip.to_sym] = item_data
          else
            if item_name == 'queue_apps' && item_data.count(':') != 2 then
              errs << "#{file_name}の#{line_num}行目: queue_appsの「データ」列の値が正しくありません。"
              next
            elsif item_name == 'manual_products' then
              if !product_names[item_data] && !version_names[item_data] then
                errs << "#{file_name}の#{line_num}行目: manual_productsの「データ」列の値に、登録されていないプロダクト名が入力されています。" 
                next
              elsif !version_names[item_data] then
                pid = SidfmdProduct.find_by(:name => item_data).pid
                if unregistrable_product_ids.include?(pid) then
                  errs << "#{file_name}の#{line_num}行目: manual_productsの「データ」列の値に、 仕様によりバージョンを指定する場合にのみしか登録することができないプロダクト名が入力されています。"
                  next
                end
              end
            elsif item_name == 'local_products' && !lproduct_names[item_data] && !lversion_names[item_data] then
              errs << "#{file_name}の#{line_num}行目: local_productsの「データ」列の値に、登録されていないプロダクト名が入力されています。"
              next
            end
            session[:new_host_opts_info][new_host_id][item_name.to_sym] << case item_name
                                                                           when 'manual_products' then 
                                                                             (product_names[item_data] || version_names[item_data])
                                                                           when 'local_products'
                                                                             (lproduct_names[item_data] || lversion_names[item_data])
                                                                           else
                                                                             item_data
                                                                           end

          end
        elsif host_id && host_id.strip.size > 0 then
          host_id = host_id.strip
          unless session[:host_opts_info][host_id] then
            session[:host_opts_info][host_id] = info_hash
          end
          if item_name == 'queue_dist' then
            if item_data.count(':') != 2 then
              errs << "#{file_name}の#{line_num}行目: queue_distの「データ」列の値が正しくありません。"
              next
            end
            if session[:host_opts_info][host_id][item_name.strip.to_sym] then
              session[:import_notices] << "#{file_name}の#{line_num}行目: ホストID「#{host_id}」のqueue_distのデータが複数あります。"
            end
            session[:host_opts_info][host_id][item_name.strip.to_sym] = item_data
          else
            if item_name == 'queue_apps' && item_data.count(':') != 2 then
              errs << "#{file_name}の#{line_num}行目: queue_appsの「データ」列の値が正しくありません。"
              next
            elsif item_name == 'manual_products' then
              if !product_names[item_data] && !version_names[item_data] then
                errs << "#{file_name}の#{line_num}行目: manual_productsの「データ」列の値に、登録されていないプロダクト名が入力されています。"
                next
              elsif !version_names[item_data] then
                pid = SidfmdProduct.find_by(:name => item_data).pid
                if unregistrable_product_ids.include?(pid) then
                  errs << "#{file_name}の#{line_num}行目: manual_productsの「データ」列の値に、 仕様によりバージョンを指定する場合にのみしか登録することができないプロダクト名が入力されています。"
                  next
                end
              end
            elsif item_name == 'local_products' && !lproduct_names[item_data] && !lversion_names[item_data] then
              errs << "#{file_name}の#{line_num}行目: local_productsの「データ」列の値に、登録されていないプロダクト名が入力されています。"
              next
            end
            session[:host_opts_info][host_id][item_name.to_sym] << case item_name
                                                                   when 'manual_products' then
                                                                     (product_names[item_data] || version_names[item_data])
                                                                   when 'local_products'
                                                                     (lproduct_names[item_data] || lversion_names[item_data])
                                                                   else
                                                                     item_data
                                                                   end

          end
        else
          next
        end
      end
    }

    begin
      is_complete = false
      executed_size = 0
      if session[:host_opts_info_files].size > 0 then
        files = session[:host_opts_info_files].shift(100)
        executed_size = files.size
        files.each{|file| create_opts_info.(file[:file], file[:name])}
      elsif session[:new_host_opts_info_files].size > 0 then
        files = session[:new_host_opts_info_files].shift(100)
        executed_size = files.size
        files.each{|file| create_opts_info.(file[:file], file[:name])}
      else
        is_complete = true
      end 
    rescue Exception => e
      log_exception(e)
      is_exception = true
    end
    session[:import_errs] = session[:import_errs] + errs
    render :layout => false, :json => {:err => is_exception, :is_complete => is_complete, :executed_size => executed_size}
  end

  def progress_import_hosts
    #importの処理を進める
    @hosts = session[:lmhosts].symbolize_keys[:hosts]
    @actions = []
    all_category_names = []
    all_s_category_names =  []
    SidfmdCategory.all.each do |cat|
      all_category_names << cat.name
      all_s_category_names << cat.name if cat.is_system
    end
    old_tokens = []
    @hosts.each do |host|
      if host['sidfmd_host'] then
        apikey = host['sidfmd_host']['host_apikey']
        old_tokens << apikey if apikey
      end
    end
    if !@role_admin then
      user_c_cat_names = User.current.sidfmd_categories.map{|c| c.name}
    end

    get_filtered_category_names = ->(target_category_names, user_category_names, all_system_category_names){
      cat_names = target_category_names.split(/,/)
                                       .map{|c| c.strip}
                                       .uniq
      filtered_cats = []
      s_cat_names = []
      cat_names.each do |c_name|
        if all_system_category_names.include?(c_name) then
          if s_cat_names.size == 0 then
            filtered_cats << c_name
          end
          s_cat_names << c_name
        elsif user_category_names.include?(c_name) then
          filtered_cats << c_name
        end
      end
      filtered_cats  = filtered_cats.sort{|a, b|
        r = (all_system_category_names.include?(a) ? 1: 10) <=> (all_system_category_names.include?(b) ? 1 : 10)
        if r == 0 then
          r = a <=> b
        end
        r
      }.join(',')
      return { :filtered_category_names => filtered_cats, :system_category_names => s_cat_names }
    }

    is_err = false
    @errs = []
    @notices = []
    host_opts_info = session[:host_opts_info]
    new_host_opts_info = session[:new_host_opts_info]
    session[:new_host_ids] = [] unless session[:new_host_ids]
    session[:import_hostnames] = {} unless session[:import_hostnames]
    hostname_errs = []
    executed_size = 0;
    session[:line_num] = 1 unless session[:line_num]
    session[:hosts_info_file_lines].shift(100).each do |line|
      session[:line_num] += 1
      executed_size += 1
      item = {}
      item[:new_host_id] = line['ホスト登録ID']
      item[:flg_host_delete] = line['ホスト削除']
      item[:host_id] = line['ホストID']
      item[:token] = line['トークン']
      item[:hostname] = line['ホスト名']
      item[:ext_name] = line['表示名']
      item[:categories] = line['カテゴリ']
      item[:refdate] = line['基準日']
      item[:flg_delete] = line['削除']
      item[:is_agent] = line['Agentフラグ']
      item[:memo] = line['メモ']
      item[:new_host_id] = item[:new_host_id].strip.gsub("\n", '').tr('０-９','0-9').to_i if item[:new_host_id] && item[:new_host_id].strip.size > 0
      item[:flg_host_delete] = item[:flg_host_delete].strip.gsub("\n", '') if item[:flg_host_delete]
      item[:host_id] = item[:host_id].strip.gsub("\n", '').tr('０-９','0-9').to_i if item[:host_id] && item[:host_id].strip.size > 0
      item[:token] = item[:token].strip.gsub("\n", '').tr('０-９ａ-ｚＡ-Ｚ','0-9a-zA-Z') if item[:token]
      if !item[:hostname] || item[:hostname].strip.size == 0 then
        @errs << "hostinfo.csvの#{session[:line_num]}行目:「ホスト名」列が入力されておりません。"
        next
      end
      item[:hostname] = item[:hostname].gsub("\n", ' ').strip.tr('０-９ａ-ｚＡ-Ｚ','0-9a-zA-Z').gsub('ー', '-')
      check = check_hostname(item[:hostname])
      if check != true then
        msg = check
        @errs << "hostinfo.csvの#{session[:line_num]}行目: #{msg}"
        next
      end
      if session[:import_hostnames][item[:hostname]] then
        hostname_errs << "hostinfo.csvの#{session[:import_hostnames][item[:hostname]]}行目と#{session[:line_num]}行目:「ホスト名」が重複しています。※同じホスト名を複数登録することはできません。"
        is_err = true
        #break
      else
         session[:import_hostnames][item[:hostname]] = session[:line_num]
      end

      item[:ext_name] = item[:ext_name].gsub("\n", ' ').strip if item[:ext_name]
      if item[:ext_name] and item[:ext_name] =~ /&&&&&/ then
        @errs << "hostinfo.csvの#{session[:line_num]}行目: 「&&&&&」の文字列(5つ連なるアンド)は使用できません。"
        next
      end
      item[:categories] = item[:categories].gsub("\n", '') if item[:categories]
      item[:flg_delete] = item[:flg_delete].gsub("\n", '').strip if item[:flg_delete]
     
      if item[:flg_host_delete] && item[:flg_host_delete].downcase != 'true' && item[:flg_host_delete].downcase != 'false' then
        @errs << "hostinfo.csvの#{session[:line_num]}行目: 「ホスト削除」列は\"TRUE\"と入力して下さい。"
        next
      end
 
      if item[:flg_delete] && item[:flg_delete].downcase != 'true' && item[:flg_delete].downcase != 'false' then
        @errs << "hostinfo.csvの#{session[:line_num]}行目: 「削除」列は\"TRUE\"もしくは \"FALSE\"と入力して下さい。"
        next
      end

      item[:is_agent] = 'false' if !item[:is_agent] || item[:is_agent].size == 0
      item[:is_agent] = item[:is_agent].strip.gsub("\n", '')
      
      if item[:refdate] && item[:refdate].size > 0 then
        begin
          item[:refdate] = item[:refdate].gsub("\n", '').strip.tr('０-９','0-9').gsub('ー', '/').gsub('-', '/').sub('年', '/').sub('月', '/').sub('日', '').tr('０-９ａ-ｚＡ-Ｚ','0-9a-zA-Z')
          refdate = Date.parse(item[:refdate])
        rescue ArgumentError => e
          log_exception(e)
          msg = "hostinfo.csvの#{session[:line_num]}行目:「基準日」列の値が正しくありません。"
          @errs << msg
        end
      end
      
      if item[:token] && item[:token].size > 0 && !old_tokens.include?(item[:token]) then
        #ライセンスサーバーとDBのホストのapikeyが異なるホストはimportの対象から除外する
        next
      end
    
      if (item[:host_id] && item[:host_id].size > 0) && (!item[:token] || item[:token].size == 0) then
        msg = "hostinfo.csvの#{session[:line_num]}行目:「ホストID」があり「トークン」がありません。更新には両方の値が必要です。"
        @errs << msg
        next
      elsif (!item[:host_id] || item[:host_id].size == 0 ) && (item[:token] && item[:token].size > 0) then
        msg = "hostinfo.csvの#{session[:line_num]}行目:「トークン」があり「ホストID」ありません。更新には両方の値が必要です。"
        @errs << msg
        next
      end

      if !item[:is_agent] && item[:host_id] && item[:token] then
        if item[:flg_host_delete] != 'TRUE' && item[:flg_host_delete] != 'true' then 
          @errs << "hostinfo.csvの#{session[:line_num]}行目:「Agentフラグ」列が入力されておりません。"
          next
        end
      end 

      if item[:host_id] && item[:token] && item[:is_agent] && item[:is_agent].downcase != 'true' && item[:is_agent].downcase != 'false' then
        @errs << "hostinfo.csvの#{session[:line_num]}行目:「Agentフラグ」列は \"TRUE\"もしくは \"FALSE\"と入力して下さい。"
        next
      end
      
      if !@role_admin && item[:host_id] && !@user_host_ids.include?(item[:host_id]) then
        @errs << "hostinfo.csvの#{session[:line_num]}行目: ログインユーザでの操作は許可されておりません。"
        next
      end

      #削除チェック
      if item[:flg_host_delete] && item[:flg_host_delete].downcase == 'true' then
        t_host = @hosts.find {|y| item[:token] == y['apikey'] && !y['flg_delete']}
        if t_host then
          @actions << {
            :action => 'remove_host',
            :action_name => '削除',
            :old_data => [ { :token => item[:token], :host => t_host } ],
            :new_data => nil,
            :options => nil,
          }
        end
        next
      end
      if item[:categories] then
        #カテゴリ名をfilter
        if @role_admin then
          filtered_category_names = get_filtered_category_names.(item[:categories], all_category_names, all_s_category_names)
        else
          filtered_category_names = get_filtered_category_names.(item[:categories], user_c_cat_names, all_s_category_names)
        end
        filtered_cats = filtered_category_names[:filtered_category_names]
        s_cat_names = filtered_category_names[:system_category_names]
        #システムカテゴリが無いホストはエラー
        if s_cat_names.size == 0 && (item[:flg_delete].downcase == 'false' || (!item[:host_id] && !item[:token])) then
          msg = "hostinfo.csvの#{session[:line_num]}行目:「カテゴリ」列に、有効なシステムカテゴリ名が入力されていません。"
          @errs << msg
          next
        end
        if !@role_admin then
          f_cats = filtered_cats.split(/,/).reject{|c_name| all_s_category_names.include?(c_name)}
          if f_cats.size == 0 then
            msg = "hostinfo.csvの#{session[:line_num]}行目: ログインユーザのカテゴリから外れるホストは操作できません。"
            @errs << msg
            next
          end
        end
        item[:categories] = filtered_cats
      else
        if (item[:flg_delete] && item[:flg_delete].downcase == 'false') || (!item[:host_id] && !item[:token])then
          msg = "hostinfo.csvの#{session[:line_num]}行目:「カテゴリ」列に、カテゴリが入力されていません。"  
          @errs << msg
          next
        end
      end
      #新規登録
      if (item[:new_host_id] && item[:new_host_id].size > 0) || (!item[:host_id] && !item[:token]) then
        if session[:new_host_ids].include?(item[:new_host_id]) then
          session[:import_notices] = ['hostinfo.csvの「ホスト登録ID」列に、同じIDが入力されているデータが複数あります。']
          is_err = true
          break
        end
        if SidfmdHost.where(:flg_delete => false).where(:hostname => item[:hostname]).count > 0 then
          msg = "hostinfo.csvの#{session[:line_num]}行目: 既存の削除されていないホストと同じホスト名は使用できません。"
          @errs << msg
          next
        end
        session[:new_host_ids] << item[:new_host_id] if item[:new_host_id]
        #報告者権限でログインユーザのカスタムカテゴリに属さないものはエラー
        if !@role_admin && item[:categories].split(/,/).size == 1 then
          msg = "hostinfo.csvの#{session[:line_num]}行目: ログインユーザのカスタムカテゴリの範囲外となるため、登録の対象となりません。"
          @errs << msg
          next
        end
        if !item[:refdate] then
          #新規登録で基準日がないデータは登録当日の値にする
          @notices << "hostinfo.csvの#{session[:line_num]}行目: 基準日が入力されていない新規ホストデータには、本日の日付が登録されます"
          item[:refdate] = Date.today.strftime('%Y/%m/%d')
        end
        opts_info = new_host_opts_info[item[:new_host_id].to_s]
        if opts_info then
          opts_info = opts_info.symbolize_keys
          item[:packages] = opts_info[:packages].uniq
          item[:apps] = opts_info[:queue_apps].uniq
          item[:manual_products] = opts_info[:manual_products].uniq
          item[:local_products] = opts_info[:local_products].uniq
          item[:dist] = opts_info[:queue_dist]
          if item[:is_agent].downcase == 'false' then
            if (item[:packages] && item[:packages].size > 0) || (item[:apps] && item[:apps].size > 0) || (item[:dist] && item[:dist].size > 0) then
              @notices << "hostinfo.csvの#{session[:line_num]}行目:「Agentフラグ」列が \"FALSE \"のホストには、registフォルダ下の「packages」「queue_apps」「queue_dist」のデータは登録できません。※脆弱性同期時に自動生成されます。"
              item[:packages] = []
              item[:apps] = []
              item[:dist] = nil
            end
          end
          if (item[:manual_products].size + item[:local_products].size) > SidfmdHost::MAX_REGISTER_PRODUCTS then
            @notices << "hostinfo.csvの#{session[:line_num]}行目: プロダクトの最大登録数は#{SidfmdHost::MAX_REGISTER_PRODUCTS}個のため、このホストのプロダクトの登録は行われません。optsフォルダ以下のファイルをご確認下さい。"
            item[:manual_products] = []
            item[:local_products] = []
          end
        end
        @actions << {
          :action => 'create_new_host',
          :action_name => '新規作成',
          :new_data => item,
          :old_data => nil,
          :options => nil,
          #:new_host_id => item[:new_host_id]
        }
        next
      end

      same_hostname = @hosts.find {|x| x['hostname'] == item[:hostname] }
  
      if item[:flg_delete] and item[:flg_delete].downcase == 'false' and same_hostname and same_hostname['sidfmd_host'] and same_hostname['sidfmd_host']['host_apikey'] != item[:token] and !same_hostname['flg_delete'] then
        msg = "hostinfo.csvの#{session[:line_num]}行目: 既存の削除されていないホストと同じホスト名は使用できません。"
        @errs << msg
        next
      end
      if same_hostname and same_hostname['flg_delete'] and same_hostname['sidfmd_host'] and
         item[:flg_delete] and item[:flg_delete].downcase == 'false'
      then
      # 復元
        if item[:categories] then
          if s_cat_names.size > 1 then
            @notices << "hostinfo.csvの#{session[:line_num]}行目: システムカテゴリの登録は1つのみ可能なため、システムカテゴリは「#{s_cat_names[0]}」が登録されます。"
          end
        end
        host_opts = host_opts_info[item[:host_id].to_s.strip]
        if host_opts then
          host_opts = host_opts.symbolize_keys
          if item[:is_agent].downcase == 'false' && ((host_opts[:packages] && host_opts[:packages].size > 0) || (host_opts[:apps] && host_opts[:apps].size > 0) || (host_opts[:dist] && host_opts[:dist].size > 0)) then
            @notices << "hostinfo.csvの#{session[:line_num]}行目:「Agentフラグ」列が \"FALSE \"のホストには、optsフォルダ下の「packages」「queue_apps」「queue_dist」のデータは登録できません。※脆弱性同期時に自動生成されます。"
          else
            item[:packages] = host_opts[:packages].uniq
            item[:apps] = host_opts[:queue_apps].uniq
            item[:dist] = host_opts[:queue_dist]
          end
          item[:manual_products] = host_opts[:manual_products].uniq
          item[:local_products] = host_opts[:local_products].uniq
          if (item[:manual_products].size + item[:local_products].size) > SidfmdHost::MAX_REGISTER_PRODUCTS then
            @notices << "hostinfo.csvの#{session[:line_num]}行目: プロダクトの最大登録数は#{SidfmdHost::MAX_REGISTER_PRODUCTS}個のため、このホストのプロダクトの登録は行われません。optsフォルダ以下のファイルをご確認下さい。"
            item[:manual_products] = []
            item[:local_products] = []
          end
        end
        @actions << {
          :action => 'undelete_host',
          :action_name => '復元',
          :new_data => item,
          :old_data => same_hostname,
          :options => nil,
        }
        next
      end
      same_token = @hosts.find {|x| x['apikey'] == item[:token]}
      # 更新のチェック
      if same_token and same_token['sidfmd_host'] then
        target = []
        # カテゴリ比較
        i_categories = item[:categories] ? item[:categories].split(/,/) : []
        t_host = SidfmdHost.find_by(:id => same_token['sidfmd_host']['id'])
        if !t_host then
          @err << [ "hostinfo.csvの#{session[:line_num]}行目: 変更対象のホストが削除されています。" ]
          next
        end
        same_token['sidfmd_host'] = t_host
        if @role_admin then
          s_categories = t_host.sidfmd_categories.map{|x| x.name.strip}
        else
          s_categories = t_host.sidfmd_categories.map{|x|
            (user_c_cat_names.include?(x.name.strip) || x.is_system) ? x.name.strip : nil
          }.compact
        end
        s_categories = s_categories.uniq.sort{|a, b|
          r = (all_s_category_names.include?(a) ? 1: 10) <=> (all_s_category_names.include?(b) ? 1 : 10)
          if r == 0 then
            r = a <=> b
          end
          r
        }
        #報告者権限で、ログインユーザのカスタムカテゴリが対象ホストのカスタムカテゴリから外れる場合はエラーとする
        if !@role_admin && s_categories.size > 1 && filtered_cats.size == 1 then
          #元データにシステムカテゴリが1つとカスタムカテゴリが1つ以上あり、
          #カスタムカテゴリが全てログインユーザのカスタムカテゴリから外れている場合
          @notices << "hostinfo.csvの#{session[:line_num]}行目: ログインユーザのカスタムカテゴリの範囲外になるため、登録されません。"
          next
        end
        is_multiple_system_categories = false
        notice = ""
        if i_categories.size > 1 && s_cat_names.size > 1 then
            is_multiple_system_categories = true
            notice = "hostinfo.csvの#{session[:line_num]}行目: システムカテゴリの登録は1つのみ可能なため、ホストのシステムカテゴリは変更されません。"
        end
        if (i_categories - s_categories).size != 0 or
            (s_categories - i_categories).size != 0 then
          if is_multiple_system_categories then
            notice =  "hostinfo.csvの#{session[:line_num]}行目: システムカテゴリの登録は1つのみ可能なため、システムカテゴリは「#{s_cat_names[0]}」が登録されます。"
          end
          target << {
            :action => 'category_change',
            :action_name => 'カテゴリ変更',
            :old => s_categories,
            :new => i_categories,
          }
        end
        if is_multiple_system_categories then
          @notices << notice
        end
        # ホスト名比較
        if item[:hostname] != t_host.hostname then
          if same_hostname and same_hostname['sidfmd_host'] and same_hostname['sidfmd_host']['host_apikey'] != item[:token] and !same_hostname['flg_delete'] then
            msg = "hostinfo.csvの#{session[:line_num]}行目: 既存の削除されていないホストと同じホスト名は使用できません。"
            @errs << msg
            next 
          else  
            target << {
              :action => 'rename_host',
              :action_name => 'ホスト名変更',
              :old => t_host.hostname,
              :new => item[:hostname],
            }
          end
        end
        # 表示名比較
        old_extname = t_host.ext_name ? t_host.ext_name : ''
        new_extname = item[:ext_name] ? item[:ext_name]  : ''
        if old_extname != new_extname then
          target << {
            :action => 'rename_extname',
            :action_name => '表示名変更',
            :old => old_extname,
            :new => new_extname,
          }
        end
        # 基準日比較
        i_refdate = item[:refdate].nil? ? '' : item[:refdate]
        s_refdate = t_host.refdate
        if s_refdate then
          s_refdate = s_refdate.strftime('%Y/%m/%d')
        else
          s_refdate = ''
        end
        ev_i_refdate = ''
        if i_refdate.size > 0 then
          begin
            ev_i_refdate = Date.parse(i_refdate)
            if ev_i_refdate then
              ev_i_refdate = ev_i_refdate.strftime('%Y/%m/%d')
            end
          rescue Exception => e
            log_exception(e)
          end
        end
        if ev_i_refdate != s_refdate then
          target << {
            :action =>'refdate_change',
            :action_name => '基準日変更',
            :old => s_refdate,
            :new => i_refdate,
          }
        end
        # メモ比較
        i_memo = item[:memo]
        s_memo = t_host.memo
        if (item[:memo] ? item[:memo] : '') != (t_host.memo ? t_host.memo : '') then
          if i_memo and i_memo.size > SidfmdHost::MEMO_MAX_LEN then
            @errs << "hostinfo.csvの#{session[:line_num]}行目:「メモ」の長さが最大値の#{SidfmdHost::MEMO_MAX_LEN}文字を超えています。"
            next
          end
          target << {
            :action => 'memo_change',
            :action_name => 'メモ変更',
            :old => s_memo,
            :new => i_memo,
          }
        end

        host_opts = host_opts_info[t_host.id.to_s]
        if item[:is_agent].downcase == 'false' && t_host.is_agent then
           @notices << "hostinfo.csvの#{session[:line_num]}行目: 「Agentフラグ」列が \"FALSE \"のため、現在登録されている「queue_packages」「queue_apps」「queue_dist」の情報は、次の脆弱性同期のタイミングでホストの情報として反映されなくなります。"
        end
        if host_opts then
          host_opts = host_opts.symbolize_keys
          if item[:is_agent].downcase == 'false' && ((host_opts[:packages] && host_opts[:packages].size > 0) || (host_opts[:apps] && host_opts[:apps].size > 0) || (host_opts[:dist] && host_opts[:dist].size > 0)) then
            @notices << "hostinfo.csvの#{session[:line_num]}行目:「Agentフラグ」列が \"FALSE \"のホストには、optsフォルダ下の「packages」「queue_apps」「queue_dist」のデータは登録できません。※脆弱性同期時に自動生成されます。"
            new_pkgs = []
            new_apps = []
            new_dist = ''
          else
            new_pkgs = host_opts[:packages].uniq
            new_apps = host_opts[:queue_apps].uniq
            new_dist = host_opts[:queue_dist] ? host_opts[:queue_dist] : ''
          end
          new_products = host_opts[:manual_products].uniq
          new_lproducts = host_opts[:local_products].uniq
        else
          new_pkgs = []
          new_apps = []
          new_products = []
          new_lproducts = []
          new_dist = ''
        end
        if item[:is_agent].downcase == 'false' && t_host.is_agent then
          is_os_prod = false
          os_pids = []
          SidfmdProduct.where(:category => 'OS').each do |p|
            if new_products.include?(p.name) then
              is_os_prod = true
              break
            end
            os_pids << p.id
          end
          SidfmdVersion.where(:sidfmd_product_id => os_pids).each do |v|
            if new_products.include?(v.fullname) then
              is_os_prod = true
              break
            end
          end
          unless is_os_prod then
            @notices << "hostinfo.csvの#{session[:line_num]}行目: OSのプロダクトが登録されないため、ホストのOSは出力されません。"
          end
        end
        if t_host.is_agent == false && item[:is_agent].downcase == 'true' && new_pkgs.size == 0 && new_apps.size == 0 && new_dist.size == 0 then
          @notices << "hostinfo.csvの#{session[:line_num]}行目: 「queue_packages」「queue_apps」「queue_dist」の情報が入力されておりません。"
        end
        queue = t_host.get_last_queue
        if queue && item[:is_agent].downcase == 'true'then
          # パッケージ比較
          s_pkgs = queue.sidfmd_queue_pkgs.map{|pkg| pkg.pkg}
          if (new_pkgs - s_pkgs).size != 0 ||
             (s_pkgs - new_pkgs).size != 0
          then
            target << {
              :action => 'packages_change',
              :action_name => 'パッケージ変更',
              :old => s_pkgs,
              :new => new_pkgs,
            }
          end
          # 最新同期アプリケーション比較
          s_apps = queue.sidfmd_queue_apps.map{|app| "#{app.name}:#{app.version}:#{app.path}"}
          if (new_apps - s_apps).size != 0 ||
             (s_apps - new_apps).size != 0
          then
            target << {
              :action => 'apps_change',
              :action_name => '最新同期アプリケーション変更',
              :old => s_apps,
              :new => new_apps,
            }
          end
          # 最新同期OS情報比較
          dist = queue.sidfmd_queue_dist
          if dist then
            dist = "#{dist.ident}:#{dist.version}:#{dist.name}"
          else
             dist = ''
          end
          if dist != new_dist then
            target << {
              :action => 'dist_change',
              :action_name => '最新同期OS変更',
              :old => dist,
              :new => new_dist,
            }
          end
        else
          #queueが無い場合
          if new_pkgs.size > 0 then
            target << {
              :action => 'packages_change',
              :action_name => 'パッケージ変更',
              :old => [],
              :new => new_pkgs,
            }
          end
          if new_apps.size > 0 then
            target << {
              :action => 'apps_change',
              :action_name => '最新同期アプリケーション変更',
              :old => [],
              :new => new_apps,
            }
          end
          if new_dist.size > 0 then
            target << {
              :action => 'dist_change',
              :action_name => '最新同期OS変更',
              :old => '',
              :new => new_dist,
            }
          end
        end
        # 製品比較
        if (new_products.size + new_lproducts.size) > SidfmdHost::MAX_REGISTER_PRODUCTS then
          @notices << "hostinfo.csvの#{session[:line_num]}行目: プロダクトの最大登録数は#{SidfmdHost::MAX_REGISTER_PRODUCTS}個のため、このホストのプロダクトの変更は行われません。optsフォルダ以下のファイルをご確認下さい。"
        else
          same_token_products = []
          same_token_lproducts = []
          prod_ids = []
          ver_ids = []
          lprod_ids = []
          lver_ids = []
          is_sidfmd_products = false
          t_host.sidfmd_host_apps.each do |app|
            if app.version_id && app.version_id != 0 then
              ver_ids << app.version_id
            else
              prod_ids << app.product_id
            end
          end
          SidfmdVersion.where(:vid => ver_ids).each do |ver|
            same_token_products << ver.fullname
          end
          SidfmdProduct.where(:pid => prod_ids).each do |prd|
            same_token_products << prd.name
          end
          t_host.sidfmd_host_lapps.each do |lapp|
            if lapp.sidfmd_lversion_id == 0 then
              lprod_ids << lapp.sidfmd_lproduct_id
            else
              lver_ids << lapp.sidfmd_lversion_id
            end
          end
          SidfmdLproduct.where(:id => lprod_ids).each do |lprd|
            same_token_lproducts << lprd.name
          end
          SidfmdLversion.where(:id => lver_ids).each do |lver|
            same_token_lproducts << lver.fullname
          end
          same_token_products.compact!
          same_token_products.uniq!
          same_token_lproducts.compact!
          same_token_lproducts.uniq!
          if (new_products - same_token_products).size != 0 or
             (same_token_products - new_products).size != 0 or
             (new_lproducts - same_token_lproducts).size != 0 or
             (same_token_lproducts - new_lproducts).size != 0
          then
            is_sidfmd_products = true if new_products.size > 0 
            target << {
              :action => 'product_change',
              :action_name => '製品変更',
              :old => same_token_products + same_token_lproducts,
              :new => new_products + new_lproducts,
              :new_manual_products => new_products,
              :new_local_products => new_lproducts
            }
          end
        end

        #is_agent比較
        i_is_agent = item[:is_agent].downcase
        s_is_agent = t_host.is_agent.to_s.downcase
        if i_is_agent != s_is_agent then
          if i_is_agent == 'false' && s_is_agent == 'true' && 
            !is_sidfmd_products && t_host.sidfmd_host_apps.size == 0 
          then
            if target.size > 0 then
              @notices << "hostinfo.csvの#{session[:line_num]}行目: Agentフラグを \"TRUE \"から \"FALSE \"にする場合は、ローカルプロダクトではない製品の登録が必要なため、Agentフラグは変更されません。"
            else
              @notices = @notices.reject{|n| n.index("hostinfo.csvの#{session[:line_num]}")}
              @errs << "hostinfo.csvの#{session[:line_num]}行目: Agentフラグを \"TRUE \"から \"FALSE \"にする場合は、ローカルプロダクトではない製品の登録が必要です。"
            end
          else
            target << {
              :action => 'agent_change',
              :action_name => 'Agentフラグ変更',
              :old => s_is_agent,
              :new => i_is_agent
            }
          end
        end

        if target.size > 0 then
          if (item[:flg_delete] && item[:flg_delete].downcase == 'true') then
            #@errs << "hostinfo.csvの#{session[:line_num]}行目: 削除されているホストの設定を変更することはできません。"
            next
          end
          @actions << {
            :action => 'modify_host',
            :action_name => '設定変更',
            :old_data => same_token,
            :new_data => item,
            :options => target,
          }
        end
      end
    end

    session[:import_actions] = session[:import_actions] + @actions
    session[:import_errs] = session[:import_errs] + @errs
    session[:import_notices] = hostname_errs.size > 0 ? hostname_errs : session[:import_notices] + @notices
    session[:is_new_import] = true

    is_change = true
    #変更対象のデータが全てエラーになる場合は、index_rでエラーメッセージ表示
    if session[:import_actions].size == 0 && (session[:import_notices].size > 0 || session[:import_errs].size > 0) && executed_size == 0 then
      #:noticeに入れると、パラメーター制限値を超えてしまう可能性があるため、sessionに保存。
      if !is_err then
        session[:import_notices] = ["変更が無いため、インポートを終了しました。", "※インポートしたファイルにエラーがあります。"] + session[:import_errs] + session[:import_notices]
      end
      is_change = false
    elsif session[:import_actions].size == 0 && executed_size == 0 then
      #変更もエラーもない場合
      if !is_err then
        session[:import_notices] = ['変更が無いため、インポートを終了しました。']
      end
      is_change = false
    elsif executed_size == 0 then
      # ライセンスチェック
      create_undelete_count = @actions.count {|x| [ 'create_new_host', 'undelete_host' ].include?(x[:action]) }
      delete_count = @actions.count {|x| 'remove_host' == x[:action] }
      create_count = create_undelete_count - delete_count
      if create_count > 0 then
        # ライセンス追加がある場合は、ライセンス数の確認を行う。
        if @hosts and @hosts[:license_number] - @hosts[:used_license_number] < create_count then
          if !is_err then
            session[:import_notices] = ['ライセンス数を超えるため、インポートを実行することができませんでした。']
          end
          is_change = false
        end
      end
    end
    #logger.info(session[:import_actions].inspect)
  
    render :layout => false, :json => {:err => is_err, :executed_size => executed_size, :is_change => is_change}
  end

  def new_import_apply
    f_params = params.permit(:action_data, :rule_info, :is_exec_synchro_hosts)
    err_message = ''
    if !f_params[:action_data] || !f_params[:is_exec_synchro_hosts] then
      err_message = 'パラメータが不正です。'
    end
    begin
      action = JSON.parse(f_params[:action_data], :symbolize_names => true)
      rule_info = (f_params[:rule_info] ? JSON.parse(f_params[:rule_info]).symbolize_keys : nil) 
      #rule_info = JSON.parse(f_params[:rule_info] || '{}').symbolize_keys
    rescue JSON::ParserError => e
      log_exception(e)
      err_message = 'パラメータが不正です。'
    end
    is_exec_synchro_hosts = f_params[:is_exec_synchro_hosts].to_s
    if err_message.size > 0 then
      render :layout => false, :json => { :err_message => err_message  } and return
    end
    host_ids_for_view = []
    target_html_ids = []
    is_remove_host = false #remove_hostかmodify_categoryの時に関連するルールをチェックする
    is_mod_cat_host = false
    case action[:action]
    when 'create_new_host' then # 新規登録
      all_categories = {}
      all_system_categories ={}
      SidfmdCategory.all.each do |c|
        all_categories[c.name] = c.id
        all_system_categories[c.id] = c.name if c.is_system
      end
      user_c_cat_ids = (@role_admin ? all_categories.values : @user_category_ids)
      if !@role_admin then
        user_cat_names = SidfmdCategory.where(:id => user_c_cat_ids).where(:is_system => false).pluck(:name)
        if (user_cat_names - action[:new_data][:categories].split(/,/).map{|c| c.strip}).size == user_cat_names.size then
          err_message = "不正な操作のため、「#{action[:new_data][:hostname]}」の登録はされませんでした。"
          render :layout => false, :json => { :err_message => err_message  } and return
        end
      end

      if action[:new_data][:ext_name] and action[:new_data][:ext_name] =~ /&&&&&/ then
        err_message = "不正な操作のため、「#{action[:new_data][:hostname]}」の登録はされませんでした。"
        render :layout => false, :json => { :err_message => err_message  } and return
      end

      # ライセンスチェック
      lmhosts = SidfmdHost.get_hosts_by_lm(:limit => 0)
      if lmhosts and lmhosts[:license_number] <= lmhosts[:used_license_number] then
        logger.info('license number is upper limit')
        err_message = 'ホスト数がライセンスの上限に達しています。'
        render :layout => false, :json => { :err_message => err_message  } and return
      end
      
      is_hostname_correct = check_hostname(action[:new_data][:hostname])
      if is_hostname_correct != true then
        logger.info('hostname is invalid')   
        err_message = "「#{action[:new_data][:hostname]}」のホスト名は使用できないため、「#{action[:new_data][:hostname]}」の登録に失敗しました。"
        render :layout => false, :json => { :err_message => err_message  } and return 
      end

      if !@role_admin && SidfmdHost.find_by(:hostname => action[:new_data][:hostname]) then
        err_message = "「#{action[:new_data][:hostname]}」のホスト名は既に使用されているため、「#{action[:new_data][:hostname]}」の登録に失敗しました。"
        render :layout => false, :json => { :err_message => err_message  } and return
      end

      # ライセンスサーバとVMでホストの登録を行う。
      temp = SidfmdHost.register_host(action[:new_data][:hostname], nil, false)
      if temp.nil? then
        logger.info('failed to register host')
        err_message = "「#{action[:new_data][:hostname]}」の登録に失敗しました。"
        render :layout => false, :json => { :err_message => err_message  } and return
      end
      
    
      # 作成されたホストを取得
      host = SidfmdHost.where(:hostname => action[:new_data][:hostname]).first
      unless host then
        logger.info('failed to register host')
        err_message = "「#{action[:new_data][:hostname]}」の登録に失敗しました。"
        render :layout => false, :json => { :err_message => err_message  } and return
      end
     
      SidfmdHost.transaction do
        begin  
          host.lock!
          #Agentフラグ
          host.update_attribute :is_agent, (action[:new_data][:is_agent].to_s.downcase == 'true' ? true : false)

          #表示名登録
          host.update_attribute :ext_name, action[:new_data][:ext_name]
      
          # カテゴリ登録
          if action[:new_data][:categories] then
            is_s_cat_registered = false
            action[:new_data][:categories].split(/,/).map{|c| c.strip}.uniq.each do |category_name|
              cat_id = all_categories[category_name]
              if cat_id then
                if all_system_categories[cat_id] then
                  if !is_s_cat_registered then
                    #公開サーバがデフォルトで登録されるため、システムカテゴリが２つにならないように、削除してから登録する
                    SidfmdCategoriesHost.where(:sidfmd_host_id => host.id)
                                        .where(:sidfmd_category_id => all_system_categories.keys)
                                        .destroy_all
                    SidfmdCategoriesHost.create(:sidfmd_host_id => host.id,:sidfmd_category_id => cat_id)
                    is_s_cat_registered = true
                  end
                else
                  if user_c_cat_ids.include?(cat_id) then
                    SidfmdCategoriesHost.create(:sidfmd_host_id => host.id,:sidfmd_category_id => cat_id)
                  end
                end
              end
            end
          end
         
          #基準日登録
          refdate = action[:new_data][:refdate]
          if refdate then
            begin
              refdate = Date.parse(refdate)
            rescue ArgumentError => e
              log_exception(e)
              raise Exception
            end
          else
            refdate = Date.today
          end
          host.update_attribute :refdate, refdate
          # メモ登録
          if action[:new_data][:memo] and action[:new_data][:memo].size > SidfmdHost::MEMO_MAX_LEN then
            logger.info('memo is over max length')
            raise Exception
          end
          host.update_attribute :memo, action[:new_data][:memo]

          #削除フラグ
          #host.update_attribute :flg_delete, (action[:new_data][:flg_delete].to_s.downcase == 'true' ? true : false)

          host.save
        rescue Exception => e
          log_exception(e)
          err_message = "「#{host.hostname}」に関連する情報の登録に失敗しました。※ホストの登録は正常に行われました。"  
          Rails.cache.clear
          render :layout => false, :json => { :err_message => err_message  } and return
        end
      end
      #パッケージ・最新同期アプリケーション・最新同期OS情報登録
      regist_queue_info = regist_host_queue_opts(host.id, action[:new_data][:packages], action[:new_data][:apps], action[:new_data][:dist])
      if !regist_queue_info then
        err_message = "「#{host.hostname}」の付属情報の登録に失敗しました。※ホストの登録は正常に行われました。"
        Rails.cache.clear
        render :layout => false, :json => { :err_message => err_message  } and return
      end
      #プロダクト登録
      host.sidfmd_host_apps = []
      host.sidfmd_host_lapps = []
      host.save
      manual_prds = action[:new_data][:manual_products] ? action[:new_data][:manual_products] : []
      local_prds = action[:new_data][:local_products] ? action[:new_data][:local_products] : []
      if (manual_prds.size > 0 || local_prds.size > 0) && (manual_prds.size + local_prds.size) <= SidfmdHost::MAX_REGISTER_PRODUCTS then
        begin 
          regist_products(manual_prds, local_prds, host.id)
        rescue Exception => e
          log_exception(e)
          err_message = "「#{host.hostname}」のプロダクトの登録に失敗しました。※ホストの登録は正常に行われました。"
          Rails.cache.clear
          render :layout => false, :json => { :err_message => err_message  } and return
        end
      end
      SidfmdHostView.sync_data([host.id])
    when 'undelete_host' then # 復元（削除状態のものを復元する）
      host_id = action[:old_data][:id].to_i
      hostname = action[:old_data][:hostname]
      sidfmd_hostid = (action[:old_data][:sidfmd_host] ? action[:old_data][:sidfmd_host][:id] : nil)
      host = nil
      if !@role_admin && hostname then
        err_message = "「#{hostname}」の復元は、ログインユーザでは許可されていない操作です。"
        render :layout => false, :json => { :err_message => err_message  } and return
      end
      if action[:new_data][:ext_name] and action[:new_data][:ext_name] =~ /&&&&&/ then
        err_message = "不正な操作のため、「#{action[:new_data][:hostname]}」の登録はされませんでした。"
        render :layout => false, :json => { :err_message => err_message  } and return
      end
      if host_id > 0 and hostname and hostname.size > 0 and sidfmd_hostid then
        SidfmdHost.transaction do
          all_categories = {}
          all_system_categories ={}
          SidfmdCategory.all.each do |c|
            all_categories[c.name] = c.id
            all_system_categories[c.id] = c.name if c.is_system
          end
          res = SidfmdHost.toggle_host_delete(host_id)
          if res then
            if res[:success] then
              host = SidfmdHost.where(:id => sidfmd_hostid ).first
              if host then
                host.lock!
                host.update_attribute :flg_delete, false
              else
                res = SidfmdHost.establish_consistency
              end
            else
              err_message = "ライセンス数を超過するため、「#{hostname}」の復元に失敗しました。"
              render :layout => false, :json => { :err_message => err_message  } and return
            end
          else
            err_message = "ライセンスサーバとの通信に失敗したため、「#{hostname}」の復元に失敗しました。"
            render :layout => false, :json => { :err_message => err_message  } and return
          end
        end
      else
        err_message = "「#{hostname}」の復元に失敗しました。"
        render :layout => false, :json => { :err_message => err_message  } and return
      end
      # 復元されたホストが取得できなかった場合。
      unless host then
        host = SidfmdHost.where(:hostname => hostname).first
        unless host then
          err_message = "「#{hostname}」の復元に失敗しました。"
          render :layout => false, :json => { :err_message => err_message  } and return
        end
      end
      begin
        SidfmdHost.transaction do
          host.lock!
          #表示名更新
          host.update_attribute :ext_name, action[:new_data][:ext_name]
          # カテゴリ更新
          cat_names = action[:new_data][:categories]
          if cat_names and cat_names.size > 0 then
            c_category_ids = []
            system_category_id = nil
            cat_names.split(/,/).map{|c| c.strip}.uniq.each do |c_name|
              cat_id = all_categories[c_name]
              if cat_id then
                if all_system_categories[cat_id] then
                  if !system_category_id then
                    system_category_id = cat_id
                  end
                else
                  c_category_ids << cat_id
                end
              end
            end
            if system_category_id then
              res = host.modify_categories(system_category_id, c_category_ids)
              unless res then
                logger.info('SidfmdHost#modify_categories failure')
                raise Exception
                #raise StandardError, 'SidfmdHost#modify_categories failure'
              end
            end
          else
            SidfmdCategoriesHost.destroy_all(:sidfmd_host_id => host.id)
          end
          #is_agent更新
          host.update_attribute :is_agent, (action[:new_data][:is_agent].to_s.downcase == 'true' ? true : false)

          # 基準日更新
          refdate = action[:new_data][:refdate]
          if refdate then
            begin
              refdate = Date.parse(refdate)
            rescue ArgumentError => e
              log_exception(e)
              logger.info('ignore invalid refdate:' + e.inspect)
               refdate = nil
            end
          end
          host.update_attribute :refdate, refdate
          # メモ更新
          memo = action[:new_data][:memo]
          if memo and memo.size > SidfmdHost::MEMO_MAX_LEN then
            logger.info('memo is over max length')
            raise Exception
          elsif memo then
            host.update_attribute :memo, memo
          end
        end
        SidfmdHostView.sync_data([host.id])
      rescue Exception => e
        log_exception(e)
        Rails.cache.clear
        err_message = "「#{hostname}」に関連する情報の登録に失敗しました。※ホストの復元は正常に行われました。"
        render :layout => false, :json => { :err_message => err_message  } and return
      end
      #パッケージ・最新同期アプリケーション・最新同期OS情報登録
      regist_queue_info = regist_host_queue_opts(host.id, action[:new_data][:packages], action[:new_data][:apps], action[:new_data][:dist])
      if !regist_queue_info then
        Rails.cache.clear
        err_message = "「#{host.hostname}」の付属情報の登録に失敗しました。※ホストの登録は正常に行われました。"
        render :layout => false, :json => { :err_message => err_message  } and return
      end
      #プロダクト更新
      manual_prds = action[:new_data][:manual_products] ? action[:new_data][:manual_products] : []
      local_prds = action[:new_data][:local_products] ? action[:new_data][:local_products] : []
      if (manual_prds.size > 0 || local_prds.size > 0) && (manual_prds.size + local_prds.size) <= SidfmdHost::MAX_REGISTER_PRODUCTS then
        host.sidfmd_host_apps = []
        host.sidfmd_host_lapps = []
        host.save
        begin
          regist_products(manual_prds, local_prds, host.id)
        rescue Exception => e
          log_exception(e)
          Rails.cache.clear
          err_message = "「#{host.hostname}」のプロダクトの登録に失敗しました。※ホストの復元は正常に行われました。"
          render :layout => false, :json => { :err_message => err_message  } and return
        end
      else
        SidfmdHostApp.destroy_all(:sidfmd_host_id => host.id)
        SidfmdHostLapp.destroy_all(:sidfmd_host_id => host.id)
      end 
    when 'remove_host' then
      is_remove_host = true
      # ライセンスサーバとVMでホストの削除を行う。
      action[:old_data].each do |i|
        host_id = i[:host][:id]
        hostname = i[:host][:hostname]
        sidfmdhostid = i[:host][:sidfmd_host] ? i[:host][:sidfmd_host][:id] : nil
        if !@role_admin && hostname then
          if sidfmdhostid then
            t_host = SidfmdHost.find_by(:id => sidfmdhostid)
            if t_host then
              host_cat_ids = t_host.sidfmd_categories.pluck(:id)
              if (@user_category_ids - host_cat_ids).size == @user_category_ids.size then
                err_message = "ログインユーザでの操作は許可されていないため、「#{host.hostname}」の削除に失敗しました。"
                render :layout => false, :json => { :err_message => err_message  } and return
              end
            else
              err_message = "「#{host.hostname}」の削除に失敗しました。"
              render :layout => false, :json => { :err_message => err_message  } and return
            end
          else
            err_message = "ログインユーザでの操作は許可されていないため、「#{host.hostname}」の削除に失敗しました。"
            render :layout => false, :json => { :err_message => err_message  } and return
          end
        end
        if host_id and hostname and sidfmdhostid then
          SidfmdHost.transaction do
            res = SidfmdHost.toggle_host_delete(host_id)
            if res then
              if res[:success] then
                host = nil
                if sidfmdhostid > 0 then
                  host = SidfmdHost.find_by(:id => sidfmdhostid)
                else
                  host = SidfmdHost.where(:hostname => hostname).first
                end
                if host then
                  t_html_ids = SidfmdHostView.relate_htmlids([ host.id ])
                  host.lock!
                  host.update_attribute :flg_delete, true
                  host.last_checked_at = nil
                  host.memo = nil
                  host.is_agent = false
                  host.save

                  SidfmdCategoriesHost.where(:sidfmd_host_id => sidfmdhostid).delete_all
                  SidfmdHostApp.where(:sidfmd_host_id => sidfmdhostid).delete_all
                  SidfmdHostLapp.where(:sidfmd_host_id => sidfmdhostid).delete_all

                  sidfmd_queue_ids = SidfmdQueue.where(:sidfmd_host_id => sidfmdhostid).collect {|x| x.id }.uniq
                  sidfmd_queue_ids.each do |sidfmd_queue_id|
                    SidfmdQueueDist.where(:sidfmd_queue_id => sidfmd_queue_id).delete_all
                    SidfmdQueuePkg.where(:sidfmd_queue_id => sidfmd_queue_id).delete_all
                    SidfmdQueueApp.where(:sidfmd_queue_id => sidfmd_queue_id).delete_all
                    SidfmdQueueLapp.where(:sidfmd_queue_id => sidfmd_queue_id).delete_all
                    SidfmdQueueWin.where(:sidfmd_queue_id => sidfmd_queue_id).delete_all
                    sidfmd_queue_vul_ids = SidfmdQueueVul.where(:sidfmd_queue_id => sidfmd_queue_id).collect {|x| x.id }.uniq
                    sidfmd_queue_vul_ids.each do |sidfmd_queue_vul_id|
                      SidfmdQueueVulCve.where(:sidfmd_queue_vul_id => sidfmd_queue_vul_id).delete_all
                      SidfmdQueueVulSri.where(:sidfmd_queue_vul_id => sidfmd_queue_vul_id).delete_all
                      SidfmdQueueVulAffprod.where(:sidfmd_queue_vul_id => sidfmd_queue_vul_id).delete_all
                      SidfmdQueueVulAffver.where(:sidfmd_queue_vul_id => sidfmd_queue_vul_id).delete_all
                      SidfmdQueueVulAffpkg.where(:sidfmd_queue_vul_id => sidfmd_queue_vul_id).delete_all
                    end
                    SidfmdQueueVul.where(:sidfmd_queue_id => sidfmd_queue_id).delete_all
                  end
                  SidfmdQueue.where(:id => sidfmd_queue_ids).delete_all

                  sidfmd_issue_ids = SidfmdIssue.where(:sidfmd_host_id => sidfmdhostid).collect {|x| x.id }.uniq
                  sidfmd_issue_ids.each do |sidfmd_issue_id|
                    SidfmdIssueCve.where(:sidfmd_issue_id => sidfmd_issue_id).delete_all
                    SidfmdIssueSri.where(:sidfmd_issue_id => sidfmd_issue_id).delete_all
                    SidfmdIssueAffprod.where(:sidfmd_issue_id => sidfmd_issue_id).delete_all
                    SidfmdIssueAffver.where(:sidfmd_issue_id => sidfmd_issue_id).delete_all
                    SidfmdIssueAffpkg.where(:sidfmd_issue_id => sidfmd_issue_id).delete_all
                  end
                  SidfmdIssue.where(:id => sidfmd_issue_ids).delete_all
                  SidfmdHostView.sync_data([host.id])
                  SidfmdVulView.sync_data(t_html_ids)
                else
                  res = SidfmdHost.establish_consistency
                end
              else
                err_message = "「#{host.hostname}」の削除に失敗しました。"
                render :layout => false, :json => { :err_message => err_message  } and return
              end
            else
              err_message = 'ライセンスサーバとの通信に失敗しました。' 
              render :layout => false, :json => { :err_message => err_message  } and return
            end
          end
        end
      end
    when 'modify_host' then
      host_id = action[:old_data][:id]
      hostname = action[:old_data][:hostname]
      sidfmd_hostid = (action[:old_data][:sidfmd_host] ? action[:old_data][:sidfmd_host][:id] : nil)
      unless sidfmd_hostid then
        err_message = "「#{hostname}」の登録情報の変更に失敗しました。"
      else
        host = SidfmdHost.where(:id => sidfmd_hostid).first
        if host then
          if !@role_admin then
            if (@user_category_ids - host.sidfmd_categories.pluck(:id)).size == @user_category_ids.size then
              err_message = "ログインユーザでの操作は許可されていないホストのため、「#{hostname}」の登録情報の変更に失敗しました。"
              render :layout => false, :json => { :err_message => err_message  } and return
            end
          end
          host_ids_for_view << host.id
          host_queue_data = {}
          option_err_messages = []
          action[:options].each do |option|
            case option[:action]
            when 'agent_change' then
              SidfmdHost.transaction do
                begin 
                  host.lock!
                  host.update_attribute :is_agent, (option[:new].to_s.downcase == 'true' ? true : false)
                rescue Exception => e
                  log_exception(e)
                  err_message = "「#{hostname}」のAgentフラグの変更に失敗しました。"
                  option_err_messages << err_message
                end
              end
            when 'rename_host' then
              new_name = option[:new]
              SidfmdHost.transaction do
                begin 
                  res = SidfmdHost.rename_host(host_id, new_name)
                  if res.respond_to?(:[]) then
                    if res[:success] then
                       host.lock!
                       host.update_attribute :hostname, new_name
                    else
                      err_message = "登録済のホスト名、もしくは不正な文字を指定したため、「#{hostname}」のホスト名変更に失敗しました。"
                      option_err_messages << err_message
                    end
                  else
                    err_message = "ライセンスサーバとの通信に失敗したため、「#{hostname}」のホスト名変更は行われませんでした。"
                    option_err_messages << err_message
                  end
                rescue Exception => e
                  log_exception(e)
                  err_message = "「#{host.hostname}」のホスト名変更に失敗しました。"
                  option_err_messages << err_message
                end
              end
            when 'rename_extname' then
              SidfmdHost.transaction do
                begin
                  host.lock!
                  new_extname = option[:new]
                  if new_extname =~ /&&&&&/ then
                    raise Exception
                  end
                  host.update_attribute :ext_name, (new_extname.size > 0 ? new_extname : nil)
                rescue Exception => e
                  log_exception(e)
                  err_message = "「#{host.hostname}」の表示名変更に失敗しました。"
                  option_err_messages << err_message
                end
              end
            when 'packages_change', 'apps_change', 'dist_change' then
              #後でまとめて処理
              if host then
                host_queue_data[option[:action].to_sym] = option[:new]
              end  
            when 'product_change' then
              # 製品変更
              SidfmdHost.transaction do
                begin
                  host.lock!
                  host.sidfmd_host_apps = []
                  host.sidfmd_host_lapps = []
                  host.save
                  manual_prds = option[:new_manual_products] 
                  local_prds = option[:new_local_products]
                  if (manual_prds.size > 0 || local_prds.size > 0) && (manual_prds.size + local_prds.size) <= SidfmdHost::MAX_REGISTER_PRODUCTS then
                    regist_products(manual_prds, local_prds, host.id)
                  end
                rescue Exception => e
                  log_exception(e)
                  err_message = "「#{host.hostname}」のプロダクト変更に失敗しました。"
                  option_err_messages << err_message
                end
              end
            when 'category_change' then 
              is_mod_cat_host = true
              # カテゴリ変更
              host = SidfmdHost.where(:hostname => hostname).first
              if host then
                old_categories = option[:old].map{|c_name| c_name.strip}.uniq
                new_categories = option[:new].map{|c_name| c_name.strip}.uniq
                remove_categories = old_categories - new_categories
                add_categories = new_categories - old_categories
                old_s_cat = SidfmdCategory.system.where(:name => old_categories).first
                new_s_cat = SidfmdCategory.system.where(:name => new_categories).first
                SidfmdCategoriesHost.transaction do
                  begin 
                    cids = []
                    remove_categories.each do |category_name|
                      cat = SidfmdCategory.where(:name => category_name).first
                      if cat then
                        cids << cat.id
                      end
                    end
                    if cids.size > 0 then
                      SidfmdCategoriesHost.where(:sidfmd_host_id => host.id,
                                                 :sidfmd_category_id => cids).destroy_all
                    end
                      add_categories.each do |category_name|
                      cat = SidfmdCategory.where(:name => category_name).first
                      if cat then
                        SidfmdCategoriesHost.create(:sidfmd_host_id => host.id,
                                                    :sidfmd_category_id => cat.id)
                      end
                    end
                  rescue Exception => e
                    log_exception(e)
                    err_message = "「#{host.hostname}」のカテゴリ変更に失敗しました。"
                    option_err_messages << err_message
                  end
                end
              else
                logger.info('ignore change categories: old_s_cat=' + old_s_cat.inspect + ', new_s_cat=' + new_s_cat.inspect)
                err_message = "「#{host.hostname}」のカテゴリ変更に失敗しました。"  
                option_err_messages << err_message
              end
            when 'refdate_change' then
              # 基準日変更
              SidfmdHost.transaction do
                begin
                  host.lock!
                  if option[:new] and option[:new].size > 0 then
                    host.update_attribute :refdate, Date.parse(option[:new])
                  else
                    host.update_attribute :refdate, nil
                  end
                rescue ArgumentError => e
                  log_exception(e)
                  logger.info('ignore change refdate: new_refdate=' + option[:new].inspect)
                  err_message = "「#{host.hostname}」の基準日変更に失敗しました。"
                  option_err_messages << err_message
                end
              end
            when 'memo_change' then 
              # メモ変更
              SidfmdHost.transaction do
                begin 
                  host.lock!
                  if option[:new] and option[:new].size > 0 then
                    if option[:new].size > SidfmdHost::MEMO_MAX_LEN then
                      logger.info("ignore change memo(over length): new_memo=#{option[:new].inspect}")
                      err_message = "「#{host.hostname}」のメモ変更に失敗しました。"
                      option_err_messages << err_message
                    else
                      host.update_attribute :memo, option[:new]
                    end
                  else
                    host.update_attribute :memo, nil
                  end
                rescue Exception => e
                  log_exception(e)
                  err_message = "「#{host.hostname}」のメモ変更に失敗しました。"
                  option_err_messages << err_message
                end
              end
            end
          end
          packages_data = host_queue_data[:packages_change]
          apps_data = host_queue_data[:apps_change]
          dist_data = host_queue_data[:dist_change]
          if packages_data || apps_data || dist_data then
            queue = host.get_last_queue
            if queue then
              if !packages_data then
                packages_data = SidfmdQueuePkg.where(:sidfmd_queue_id => queue.id).pluck(:pkg)
              end
              if !apps_data then
                apps_data = SidfmdQueueApp.where(:sidfmd_queue_id => queue.id).map{|app| "#{app.name}:#{app.version}:#{app.path}"}
              end
              if !dist_data then
                dist = SidfmdQueueDist.where(:sidfmd_queue_id => queue.id).first
                if dist then
                  dist_data = "#{dist.ident}:#{dist.version}:#{dist.name}"
                end
              end
              regist_queue_info = regist_host_queue_opts(host.id, packages_data, apps_data, dist_data)
              if !regist_queue_info then
                err_message = "「#{hostname}」の「packages」「queue_apps」「queue_dist」の変更に失敗しました。"
                option_err_messages << err_message
              end
            else
              regist_queue_info = regist_host_queue_opts(host.id, packages_data, apps_data, dist_data)
              if !regist_queue_info then
                err_message = "「#{hostname}」の「packages」「queue_apps」「queue_dist」の変更に失敗しました。"
                option_err_messages << err_message
              end
            end
          else
            if action[:old_data][:sidfmd_host] && 
               action[:old_data][:sidfmd_host][:is_agent].to_s.downcase == 'false' && 
               action[:new_data][:is_agent] && action[:new_data][:is_agent].to_s.downcase == 'true'
            then
              SidfmdQueue.create(sidfmd_host_id: action[:old_data][:sidfmd_host][:id].to_i) 
            end
          end
          if is_exec_synchro_hosts == 'true' then
            SidfmdHost.transaction do
              begin
                host.lock!
                host.update_attribute :last_checked_at, nil
              rescue Exception => e
                log_exception(e)
              end
            end
          end
          SidfmdHostView.sync_data([host.id]) if host.sidfmd_categories.where(:is_system => true).any? && host.last_checked_at
          err_message = option_err_messages.join("\n")
        else
          err_message = "「#{hostname}」の登録情報の変更に失敗しました。"
        end
      end
    end
    if (is_remove_host || is_mod_cat_host) && rule_info then
      delete_rule_ids = rule_info[:delete_rule_ids]
      delete_cond_ids = rule_info[:delete_cond_ids]
      delete_host_cond_ids = rule_info[:delete_host_cond_ids]
      if delete_rule_ids && delete_rule_ids.size > 0 then
        SidfmdRule.where(:id => delete_rule_ids).destroy_all
      end
      if delete_cond_ids && delete_cond_ids.size > 0 then
        SidfmdRuleCond.where(:id => delete_cond_ids).destroy_all
      end
      if delete_host_cond_ids && delete_host_cond_ids.size > 0 then
        SidfmdRuleCmHostSub.where(:id => delete_host_cond_ids).destroy_all
      end
    end
    if host_ids_for_view.any? then
      host_ids_for_view.uniq!
      target_html_ids |= SidfmdHostView.relate_htmlids(host_ids_for_view)
      SidfmdHostView.sync_data(host_ids_for_view)
      SidfmdVulView.sync_data(target_html_ids)
    end
    Rails.cache.clear
    render :layout => false, :json => { :err_message => err_message }
  end

  CSV_BOM = "\xff\xfe".force_encoding('UTF-16LE').freeze

  def new_export
    f_params = params.permit(:checked, :process, :executed_size)
    process = f_params[:process]
    err = false
    message = ''
    r = {}
    if process == 'start_export' then
      begin
        checked = JSON.parse(f_params[:checked])
      rescue JSON::ParserError => e
        log_exception(e)
        err = true
        message = '不正な操作です。'
      end
      hosts = nil
      begin
        hosts = SidfmdHost.get_hosts_by_lm(:'sort[hostname]' => :asc)
      rescue Exception => e
        log_exception(e)
        err = true
        message = 'ライセンスサーバへのアクセスに失敗しました。'
      end
      unless hosts then
        err = true
        message = 'データ取得に失敗しました。'
      end
      if hosts and hosts[:success] then
        hosts = hosts[:hosts]
        if hosts.nil? then
          err = true
          message = 'データ取得に失敗しました。'
        end
      end
      if checked.size > 0 then
        hosts = hosts.select{|d| checked.include?(d[:hostname])}
      end
      clear_export_session
      session[:export_hosts] = hosts
      r[:total_hosts] = hosts.size
      r[:err] = err
      r[:message] = message
      render :layout => false, :json => r
    elsif process == 'progress_export' then
      hosts = session[:export_hosts] 
      if !hosts then
        err = true
        message = 'データ取得に失敗しました。'
      else
        begin
          executed_size = f_params[:executed_size].to_i
          t_hosts = hosts.slice(executed_size, 50)
          t_hosts = [] unless t_hosts
          executed_hosts = t_hosts.size
          session[:export_hosts_lines] = [] unless session[:export_hosts_lines]
          session[:export_opts_lines] = {} unless session[:export_opts_lines]
          t_hosts.each do |host|
            sidfmd_host_id = (host['sidfmd_host'] ? host['sidfmd_host']['id'] : nil)
            if sidfmd_host_id then
              sidfmd_host = SidfmdHost.find_by(:id => sidfmd_host_id)
              if sidfmd_host then
                if !@role_admin && ((@user_category_ids - sidfmd_host.sidfmd_categories.pluck(:id)).size == @user_category_ids.size || sidfmd_host.flg_delete) then
                  sidfmd_host = nil
                end
              end
            else
              sidfmd_host = nil
            end
            csv_data = create_new_export_hosts_info(host, sidfmd_host)
            session[:export_hosts_lines] << csv_data if csv_data.size > 0
            if sidfmd_host then
              opts = []
              if sidfmd_host.is_agent then
                pkgs_info = create_host_pkgs_info_for_new_export(sidfmd_host)
                pkgs_info.each do |pkg_info|
                  opts << pkg_info
                end
              end
              manual_prods_info = create_host_manual_products_info_for_new_export(sidfmd_host)
              manual_prods_info.each do |prod_info|
                opts << prod_info
              end
              local_prods_info = create_host_local_products_info_for_new_export(sidfmd_host)
              local_prods_info.each do |lprod_info|
                opts << lprod_info
              end
              if sidfmd_host.is_agent then
                apps_info = create_host_apps_info_for_new_export(sidfmd_host)
                apps_info.each do |app_info|
                  opts << app_info
                end
                os_info = create_host_os_info_for_new_export(sidfmd_host)
                opts << os_info if os_info.size > 0
              end
              session[:export_opts_lines][sidfmd_host.id] = opts
            end
          end  
          r[:executed_hosts] = executed_hosts
        rescue Exception => e
          log_exception(e)
          err = true
          message = 'データの処理に失敗しました。'
        end
      end
      r[:err] = err
      r[:message] = message
      render :layout => false, :json => r
    else
      hostinfo_header = [
        'ホスト登録ID'.encode('UTF-16LE'),
        'ホスト削除'.encode('UTF-16LE'),
        'ホストID'.encode('UTF-16LE'),
        'トークン'.encode('UTF-16LE'),
        'ホスト名'.encode('UTF-16LE'),
        '表示名'.encode('UTF-16LE'),
        'カテゴリ'.encode('UTF-16LE'),
        '基準日'.encode('UTF-16LE'),
        '削除'.encode('UTF-16LE'),
        'Agentフラグ'.encode('UTF-16LE'),
        'メモ'.encode('UTF-16LE')
      ]
      opt_header = [
        'ホストID'.encode('UTF-16LE'),
        '種類'.encode('UTF-16LE'),
        'データ'.encode('UTF-16LE')
      ]
      new_host_opt_header = [
        'ホスト登録ID'.encode('UTF-16LE'),
        '種類'.encode('UTF-16LE'),
        'データ'.encode('UTF-16LE')
      ]
      sorted_opts_host_ids = session[:export_opts_lines].keys.sort() 
      Tempfile.create('temp') do |f|
        path = f.path
        Zip::File.open(path, Zip::File::CREATE) do |zipfile|
          zipfile.get_output_stream("hostinfo.csv") do |f|
            f.puts(
              CSV.generate("#{CSV_BOM}", :col_sep => "\t".encode('UTF-16LE')) do |o|
                o << hostinfo_header
                session[:export_hosts_lines].each do |line|
                  o << line.map{|item| item.encode('UTF-16LE')}
                end
              end
            )
          end
          sorted_opts_host_ids.each do |host_id|
            zipfile.get_output_stream("regist/1.csv") do |f|
              f.puts(
                CSV.generate("#{CSV_BOM}", :col_sep => "\t".encode('UTF-16LE')) do |o|
                  o << new_host_opt_header  #新規ホストの付属情報登録用のファイル
                end
              )
            end
            zipfile.get_output_stream("opt/#{host_id}.csv") do |f|
              f.puts(
                CSV.generate("#{CSV_BOM}", :col_sep => "\t".encode('UTF-16LE')) do |o|
                  o << opt_header
                  host_opts_lines = session[:export_opts_lines][host_id]
                  host_opts_lines.each do |line|
                    o << line.map{|item| item.encode('UTF-16LE')}
                  end
                end
              )
            end
          end
        end
        send_data(File.read(path), :filename => "vm-export-#{Time.now.strftime('%Y%m%d%H%M%S')}.zip", :disposition => 'attachment')
        cookies['export_flg'] = true
        clear_export_session
      end
    end
  end

  #プロダクトマスタ情報のダウンロードを行う
  def download_products
    begin
      header = [
        '種類'.encode('UTF-16LE'),
        'サポート終了'.encode('UTF-16LE'),
        'プロダクトID'.encode('UTF-16LE'),
        'バージョンID'.encode('UTF-16LE'),
        'プロダクト名'.encode('UTF-16LE'),
        'バージョン名'.encode('UTF-16LE')
      ]
      csv = CSV.generate("#{CSV_BOM}", :col_sep => "\t".encode('UTF-16LE')) do |o|
        o << header
        SidfmdProduct.order(:name).each do |prd|
          pid = prd.pid
          name = prd.name
          o << [
            'sidfm'.encode('UTF-16LE'),
            prd.is_support_end.to_s.upcase.encode('UTF-16LE'),
            pid.to_s.encode('UTF-16LE'),
            0.to_s.encode('UTF-16LE'),
            name.encode('UTF-16LE'),
            ''.encode('UTF-16LE')
          ]
          SidfmdVersion.where(:sidfmd_product_id => prd.id).order(:version).each do |ver|
            o << [
              'sidfm'.encode('UTF-16LE'),
               ver.is_support_end.to_s.upcase.encode('UTF-16LE'),
               pid.to_s.encode('UTF-16LE'),
               ver.vid.to_s.encode('UTF-16LE'), 
               name.encode('UTF-16LE'),
               ver.fullname.encode('UTF-16LE')
            ]
          end
        end
        SidfmdLproduct.order(:priority).each do |lprd|
          lpid = lprd.id
          name = lprd.name
          o << [
            'local'.encode('UTF-16LE'),
            'FALSE'.encode('UTF-16LE'),
            lpid.to_s.encode('UTF-16LE'),
            0.to_s.encode('UTF-16LE'),
            name.encode('UTF-16LE'),
            ''.to_s.encode('UTF-16LE')
          ]
          SidfmdLversion.where(:sidfmd_lproduct_id => lpid).order(:priority).each do |lver|
            o << [
              'local'.encode('UTF-16LE'),
              'FALSE'.encode('UTF-16LE'),
              lpid.to_s.encode('UTF-16LE'),
              lver.id.to_s.encode('UTF-16LE'),
              name.encode('UTF-16LE'),
              lver.fullname.encode('UTF-16LE')
            ]
          end
        end
      end
    rescue Exception => e
      logger.info(e)
      redirect_to :action => :index_r, :notice => 'データ取得に失敗しました。' and return
    end
    send_data(csv, :filename => 'vm-products-master.csv', :disposition => 'attachment')
  end

  ### 以下、内部API

  # アプリケーション名の一部からアプリケーション情報を返す。
  def api_substr_search
    # params[keyword]
    f_params = params.fetch(:params, {}).permit(:keyword)
    keyword = f_params[:keyword]
    ret = {}
    if keyword and keyword.size > 1 then
      products = SidfmdProduct
        .where('LOWER(name) LIKE ? OR LOWER(name) LIKE ?',
               ActiveRecord::Base.send(:sanitize_sql_like, keyword.downcase) + '%',
               '% ' + ActiveRecord::Base.send(:sanitize_sql_like, keyword.downcase) + '%')
        .order(:category)
      products.each do |product|
        ret[product.pid] = { 'product' => product }
      end
      versions = SidfmdVersion
        .where.not(:version => 'ALL')
        .where('LOWER(fullname) LIKE ? OR LOWER(fullname) LIKE ?',
               ActiveRecord::Base.send(:sanitize_sql_like, keyword.downcase) + '%',
               '% ' + ActiveRecord::Base.send(:sanitize_sql_like, keyword.downcase) + '%')
        .order(:position => :desc)
      versions.each do |version|
        product = version.sidfmd_product
        if !ret.has_key?(product.pid) then
          ret[product.pid] = { 'product' => product }
        end
        ret[product.pid]['versions'] = [] unless ret[product.pid].has_key?('versions')
        ret[product.pid]['versions'] << version
      end
      #ローカルプロダクト
      temp_lproducts_info = []
      lproducts = SidfmdLproduct.where(
        'LOWER(name) LIKE ? OR LOWER(name) LIKE ?', ActiveRecord::Base.send(:sanitize_sql_like, keyword.downcase) + '%','% ' + ActiveRecord::Base.send(:sanitize_sql_like, keyword.downcase) + '%'
      ).where(:flg_enable => true).order(:priority)
      lproducts.each do |p|
        temp_lproducts_info << {
          :product => p,
          :versions => p.sidfmd_lversions.where(:flg_enable => true).order(:priority)
        } 
      end
      ret['lproducts'] = temp_lproducts_info
    end
    render :layout => false, :json => ret.to_json
  end

  # hostnameに関連づけられたアプリケーション情報を返す
  def api_get_apps_by_hostname
    f_params = params.fetch(:params, {}).permit(:hostname)
    hostname = f_params[:hostname]
    ret = {}
    host = SidfmdHost.where(:hostname => hostname).first
    if host then
      apps = host.sidfmd_host_apps
      if apps then
        apps.each do |app|
          ret[app.product_id] = [] unless ret.has_key?(app.product_id)
          pid = app.product_id
          vid = app.version_id
          if !(ret[app.product_id].find {|x| x[:vid] == vid }) then
            name = ''
            fullname = ''
            if pid > 0 then
              product = SidfmdProduct.where(:pid => pid).first
              if product then
                name = product.name
              end
            end
            if vid && vid > 0 then
              version = SidfmdVersion.where(:vid => vid).first
              if version then
                fullname = version.fullname
              end
            end
            ret[app.product_id] << {
              :pid => pid, :vid => vid,
              :name => name, :fullname => fullname, }
          end
        end
      end
      #ローカルプロダクト
      temp_lproducts_info = {}
      lapps = host.sidfmd_host_lapps
      lapps.each do |lapp|
        pid = lapp.sidfmd_lproduct_id
        vid = lapp.sidfmd_lversion_id
        temp_lproducts_info[pid] = [] unless temp_lproducts_info.has_key?(pid)
        name = ''
        fullname = ''
        if pid > 0 then
          product = SidfmdLproduct.find_by_id(pid)
          if product then
            name = product.name
          end
        end
        if vid > 0 then
          version = SidfmdLversion.find_by_id(vid)
          if version then
            fullname = version.fullname
          end
        end 
        temp_lproducts_info[pid] << {
          :pid => pid, :vid => vid, :name => name, :fullname => fullname
        }
      end
      ret['lproducts'] = temp_lproducts_info
    end
    render :layout => false, :json => ret.to_json
  end
 
  #新規ホスト登録時のvalidation
  def api_validate_new_host
    all_hostnames = SidfmdHost.where(:flg_delete => false).map{|host| host.hostname}
    render :layout => false, :json => all_hostnames.to_json
  end
 

  # オフラインデータの取り込み
  def load_offline_data
    f_params = params.permit(:file_data, :fileId, :initialPreview, :initialPreviewConfig, :initialPreviewThumbTags, :system_category_id, :custom_category_id)
    file_data = f_params[:file_data]
    #fileId = f_params[:fileId]
    #initialPreview = f_params[:initialPreview]
    #initialPreviewConfig = f_params[:initialPreviewConfig]
    #initialPreviewThumbTags = f_params[:initialPreviewThumbTags]
    system_category_id = (f_params[:system_category_id] || SidfmdCategory.label2id(:public_server)).to_i
    custom_category_id = (f_params[:custom_category_id] || '0').to_i
    ret = {}
    hostname = nil
    host = nil
    agent_type = :unix
    # 利用するパラメータのチェック
    begin
      if file_data then
        file_data = JSON.parse((file_data.read || '{}'), :symbolize_names => true)
        hostname = file_data[:hostname]
        unless hostname then
          logger.error('LackHostname')
          ret[:error] = 'ファイルが不正です'
          render :layout => false, :json => ret.to_json and return
        end
        host = SidfmdHost.enable_hosts.where(:hostname => hostname).first
        if host and !@role_admin and !@user_host_ids.include?(host.id) then
          logger.error('RejectByRole')
          ret[:error] = 'アクセス権限がありません'
          render :layout => false, :json => ret.to_json and return
        end
        if file_data.has_key?(:kb) then
          agent_type = :windows
        end
      else
        logger.error('InvalidParameter')
        ret[:error] = 'ファイルが不正です'
        render :layout => false, :json => ret.to_json and return
      end
    rescue JSON::ParserError => e
      log_exception(e)
      ret[:error] = 'ファイルが不正です'
      render :lauout => false, :json => ret.to_json and return
    rescue Exception => e
      log_exception(e)
      ret[:error] = 'ファイルが不正です'
      render :layout => false, :json => ret.to_json and return
    end
    unless host then
      if @sidfm_roles.include?(:admin) then
        system_category = SidfmdCategory.find(system_category_id)
        if system_category.nil? or !system_category.is_system then
          logger.error('InvalidSystemCategoryId')
          ret[:error] = 'システムカテゴリが不正です'
          render :layout => false, :json => ret.to_json and return
        end
        if custom_category_id > 0 then
          custom_category = SidfmdCategory.find(custom_category_id)
          if custom_category.nil? or custom_category.is_system then
            logger.error('InvalidCustomCategoryId')
            ret[:error] = 'カスタムカテゴリが不正です'
            render :layout => false, :json => ret.to_json and return
          end
        end
      elsif @sidfm_roles.include?(:use) then
        system_category = SidfmdCategory.find(system_category_id)
        if system_category.nil? or !system_category.is_system then
          logger.error('InvalidSystemCategoryId')
          ret[:error] = 'システムカテゴリが不正です'
          render :layout => false, :json => ret.to_json and return
        end
        if custom_category_id > 0 then
          custom_category = SidfmdCategory.find(custom_category_id)
          if custom_category.nil? or custom_category.is_system then
            logger.error('InvalidCustomCategoryId')
            ret[:error] = 'カスタムカテゴリが不正です'
            render :layout => false, :json => ret.to_json and return
          end
          unless @user_category_ids.include?(custom_category_id) then
            logger.error('RejectByRole')
            ret[:error] = 'アクセス権限がありません'
            render :layout => false, :json => ret.to_json and return
          end
        else
          logger.error('InvalidCustomCategoryId')
          ret[:error] = 'カスタムカテゴリが不正です'
          render :layout => false, :json => ret.to_json and return
        end
      else
        logger.error('RejectByRole')
        ret[:error] = 'アクセス権限がありません'
        render :layout => false, :json => ret.to_json and return
      end
    end
    unless host then
      # ホスト登録処理
      # 登録APIキーのチェックは無視。
      host_apikey = SidfmdHost.register_host(hostname, nil, true, false)
      if host_apikey then
        SidfmdHost.transaction do
          host = SidfmdHost.where(:hostname => hostname).first
          unless host then
            logger.error('FailureRegistHost')
            ret[:error] = 'ホストの登録に失敗しました'
            render :layout => false, :json => ret.to_json and return
          end
          host.lock!
          host.update_attribute :flg_delete, false if host.flg_delete
          SidfmdCategoriesHost.where(:sidfmd_host_id => host.id).delete_all
          SidfmdCategoriesHost.create(:sidfmd_category_id => system_category_id,
                                      :sidfmd_host_id => host.id)
          if custom_category_id > 0 then
            SidfmdCategoriesHost.create(:sidfmd_category_id => custom_category_id,
                                        :sidfmd_host_id => host.id)
          end
        end
        Rails.cache.clear
      else
        logger.error('FailureRegistHost')
        ret[:error] = 'ホストの登録に失敗しました'
        render :layout => false, :json => ret.to_json and return
      end
    end
    # データ登録処理
    if agent_type == :windows then
      begin
        SidfmdHost.transaction do
          host.lock!
          host.update_attribute :is_agent, true unless host.is_agent
          host.create_queue_win(file_data)
        end
        SidfmdHostView.sync_data([host.id])
      rescue Exception => e
        log_exception(e)
        ret[:error] = e.message
        render :layout => false, :json => ret.to_json and return
      end
    elsif agent_type == :unix then
      begin
        SidfmdHost.transaction do
          host.lock!
          host.update_attribute :is_agent, true unless host.is_agent
          host.create_queue(file_data)
        end
        SidfmdHostView.sync_data([host.id])
      rescue Exception => e
        log_exception(e)
        ret[:error] = e.message
        render :layout => false, :json => ret.to_json and return
      end
    else
      logger.error('LastResort')
      ret[:error] = 'システムの不具合'
      render :layout => false, :json => ret.to_json and return
    end
    render :layout => false, :json => ret.to_json
  end

  def set_last_checked_at 
    f_params = params.fetch(:form, {}).permit(:time_with_hostid)
    begin 
      SidfmdHost.transaction do 
        data = JSON.parse(f_params[:time_with_hostid])
        hostid = data['hostid'].to_i
        time = data['time'].to_i
        host = SidfmdHost.find(hostid)
        if time == -1 then
          set_time = nil
        else
          if time < 10 then
            time = '0' + time.to_s
          end
          last_time = host.last_checked_at
          if last_time then
            last_time = last_time.to_time
            set_time = (last_time.to_s.slice(0, 11) + time.to_s + ':00:00 +0900').to_time
            if set_time - last_time < 0 then
              set_time = set_time + 24*60*60
            end
            check_time = set_time - Time.current.to_time
            if check_time > 0 && (check_time / 60) / 60 >= 24 then
              set_time = set_time - 24*60*60
            end
          else
            set_time = (Date.today.yesterday.to_time.to_s.slice(0, 11) + time.to_s + ':00:00 +0900').to_time
            if ((Time.current.to_time - set_time) / 60) / 60 >= 24 then
              set_time = set_time + 24*60*60
            end
          end 
        end
        host.update_attribute :last_checked_at, set_time
      end
    rescue Exception => e
      log_exception(e)
      redirect_to :action => :index_r, :notice => '同期時間の設定に失敗しました。' and return
    end
    Rails.cache.clear
    redirect_to :action => :index_r and return
  end

  #同期時間の確認
  def api_get_new_last_checked_at
    f_params = params.permit(:time_with_hostid)
      data = JSON.parse(f_params[:time_with_hostid])
      hostid = data['hostid'].to_i
      time = data['time'].to_i
      host = SidfmdHost.find(hostid)
      if time == -1 then
        set_time = 'now'
        synchro_time = 'now'
      else
        if time < 10 then
          time = '0' + time.to_s
        end
        last_time = host.last_checked_at
        if last_time then
          last_time = last_time.to_time
          set_time = (last_time.to_s.slice(0, 11) + time.to_s + ':00:00 +0900').to_time
          if set_time - last_time < 0 then
            set_time = set_time + 24*60*60
          end
          check_time = set_time - Time.current.to_time
          if check_time > 0 && (check_time / 60) / 60 >= 24 then
            set_time = set_time - 24*60*60
          end
        else
          set_time = (Date.today.yesterday.to_time.to_s.slice(0, 11) + time.to_s + ':00:00 +0900').to_time
          if ((Time.current.to_time - set_time) / 60) / 60 >= 24 then
            set_time = set_time + 24*60*60
          end
        end
        hour = set_time.hour.to_s
        synchro_time = (set_time + 24*60*60).strftime("%Y年%m月%d日 " + hour + "時")
        set_time = set_time.strftime("%Y年%m月%d日 " + hour + "時")
      end
    render :layout => false, :json => [set_time, synchro_time].to_json
  end

end
