#!/usr/bin/env ruby
# -*- coding: utf-8 -*-

require 'json'
require 'uri'
require 'openssl'
require 'net/http'

require 'vm_config'

module SIDfm

  class SIDfmdlmAPI

    HEADER_LICENSE_ID = 'X-License-Id'.freeze
    HEADER_LICENSE_APIKEY = 'X-Api-Key'.freeze

    private

    def initialize(base_url, license_key, apikey = nil)
      @base_url = URI.parse(base_url[-1] == '/' ? base_url + '/' : base_url)
      @license_key = license_key
      @license_apikey = apikey
      @vm_config = SIDfm::VMConfig.new
      @cert_file = '/var/lib/redmine/plugins/sidfm-vm-server/ROR/config/cert.pem'
    end

    def connection(url)
      proxy = @vm_config.get_proxy(url.scheme == 'https')
      http = proxy.new(url.host, url.port)
      # http/https以外は排除
      raise ArgumentError if url.scheme !~ /^https?$/
      http.use_ssl = (url.scheme == 'https')
      #@http.set_debug_output(@debug_out)
      http.open_timeout = 60
      http.read_timeout = 60
      http.keep_alive_timeout = 60
      http
    end

    # URLのpathとqueryだけを抽出した文字列を返す。
    def url_path_query(url)
      url.select(:path, :query).join('?')
    end

    # HTTP GETリクエスト発行
    # 応答は応答内容の文字列をJSONフォーマットとしてデコードしたもの。
    # アクセス失敗や応答文字列の不正等はすべてnilとなる。
    def get_request(path = '')
      ret = nil
      begin
        url = @base_url + path
        ret = JSON.parse(connection(url).get(url, { 'Content-Type' => 'application/json; charset=UTF-8' }), :symbolize_names => true)
      rescue Exception => e
        File.open('/var/lib/redmine/log/sidfmdlm.log', 'a') do |f|
          f.puts e.message
          if e.backtrace then
            e.backtrace do |i|
              f.puts i
            end
          end
        end
      end
      ret
    end

    # HTTP POSTリクエスト発行
    # 応答は応答内容の文字列をJSONフォーマットとしてデコードしたもの。
    # アクセス失敗や応答文字列の不正等はすべてnilとなる。
    def post_request(path = '', form_data = {}, headers = {})
      ret = nil
      url = @base_url + path
      req = connection(url)
      res = req.post(url_path_query(url), form_data.keys.collect {|x| "#{x}=#{form_data[x]}" }.join('&'), headers)
      if res.respond_to?(:body) and res.body then
        begin
          ret = JSON.parse(res.body, :symbolize_names => true)
        rescue Exception => e
          File.open('/var/lib/redmine/log/sidfmdlm.log', 'a') do |f|
            f.puts e.message
            if e.backtrace then
              e.backtrace.each do |i|
                f.puts i
              end
            end
          end
        end
      end
      ret
    end

    # HTTP POSTリクエスト発行。なお、事前にauthを実行して成功している必要がある。
    # 応答は応答内容の文字列をJSONフォーマットとしてデコードしたもの。
    # アクセス失敗や応答文字列の不正等はすべてnilとなる。
    def post_request_with_auth(path = '', form_data = {})
      res = false
      if @license_key and @license_apikey then
        res = post_request(path, form_data,
                           { HEADER_LICENSE_ID => @license_key,
                             HEADER_LICENSE_APIKEY =>  @license_apikey })
      end
      res
    end

    public

    # 到達確認
    def reachable?
      ret = false
      res = get_request('./test')
      ret = res[:success] if res.respond_to?(:[])
      ret
    end

    # 認証
    def auth(license_password)
      res = post_request('./auth',
                         { 'license_id' => @license_key,
                           'license_password' => license_password})
      if res.respond_to?(:[]) and res[:success] then
        @license_apikey = res[:license_apikey]
      end
      res
    end

    ### 以降は認証済を前提 ###

    # ライセンス管理用APIKEYを更新する。
    # 更新に成功した場合、内部のAPIKEYを自動的に更新するが、
    # 失敗した場合はAPIKEYを削除するので、再認証が必要になることに注意。
    def update_apikey
      res = post_request_with_auth('./update_apikey')
      if res and res.respond_to?(:[]) and res[:success] then
        @license_apikey = res[:license_apikey]
      end
      res
    end

    # 登録済ホストの一覧を返す。
    def get_hosts(opts = {})
      res = post_request_with_auth('./get_hosts', opts)
      res
    end

    # ホストの削除フラグを変更する。
    # 削除されているときは復元、削除されていないときは削除する動作をする。
    # なお、本システムでは削除はフラグをたてるだけで実際に削除はしない。
    def toggle_host_delete(id)
      res = post_request_with_auth('./toggle_host_delete',
                                   { :id => id })
      res
    end

    # ホストのAPIKEYを更新する。
    # 値は指定できない。
    def update_host_apikey(id)
      res = post_request_with_auth('./update_host_apikey',
                                   { :id => id })
      res
    end

    # ホスト名を更新する。
    # ライセンス内でユニークである必要がある。
    def rename_host(id, hostname)
      res = post_request_with_auth('./rename_host',
                                   { :id => id, :hostname => hostname })
      res
    end

    # ホストを登録する。
    def create_host(hostname)
      res = post_request_with_auth('./create_host',
                                   { :hostname => hostname })
      res
    end

  end

end

if $0 == __FILE__ then

  #url = 'http://devsidfmd1.softek.co.jp:81/sidfmdlm/'
  url = 'http://artemis.softek.co.jp:10081/sidfmdlm/'
  o =  SIDfm::SIDfmdlmAPI.new(url, 'aaaa')
  puts '#reachable?'
  p o.reachable?
  puts '#auth(aaaaaaaa)'
  p o.auth('aaaaaaaa')
  puts '#update_apikey(aaaa)'
  p o.update_apikey
  puts '#get_hosts'
  p o.get_hosts(:'filters[hostname]' => 'ccc')
  p o.get_hosts(:offset => 2, :limit => 1)
  p o.get_hosts(:'sort[hostname]' => :desc)
  puts '#toggle_host_delete'
  p o.toggle_host_delete(2)
  puts '#update_host_apikey'
  p o.update_host_apikey(2)
  puts '#rename_host'
  p o.rename_host(2, 'eeee')
  p o.rename_host(2, 'ddde')
  p o.rename_host(2, 'dddd')
  puts '#create_host(gggg)'
  p o.create_host('gggg')
end
